﻿using System;
using System.Collections.Generic;
using System.Text;

namespace AvPcl
{
    public class CardHandler
    {
        //#####################################################################
        #region Member
        private int m_Hdl = -1;
        private InterruptDelegate m_RegisteredDelegate = null;
        #endregion

        //#####################################################################
        #region Types
        public enum ReturnCode
        {
            Ok = 0, // function successfully completed
            NoNewData = 1, // no new data received

            ErrorInvalidPort = -1, // invalid port id
            ErrorNotFound = -2, // pcl hardware not found
            ErrorNoHandle = -3, // no free handle found
            ErrorInvalidHandle = -4, // invalid handle
            ErrorNotOpen = -5, // handle not opened by pclOpen()
            ErrorNull = -6, // null pointer assignment
            ErrorTimeout = -7, // timeout while waiting for pcl ready
            ErrorInvalidRegNum = -8, // invalid register number
            ErrorInvalidBank = -9, // invalid transmit bank number
            ErrorRange = -10, // parameter out of range
            ErrorInit = -11, // can't initialize library
            ErrorIoctl = -12, // error in device driver while ioctl
            ErrorVersion = -13, // version of dll and sys different
            ErrorNotSupported = -14, // function not supported
            ErrorUnknownCmd = -15, // unknown command
            ErrorOpenWinDriver = -16, // can't open WinDriver lib
        };

        public enum BaseAddress
        {
            PciOffset = 0x1000,
            Pci01 = 0x1001,
            Pci02 = 0x1002,
            Pci03 = 0x1003,
            Pci04 = 0x1004,
            Pci05 = 0x1005,
            Pci06 = 0x1006,
            Pci07 = 0x1007,
            Pci08 = 0x1008,
            Pci09 = 0x1009,
            Pci10 = 0x100A,
            PciLast = Pci10,
        };

        public enum InterruptMask
        {
            None = 0x0000,
            Mixed = 0x0001,
            Ltc = 0x0004,
            Vitc = 0x0008,
            V1 = 0x0040,
            V2 = 0x0080,

            AtcLtc = 0x0100,
            AtcVitc = 0x0200,

            // Obsolete
//          AtcVitc1 = 0x0200,
            AtcVitc2 = 0x0400,
            AtcHancLtc = 0x0800,
            AtcHancVitc = 0x1000,
        };

        public enum Register
        {
            Mixed = 0x00,
            Ltc = 0x02,
            Vitc = 0x03,
            AtcLtc = 0x11,
            AtcVitc = 0x12,
            None = 0xFF,

            // Obsolete
            AtcVitc1 = 0x12,
            AtcVitc2 = 0x13,
            HancLtc = 0x14,
            HancVitc = 0x15,
        };

        #endregion

        //#####################################################################
        #region Construction/Destruction and Overrides

        ~CardHandler()
        {
            if (m_Hdl > -1) Close();
        }

        public override string ToString()
        {
        //    string s = m_BaseAddr.ToString();

        //    if (m_Firmware != 0)
        //    {
        //        switch ((m_Firmware >> 6) & 0x3)
        //        {
        //            case 0: s += " PCL3/PCL6"; break;
        //            case 1: s += " PCL PCI(e)"; break;
        //            case 2: s += " PCL10/PCL5"; break;
        //            case 3: s += " PCL12"; break;
        //        }

        //        s += " V" + ((m_Firmware >> 4) & 0x3) + "." // main version
        //                   + (m_Firmware & 0xF);             // sub version
        //    }

        //    return s;
            return base.ToString();
        }
        #endregion

        //###############################################################
        #region Handling - Open, Close

        public ReturnCode Open(BaseAddress baseAddress)
        {
            ReturnCode ret = (ReturnCode)Dll.pclOpen(out m_Hdl, (uint)baseAddress);

            if (ret == ReturnCode.Ok)
            {
                int ident, firmware, dll;
                Dll.pclGetVersion(m_Hdl, out ident, out firmware, out dll);
                if (dll < Dll.PCL_VER_DLL)
                {
//                  throw new AvException(ReturnCode.ErrorVersion);
                }
            }

            return ret;
        }

        public ReturnCode Close()
        {
            ReturnCode ret = (ReturnCode)Dll.pclClose(m_Hdl);
            if (ret == ReturnCode.Ok) m_Hdl = -1;
            return ret;
        }

        #endregion

        //###############################################################
        #region Getters
        public Timecode GetTc()
        {
            UInt32 tcBcd = GetTcBcd();
            Timecode tc = new Timecode();
            tc.SetBcd(tcBcd);
            return tc;
        }

        public Timecode GetRegister(Register reg)
        {
            UInt32 tcBcd = GetRegisterBcd(reg);
            Timecode tc = new Timecode();
            tc.SetBcd(tcBcd);
            return tc;
        }

        public UInt32 GetTcBcd()
        {
            UInt32 tcBcd;
            ReturnCode ret = (ReturnCode)Dll.pclGetTc(m_Hdl, out tcBcd);
            if (ret == ReturnCode.Ok)
            {
                return tcBcd;
            }
            throw new AvException(ret);
        }

        public UInt32 GetRegisterBcd(Register reg)
        {
            UInt32 tcBcd;
            int bits, frame, newd;
            UInt32 user;

            ReturnCode ret = (ReturnCode)Dll.pclGetRegister(m_Hdl, (int)reg, out tcBcd,
                out bits, out user, out frame, out newd);
            if (ret == ReturnCode.Ok)
            {
                return tcBcd;
            }
            throw new AvException(ret);
        }

        public ReturnCode GetUser(out UInt32 user)
        {
            return (ReturnCode)Dll.pclGetUser(m_Hdl, out user);
        }
        #endregion

        //###############################################################
        #region Interrupts
        /// <summary>Set hardware interrupt mask</summary>
        /// <param name="iMask">
        /// Interrupt mask: Combine masks from desired bits. See
        /// AvPcl.AvPclIntMask.*
        /// </param>
        public ReturnCode IntMask(int iMask)
        {
            return (ReturnCode)Dll.pclIntMask(m_Hdl, iMask);
        }

        /// <summary>Set interrupt bits</summary>
        /// <param name="iMask">
        /// Combine desired bits from AvPcl.AvPclIntMask.*
        /// </param>
        public ReturnCode SetIntMask(InterruptMask iMask)
        {
            return (ReturnCode)Dll.pclSetIntMask(m_Hdl, (int)iMask);
        }

        /// <summary>Reset interrupt bits</summary>
        /// <param name="iMask">
        /// Combine desired bits from AvPcl.AvPclIntMask.*
        /// </param>
        public ReturnCode ResetIntMask(InterruptMask iMask)
        {
            return (ReturnCode)Dll.pclResetIntMask(m_Hdl, (int)iMask);
        }

//      public delegate void AvIntDelegate(int hdl, int flag, Dll.PclIntResult intResult);
        public delegate void InterruptDelegate(CardHandler card, int flag, Dll.PclIntResult intResult);

        /// <summary>Register Interrupt Handler</summary>
        /// <param name="del">
        /// Delegate to be called upon interrupt event
        /// </param>
        public ReturnCode SetIntHandler(InterruptDelegate del)
        {
            m_RegisteredDelegate = del;
            Dll.AvIntDelegate del1 = new Dll.AvIntDelegate(DelegateWrapper);
            return (ReturnCode)Dll.SetIntHandler(m_Hdl, del1);
        }

        private void DelegateWrapper(int hdl, int flag, Dll.PclIntResult intResult)
        {
            m_RegisteredDelegate(this, flag, intResult);
        }

        #endregion
    }
}
