/***************************************************************************\
*																			*
*	Macros and function prototypes of Windows DLL 'AVPCL32'					*
*	and of Linux Kernel Driver for PCL PCI(e) Devices						*
*																			*
*	Copyright (c) 1995-2023 by PLURA Europe GmbH							*
*																			*
\***************************************************************************/

/* $Id: avpcl.h 2123 2023-10-05 14:32:37Z J.Loh $ */
/* Version: 5.45 */

#ifndef _AV_AVPCL_H
#define _AV_AVPCL_H

/***************************************************************************/

/*	Features
	--------

	This DLL allows access to the following products of	PLURA:

	PCL PCIe 3G (v2):			ATC/DVITC/LTC reader for 3G and HD video with PCIe interface
	PCL PCIe 3G:				ATC/DVITC/LTC reader for 3G and HD video with PCIe interface
	PCL PCIe HD:				ATC/DVITC/LTC reader for HD video with PCIe interface
	PCL PCIe D:					ATC/DVITC/LTC reader with PCIe interface
	PCL PCIe LV (v2):			VITC/LTC reader with PCIe interface
	PCL PCIe LV:				VITC/LTC reader with PCIe interface
	PCL PCIe L (v2):			LTC reader with PCIe interface
	PCL PCIe L:					LTC reader with PCIe interface
	PCL PCIe TS (v2):			Time Synchronisation Board
	PCL PCIe TS:				Time Synchronisation Board

	PCL PCI HD (3.3V/5V):		ATC/DVITC/LTC reader for HD video with 3.3V/5V PCI interface
	PCL PCI D (v2) (3.3V/5V):	ATC/DVITC/LTC reader with 3.3V/5V PCI interface
	PCL PCI D (3.3V/5V):		DVITC/LTC reader with 3.3V/5V PCI interface
	PCL PCI D:					DVITC/LTC reader with 5V PCI interface
	PCL PCI LV (3.3V/5V):		VITC/LTC reader with 3.3V/5V PCI interface
	PCL PCI LV:					VITC/LTC reader with 5V PCI interface
	PCL PCI TS (3.3V/5V):		Time Synchronisation Board
	PCL PCI TS:					Time Synchronisation Board
	PCL PCI L (v2) (3.3V/5V):	LTC reader with 3.3V/5V PCI interface
	PCL PCI L (3.3V/5V):		LTC reader with 3.3V/5V PCI interface

****************************************************************************/

#if defined(_Windows) || defined(_WIN32)
#include <windows.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/***************************************************************************/

/*	Basics
	------

	AVPCL exports many #defines, typedefs and functions.  To avoid naming
	conflicts, all identifiers start with a prefix: 'PCL_' for #defines
	and typedefs and 'pcl' for functions.  Identifiers starting with
	a underbar (_) are for internal use only. */

/***************************************************************************/

/*	DLL Version
	-----------
	Version of this DLL at compile time.  Compare this with the DLL
	version returned by	pclGetVersion() at run time to be sure not to
	use an outdated version of the DLL. */

#define PCL_VER_DLL (0x0500 + 45) /* 5.45 */

/***************************************************************************/

/*	Program options
	---------------

	This options enable special features of AVPCL.  To use any of these,
	a special PCL PCI(e) firmware is needed. */

#define PCL_OPT_REALTIME		/* pcl as real-time source */
#define PCL_OPT_GPI_OUT			/* event driven gpi outputs */
#define PCL_OPT_GPI_IN			/* gpi inputs */
#define PCL_OPT_GPI_IN_EVNT		/* gpi input events */

/* #define PCL_OPT_GETCMD */	/* get command bytes from device */
/* #define PCL_OPT_MTD */		/* read mtd times */

/***************************************************************************/

/* #define PCL_OBSOLETE_VER2 */
					/* include obsolete functions of driver version 2.xx */

#ifdef _WIN32
#	define PCL_OPT_VBI			/* read vbi lines */
#	define PCL_OPT_ANC			/* read user defined ancillary data */
#endif

/***************************************************************************/

/*	Error codes
	-----------

	Most functions return an int error code.  A value greater than or
	equal to zero means 'no error', negative values mean 'error'.
	A list of possible return codes is shown below: */

#define PCL_OK					  0 /* function successfully completed */

#define PCL_NO_NEWD				  1 /* no new data received */

#define PCL_ERR_INV_PORT		 -1 /* invalid port id */
#define PCL_ERR_NOT_FOUND		 -2 /* pcl hardware not found */
#define PCL_ERR_NO_HANDLE		 -3 /* no free handle found */
#define PCL_ERR_INV_HDL			 -4 /* invalid handle */
#define PCL_ERR_NOTOPEN			 -5 /* handle not opened by pclOpen() */
#define PCL_ERR_NULL			 -6 /* null pointer assignment */
#define PCL_ERR_TIMEOUT			 -7 /* timeout while waiting for pcl ready */
#define PCL_ERR_INV_REG			 -8 /* invalid register number */
#define PCL_ERR_INV_BANK		 -9 /* invalid transmit bank number */
#define PCL_ERR_RANGE			-10 /* parameter out of range */
#define PCL_ERR_INIT			-11 /* can't initialize library */
#define PCL_ERR_IOCTL			-12 /* error in device driver while ioctl */
#define PCL_ERR_VER				-13 /* version of dll and sys different */
#define PCL_ERR_NOT_SUPP		-14 /* function not supported */
#define PCL_ERR_UK_CMD			-15 /* unknown command */
#define PCL_ERR_OPEN			-16 /* can't open WinDriver lib */
#define PCL_ERR_MULTI_NOT_SUPP	-17 /* multi-mode not supported */

/***************************************************************************/

/*	Handles
	-------

	To access the PCL PCI(e) card, you need to open a communication port with
	pclOpen() described later.  It is possible to handle several PCL PCI(e)
	cards simultaneous, you need to open a communication handle for each
	PCL PCI(e) card.
	Here you find the number of handle available: */

#define PCL_HANDLES				(PCL_CARDS * 5) /* number of handles available */

/*	Cards
	-----
	The driver is able to support a limited number of PCL PCI(e) cards.
	Here you find the number of cards available: */

									/* number of supported PCL PCI(e) cards */
#define PCL_CARDS				(PCL_BASE_PCI_LAST - PCL_BASE_PCI_OFS)

/*	Access modes
	------------

	There are three different modes to access a timecode reader register.
	This modes are named PCL_REQ_xxx: */

/*		PCL_REQ_xxx						timecode request mode */
#define PCL_REQ_ASYN			0x80 /* asynchron */
#define PCL_REQ_SYNC			0x40 /* synchron */
#define PCL_REQ_PERM			0x20 /* permanent */

/***************************************************************************/

/*	BCD/Hex conversion
	------------------

	Several parameters of the PCL PCI(e) card are coded BCD.  To convert
	binary numbers to BCD and vice versa, you might find the following macros
	useful: */

#define BCD2HEX(bcd) (((bcd) & 0x0F) + ((((bcd) & 0xF0) >> 4) * 10))
								/* convert unsigned char from bcd to hex */

#define HEX2BCD(hex) ((((hex) / 10) << 4) + ((hex) % 10))
								/* convert unsigned char from hex to bcd */

/*	Some Macros from windows.h
	-------------------------- */

#ifndef MAKEWORD
#define MAKEWORD(a, b)	((WORD)(((BYTE)(a)) | ((WORD)((BYTE)(b))) << 8))
#endif

#ifndef MAKELONG
#define MAKELONG(a, b)	((LONG)(((WORD)(a)) | ((DWORD)((WORD)(b))) << 16))
#endif

#ifndef LOWORD
#define LOWORD(l)		((WORD)(l))
#endif

#ifndef HIWORD
#define HIWORD(l)		((WORD)(((DWORD)(l) >> 16) & 0xFFFF))
#endif

#ifndef LOBYTE
#define LOBYTE(w)		((BYTE)(w))
#endif

#ifndef HIBYTE
#define HIBYTE(w)		((BYTE)(((WORD)(w) >> 8) & 0xFF))
#endif

/***************************************************************************/

#ifdef LINUX
#	define LINUX_64BIT (__SIZEOF_POINTER__ == 8)
#	if defined(PCL_USE_SYSV_SEM)
		int gSemId;
#		define SEM_PROCESS_COUNT (0)
#		define SEM_LIB (1)
#		define SEM_FIRST_CARD (2)
		int lockTimeout(int semID, int semNo, unsigned int sec, unsigned int nsec);
		void unlock(int semID, int semNum);
#	endif
#else
#	define LINUX_64BIT 0
#endif

/*	Typedefs
	-------- */

/* Some Typedefs from windows.h */

#if !defined(WINAPI) || defined(_NTDDK_)

#ifndef _BYTE_DEFINED
#define _BYTE_DEFINED
typedef unsigned char BYTE;						/* size is 1 */
#endif

#ifndef _WORD_DEFINED
#define _WORD_DEFINED
typedef unsigned short WORD;					/* size is 2 */
#endif

#ifndef _UINT_DEFINED
#define _UINT_DEFINED
typedef unsigned int UINT;							/* size is int */
#endif /* !_UINT_DEFINED */

#ifndef _DWORD_DEFINED
#define _DWORD_DEFINED
#if LINUX_64BIT
typedef unsigned int DWORD;						/* size is 4 */
#else
typedef unsigned long DWORD;					/* size is 4 */
#endif
#endif

#endif

#ifndef _BOOL_DEFINED
#define _BOOL_DEFINED
typedef int BOOL;
#endif

typedef int PCL_BOOL;

typedef DWORD PCL_TC;		/* time code, bcd */

/*	Time code is coded BCD in a double word.
	0xhhmmssff with hh = hours, mm = minutes, ss = seconds, ff = frames
	e.g. 0x12345621 means 12:34:56:21 */

typedef DWORD PCL_USER;		/* user-bits, bcd */

/*	User-Bits are coded similar to time code described above. */

typedef BYTE  PCL_FULL[16]; /* full pcl register */

/*	Pcl data corresponds to data $0 to $F of any register of PCL PCI(e) card */

/***************************************************************************/

/*	Calling conventions
	------------------- */

#if defined(_WINNT) || defined(_NTDDK_) || defined(_WIN32)
#	define _PCL_FDEF	_stdcall
#	define _PCL_IN
#	define _PCL_OUT
#elif LINUX_64BIT
#	define _PCL_FDEF	__attribute__((sysv_abi))
#	define _PCL_IN
#	define _PCL_OUT
#elif defined(LINUX)
#	define _PCL_FDEF	__attribute__((cdecl))
#	define _PCL_IN
#	define _PCL_OUT
#else /* legacy 16-bit */
#	define _PCL_FDEF	_far _pascal
#	define _PCL_IN		_far
#	define _PCL_OUT		_far
#endif

/***************************************************************************\
*																			*
* Function Prototypes														*
*																			*
\***************************************************************************/

/*	The functions are described as followed:

	int _PCL_FDEF pclRecStat(	/# @44,	$91, recorder status request #/
		int hdl,				/# in:	handle from pclOpen() #/
		BYTE _PCL_OUT *stat,	/# out:	recorder status, like PCL_STAT #/
		int _PCL_OUT *ident		/# out:	recorder identification flag #/
	);

	'@44'  is the decimal dll function number as defined in the *.def file.
	'$91'  is the hex function number as described in the PCL PCI(e) manual.
	'in:'  are parameters passing something *to* the function.
	'out:' are parameters passing something *from* the function. */

/*-------------------------------------------------------------------------*/

#if !defined(_Windows) && !defined(LINUX)

int _PCL_FDEF pclInitLib(	/* initialize library avpcl */
	void					/* no parameters needed */
);

#endif

/* Initialization is only needed with DOS drivers.  Windows DLLs are
	initialized automatically */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclOpen(	/* @2,	open pcl communication channel */
	int _PCL_OUT *hdl,  /* out:	handle to access channel */
	unsigned base		/* in:	base address of pcl card, like
								PCL_BASE_xxx */
);

#define PCL_BASE_PCI_OFS	0x1000 /* offset to address PCL PCI(e) cards */
#define PCL_BASE_PCI_1		(PCL_BASE_PCI_OFS +  1) /*  1st PCL PCI(e) card */
#define PCL_BASE_PCI_2		(PCL_BASE_PCI_OFS +  2) /*  2nd PCL PCI(e) card */
#define PCL_BASE_PCI_3		(PCL_BASE_PCI_OFS +  3) /*  3rd PCL PCI(e) card */
#define PCL_BASE_PCI_4		(PCL_BASE_PCI_OFS +  4) /*  4th PCL PCI(e) card */
#define PCL_BASE_PCI_5		(PCL_BASE_PCI_OFS +  5) /*  5th PCL PCI(e) card */
#define PCL_BASE_PCI_6		(PCL_BASE_PCI_OFS +  6) /*  6th PCL PCI(e) card */
#define PCL_BASE_PCI_7		(PCL_BASE_PCI_OFS +  7) /*  7th PCL PCI(e) card */
#define PCL_BASE_PCI_8		(PCL_BASE_PCI_OFS +  8) /*  8th PCL PCI(e) card */
#define PCL_BASE_PCI_9		(PCL_BASE_PCI_OFS +  9) /*  9th PCL PCI(e) card */
#define PCL_BASE_PCI_10		(PCL_BASE_PCI_OFS + 10) /* 10th PCL PCI(e) card */
#define PCL_BASE_PCI_11		(PCL_BASE_PCI_OFS + 11) /* 11th PCL PCI(e) card */
#define PCL_BASE_PCI_12		(PCL_BASE_PCI_OFS + 12) /* 12th PCL PCI(e) card */
#define PCL_BASE_PCI_13		(PCL_BASE_PCI_OFS + 13) /* 13th PCL PCI(e) card */
#define PCL_BASE_PCI_14		(PCL_BASE_PCI_OFS + 14) /* 14th PCL PCI(e) card */
#define PCL_BASE_PCI_15		(PCL_BASE_PCI_OFS + 15) /* 15th PCL PCI(e) card */
#define PCL_BASE_PCI_16		(PCL_BASE_PCI_OFS + 16) /* 16th PCL PCI(e) card */
#define PCL_BASE_PCI_17		(PCL_BASE_PCI_OFS + 17) /* 17th PCL PCI(e) card */
#define PCL_BASE_PCI_18		(PCL_BASE_PCI_OFS + 18) /* 18th PCL PCI(e) card */
#define PCL_BASE_PCI_19		(PCL_BASE_PCI_OFS + 19) /* 19th PCL PCI(e) card */
#define PCL_BASE_PCI_20		(PCL_BASE_PCI_OFS + 20) /* 20th PCL PCI(e) card */
#define PCL_BASE_PCI_LAST	PCL_BASE_PCI_20

#define PCL_BASE_M				0x1FFF /* mask for base address */
#define PCL_BASE_OPTION_MULTI	0x2000 /* enable multi-access to card */

/*	Before you can use any of the pclXxx() functions, you have to get a
	handle by calling pclOpen().  Use this handle to call subsequent
	pclXxx functions.  Don't forget to call pclClose() before your
	program terminates.

	Example: Simple program to read a single timecode value

		#include <stdio.h>
		#include "avpcl.h"
		int main(void)
		{
			int hdl;
			PCL_TC time;
			if (pclOpen(&hdl, PCL_BASE_PCI_1) == PCL_OK) {
				pclGetTc(hdl, &time);
				printf("Current time code is: %8lx\n", time);
				pclClose(hdl);
				return 0;
			} else {
				puts("Can't open PCL PCI(e) card");
				return 1;
			}
		}
*/

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclClose(	/* @3,	close pcl communication channel */
	int hdl				/* in:	handle from pclOpen() */
);

/*	pclClose() terminates the connection to the pcl card.  See pclOpen()
	for a example. */

/*---------------------------------------------------- read mixed register */

int _PCL_FDEF pclGetTc(		/* @45,	$41, read mixed-register timecode
									asynchronous */
	int hdl,				/* in:	handle from pclOpen() */
	PCL_TC _PCL_OUT *time	/* out:	timecode from mixed-register, bcd */
);

/*	This is the simplest way to get the time from the mixed register,
	but it isn't the most efficient, because it sends a request to the pcl
	card and waits for the response.  If you need to get a timecode
	permanetly, it's better to use pclGetRegSync. */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetUser(	/* @46,	$42, read mixed-register user-bits asynchr. */
	int hdl,				/* in:	handle from pclOpen() */
	PCL_USER _PCL_OUT *user	/* out:	user-bits from mixed-register */
);

/*	The simplest way to get the user bits from the mixed register of the
	pcl card.  See pclGetTc for details. */

/*------------------------------------------------- read timecode register */

int _PCL_FDEF pclGetRegister(	/* @47,	$43, read register asynchronous */
	int hdl,					/* in:	handle from pclOpen() */
	int reg,					/* in:	registers number like PCL_REG_xxx */
	PCL_TC _PCL_OUT *time,		/* out:	timecode from register, bcd */
	int _PCL_OUT *bits,			/* out:	misc. bits of register,
										meaning depends of '*newd',
										like PCL_BITS_xxx_xxx */
	PCL_TC _PCL_OUT *user,		/* out:	user-bits from register, bcd */
	int _PCL_OUT *frame,		/* out:	frame rate as detected by pcl,
										like PCL_FRAME_xxx */
	int _PCL_OUT *newd			/* out:	identifier of timecode read,
										like PCL_NEWD_xxx */
);

/* Note:  Set several 'out' parameters to NULL if you don't need them */

/*		PCL_REG_xxx					register numbers */
#define PCL_REG_MIXED		0x00 /* mixed register */
#define PCL_REG_LTC			0x02 /* ltc register */
#define PCL_REG_VITC		0x03 /* vitc register */
#define PCL_REG_ATC_LTC		0x11 /* atc ltc register */
#define PCL_REG_ATC_VITC	0x12 /* atc vitc register */
#define PCL_REG_HANC_LTC	0x14 /* hanc ltc register */
#define PCL_REG_HANC_VITC	0x15 /* hanc vitc register */
#define PCL_REG_ATC_VITC1	0x12 /* obsolete: atc vitc 1 register */
#define PCL_REG_ATC_VITC2	0x13 /* obsolete: atc vitc 2 register */
#define PCL_REG_NONE		0xFF /* invalid register */

/*		PCL_FRAME_xxx				timecode frame rate from register */
#define PCL_FRAME_UNKNOWN	0x00 /* unknown */
#define PCL_FRAME_24		0x24 /* 24 */
#define PCL_FRAME_25		0x25 /* 25 */
#define PCL_FRAME_30		0x30 /* 30 */
#define PCL_FRAME_30DF		0xB0 /* 30 drop (29.97) */

/*		PCL_NEWD_xxx				timecode new data flag */
#define PCL_NEWD_NONE		0x00 /* no new data */
#define PCL_NEWD_VITC		0x80 /* vitc */
#define PCL_NEWD_LTC		0x40 /* ltc */

/*		PCL_BITS_xxx_xxx			timecode bits */

/*		PCL_BITS_xxx				generic bits */
#define PCL_BITS_DROP		0x01 /* drop */
#define PCL_BITS_CF			0x02 /* cf */
#define PCL_BITS_WHEEL		0x40 /* fly wheel mode */

/*		PCL_BITS_LTC_xxx			ltc bits */
#define PCL_BITS_LTC_10		0x01 /* bit 10 */
#define PCL_BITS_LTC_DROP	0x01 /* drop */
#define PCL_BITS_LTC_11		0x02 /* bit 11 */
#define PCL_BITS_LTC_CF		0x02 /* cf */
#define PCL_BITS_LTC_27		0x04 /* bit 27 */
#define PCL_BITS_LTC_43		0x08 /* bit 43 */
#define PCL_BITS_LTC_58		0x10 /* bit 58 */
#define PCL_BITS_LTC_59		0x20 /* bit 59 */
#define PCL_BITS_LTC_WHEEL	0x40 /* fly wheel mode */
#define PCL_BITS_LTC_DIR	0x80 /* direction */

/*		PCL_BITS_VITC_xxx			vitc bits */
#define PCL_BITS_VITC_14	0x01 /* bit 14 */
#define PCL_BITS_VITC_DROP	0x01 /* drop */
#define PCL_BITS_VITC_15	0x02 /* bit 15 */
#define PCL_BITS_VITC_CF	0x02 /* cf */
#define PCL_BITS_VITC_35	0x04 /* bit 35 */
#define PCL_BITS_VITC_NF	0x04 /* ntsc field mark */
#define PCL_BITS_VITC_55	0x08 /* bit 55 */
#define PCL_BITS_VITC_74	0x10 /* bit 74 */
#define PCL_BITS_VITC_75	0x20 /* bit 75 */
#define PCL_BITS_VITC_PF	0x20 /* pal field mark */
#define PCL_BITS_VITC_WHEEL	0x40 /* fly wheel mode */

/*	If you need additional informations from any timecode register of
	the pcl card, you can use pclGetRegister().

	pclGetRegister(hdl, PCL_REG_MIXED, &time, NULL, NULL, NULL, NULL);
	is equivalent to pclGetTc(hdl, &time). */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetRegSync(	/* @48,	$43, read register synchronous */
	int hdl,					/* in:	handle from pclOpen() */
	int reg,					/* in:	registers number like PCL_REG_xxx */
	PCL_TC _PCL_OUT *time,		/* out:	timecode from register, bcd */
	int _PCL_OUT *bits,			/* out:	misc. bits of register,
										meaning depends of '*newd',
										like PCL_BITS_xxx_xxx */
	PCL_TC _PCL_OUT *user,		/* out:	user-bits from register, bcd */
	int _PCL_OUT *frame,		/* out:	frame rate as detected by pcl,
										like PCL_FRAME_xxx */
	int _PCL_OUT *newd			/* out:	identifier of timecode read,
										like PCL_NEWD_xxx */
);

/*	Note:
	Set several 'out' parameters to NULL if you don't need them */

/*	pclGetRegSync doesn't wait for the pcl to
	respond.  pclGetRegSync first checks if there is a new timecode in the
	specified register.  If no, it returns PCL_NO_NEWD.  If there is
	a new timecode, it reads the value, sends a new request to
	the pcl and returns PCL_OK.  pclGetRegSync never waits for the pcl card!

	If there is a call to pclGetTc() similar to this...

		while (running) {
			PCL_TC time;
			pclGetTc(hdl, &time);
			doAnything(time)
		}

	...it might be replaced with something like that:

		pclGetRegSync(hdl, PCL_REG_MIXED, NULL, NULL, NULL, NULL, NULL)
		while (running) {
			PCL_TC time;
			if (
				pclGetRegSync(
					hdl,
					PCL_REG_MIXED, &time,
					NULL, NULL, NULL, NULL
				) == PCL_OK
			) {
				doAnything(time);
			} else {
				doAnythingElse();
			}
		}

	Please note that pclGetRegSync() only returns PCL_OK if there was a new
	timecode read in the specified register.  This is opposite to
	pclGetTc() which allways returns a timecode.  In the example above,
	with pclGetTc() doAnything() is called in every loop, but with
	pclGetRegSync() it is only called if there was a new timecode read
	by the pcl card.  The first call to pclGetRegSync() outside the loop
	gurantees that doAnything() isn't called with a timecode that was
	read before the program starts. */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetRegStatus(	/* @84,	$44, read register status */
	int hdl,					/* in:	handle from pclOpen() */
	int reg,					/* in:	registers number like PCL_REG_xxx */
	int _PCL_OUT *stat			/* out:	status bytes, like PCL_STAT_xxx */
);

#define	PCL_STAT_DBB1						0x00FF	/* ATC DBB1 mask */
#define	PCL_STAT_DBB1_LTC					0x0000	/* ATC Longitudinal time code */
#define	PCL_STAT_DBB1_VITC1					0x0001	/* ATC Vertical interval time code #1 */
#define	PCL_STAT_DBB1_VITC2					0x0002	/* ATC Vertical interval time code #2 */
#define	PCL_STAT_DBB1_FILM_DB_TRANS			0x0006	/* ATC Film data block (transferred) */
#define	PCL_STAT_DBB1_PROD_DB_TRANS			0x0007	/* ATC Production data block (transferred) */
#define	PCL_STAT_DBB1_VTAPE_DB_LOC			0x007D	/* ATC Video tape data block (locally generated) */
#define	PCL_STAT_DBB1_FILM_DB_LOC			0x007E	/* ATC Film data block (locally generated) */
#define	PCL_STAT_DBB1_PROD_DB_LOC			0x007F	/* ATC Production data block (locally generated) */

#define	PCL_STAT_DBB2						0xFF00	/* ATC DBB2 mask */
#define	PCL_STAT_DBB2_VITC_LINE_SELECT		0x1F00	/* ATC VITC line select */
#define	PCL_STAT_DBB2_VITC_LINE_DUP			0x2000	/* ATC VITC line duplication */
#define	PCL_STAT_DBB2_TC_VALIDITY			0x4000	/* ATC time code validity */
#define	PCL_STAT_DBB2_PROCESS_BIT			0x8000	/* ATC (User bits) process bits */

#define	PCL_STAT_HANC						0xFF00	/* HANC VITC status mask */
#define	PCL_STAT_HANC_VITC_LINE_SELECT		0x3F00	/* HANC VITC line select */
#define	PCL_STAT_HANC_VITC_LINE_SELECT_DIV	0x0100	/* HANC VITC line select divisor */
#define	PCL_STAT_HANC_VITC_LINE_DUP			0x4000	/* HANC VITC line duplication */
#define	PCL_STAT_HANC_VALIDITY				0x8000	/* HANC VITC validity flag */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetVersion(	/* @59,	$46, read pcl firmware version number */
	int hdl,					/* in:	handle from pclOpen() */
	int _PCL_OUT *ident,		/* out:	software identification, != 0
										means special design */
	int _PCL_OUT *firmware,		/* out:	pcl firmware revision number:

										bit 76543210
											--------
											00			PCL3/PCL6
											01			reserved
											10			PCL10/PCL5
											11			PCL12
											  xx		main version, 0..3
												xxxx	sub version,  0..15 */

	int _PCL_OUT *dll			/* out: dll revision number:

											111111
										bit 543210987654321
											---------------
											xxxxxxxx		main version
													xxxxxxx sub version  */
);

/*	Note: Set several 'out' parameters to NULL if you don't need them */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetVerSpecial(	/* @66,	$47, read special pcl firmware version */
	int hdl,					/* in:	handle from pclOpen() */
	int _PCL_OUT *ident1,		/* out:	identifier from byte $A */
	int _PCL_OUT *ident2,		/* out:	identifier from byte $B */
	int _PCL_OUT *ident3		/* out:	identifier from byte $C */
);

/*	Note: Set several 'out' parameters to NULL if you don't need them */

/***************************************************************************\
*																			*
* Commands																	*
*																			*
\***************************************************************************/

/* Please refer to chapter C1 (description of commands) for details. */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclReset(		/* @5,	$01, reset: initialize pcl card */
	int hdl,				/* in:	handle from pclOpen() */
	int imode				/* in:	operation mode of pcl, like
									PCL_IMODE_xxx */
);

/*		PCL_IMODE_xxx				operation mode of pcl card */
#define PCL_IMODE_NORMAL	0x00 /* normal mode */
#define PCL_IMODE_POR		0x80 /* force power-on reset */
#define PCL_IMODE_TEST		0xF0 /* reserved (test mode) */

/*	Use pclReset() if you want to change the operating mode of the pcl card.
	It also performs a soft reset, resetting all register values to
	their power-on values. */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclIntMask(	/* @7,	$03, interrupts: set hw interrupt mask */
	int hdl,				/* in:	handle from pclOpen() */
	int imask				/* in:	interrupt mask, like PCL_IMASK_xxx */
);

/*		PCL_IMASK_xxx				interrupt mask */
#define PCL_IMASK_NONE		0x0000 /* disable all interrups */
#define PCL_IMASK_MIXED		0x0001 /* mix disable/enable */
#define PCL_IMASK_LTC		0x0004 /* ltc disable/enable */
#define PCL_IMASK_VITC		0x0008 /* vitc disable/enable */
#define PCL_IMASK_V1		0x0040 /* odd video field */
#define PCL_IMASK_V2		0x0080 /* even video field */
#define PCL_IMASK_ATC_LTC	0x0100 /* atc ltc disable/enable */
#define PCL_IMASK_ATC_VITC	0x0200 /* atc vitc disable/enable */
#define PCL_IMASK_HANC_LTC	0x0800 /* hanc ltc disable/enable */
#define PCL_IMASK_HANC_VITC	0x1000 /* hanc vitc disable/enable */
#define PCL_IMASK_VBI		0x2000 /* vbi disable/enable */
#define PCL_IMASK_ANC		0x4000 /* anc disable/enable */
#define PCL_IMASK_GPI		0x8000 /* gpi disable/enable */

/* obsolete */
#define PCL_IMASK_ATC_VITC1	PCL_IMASK_ATC_VITC /* obsolete: atc vitc 1 disable/enable */
#define PCL_IMASK_ATC_VITC2	0x0400 /* obsolete: atc vitc 2 disable/enable */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetIntMask(	/* @8,	$04, set interrupt bits */
	int hdl,					/* in:	handle from pclOpen() */
	int imask					/* in:	interrupt mask, like PCL_IMASK_xxx */
);

/*	Set one or more bits of the interrupt mask, i.e. enable it. */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclResetIntMask(	/* @9,	$05, reset interrupt bits */
	int hdl,					/* in:	handle from pclOpen() */
	int imask					/* in:	interrupt mask, like PCL_IMASK_xxx */
);

/*	Reset one or more bits of the interrupt mask, i.e. disable it. */

/*-------------------------------------------------------------------------*/

#if LINUX_64BIT
#	define PCL_CALLBACK	__attribute__((sysv_abi))
#elif defined LINUX
#	define PCL_CALLBACK	__attribute__((cdecl))
#else /* Windows */
#	define PCL_CALLBACK	__cdecl
#endif

typedef struct
{
	DWORD counter;	/* number of interrupts received */
	DWORD lost; 	/* number of interrupts not yet dealt with */
	BOOL stopped;	/* interrupt was disabled during wait */
} PCL_INT_RESULT;

#if defined(_Windows) || defined(_WIN32) || defined(LINUX)

typedef void (PCL_CALLBACK *PCL_INT_HANDLER) (	/* prototype of interrupt
													handler */
	int hdl,							/* in:	handle from pclOpen() */
	int flag,							/* in:	interrupt flag, like
												PCL_IMASK_xxx */
	const PCL_INT_RESULT *intResult		/* in:	interrupt result */
);

int _PCL_FDEF pclSetIntHandler(	/* @69,	set interrupt handler */
	int hdl,					/* in:	handle from pclOpen() */
	PCL_INT_HANDLER intHandler	/* in:	address of interrupt handler */
);

/*	Set the interrupt handler routine.  That's a callback function
	that is called for every hardware interrupt.  See the declaration of
	PCL_INT_HANDLER above.

	The handler gets a PCL PCI(e) Handle (hdl), a interrupt request flag (flag)
	and a pointer to a interrupt information structure (intResult).  See
	declaration of PCL_INT_RESULT above.  With that informations the
	handler is able to get new data code from the source of	the interrupt.

	If several interrupts requests are pending simultaneously, the
	handler is called several times, each time with another flag bit set,
	MSB first.

	There is no need to acknowledge the interrupt request.  This is done
	by the PCL PCI(e) driver automatically.
*/

#endif

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclIntAck(	/* @70,	$48, interrupt acknowledge */
	int hdl,				/* in:	handle from pclOpen() */
	int imask				/* in:	interrupt mask, like PCL_IMASK_xxx */
);

/*	Under normal conditions, this function is never needed, because
	the PCL PCI(e) driver acknowledges every interrupt automatically.
*/

/*-------------------------------------------------------------------------*/

#ifdef _WIN32

int _PCL_FDEF pclSetIntPriority(	/* @71, set interrupt priority */
   int hdl,							/* in:	handle from pclOpen() */
   int priority						/* in:	priority of interrupt handler */
);

/*	Set the priority of the interrupt handler thread.  Setting priority
	to THREAD_PRIORITY_TIME_CRITICAL is a possible way to make the
	interrupt handling of Windows more strict to the real world.
	Refer to SetThreadPriority() in the Win32 documentation for details.
*/

#endif

/*-------------------------------------------------------------------------*/

#ifdef _WIN32

int _PCL_FDEF pclGetThreadHandle(	/* @90, get thread handle */
   int hdl,							/* in:	handle from pclOpen() */
   int thread,						/* in:	thread to use, like PCL_THREAD_xxx */
   HANDLE _PCL_OUT *phHandle		/* out:	handle of thread */
);

/*		PCL_THREAD_xxx				thread number */
#define PCL_THREAD_INT		0x0001 /* interrupt thread */
#define PCL_THREAD_EVENT	0x0002 /* event handler thread */

#endif

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclReaderEnable(	/* @10,	$10, reader on/off */
	int hdl,					/* in:	handle from pclOpen() */
	int rmask					/* in:	reader mask, like PCL_RMASK_xxx */
);

/*		PCL_RMASK_xxx				reader mask */
#define PCL_RMASK_NONE		0x0000 /* disable all readers */
#define PCL_RMASK_LTC		0x0002 /* ltc off/on */
#define PCL_RMASK_VITC		0x0004 /* vitc off/on */
#define PCL_RMASK_ATC_LTC	0x0100 /* atc ltc off/on */
#define PCL_RMASK_ATC_VITC	0x0200 /* atc vitc off/on */
#define PCL_RMASK_HANC_LTC	0x0800 /* hanc ltc off/on */
#define PCL_RMASK_HANC_VITC	0x1000 /* hanc vitc off/on */
#define PCL_RMASK_ALL		0x1B06 /* enable all readers */
#define PCL_RMASK_ATC_VITC1	0x0200 /* obsolete: atc vitc1 off/on */
#define PCL_RMASK_ATC_VITC2	0x0400 /* obsolete: atc vitc2 off/on */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclVitcSetup(	/* @65,	$14, vitc setup */
	int hdl,				/* in:	handle from pclOpen() */
	int line1,				/* in:	1st line */
	int line2,				/* in:	2nd line */
	int vline				/* in:	vitc line mode, like PCL_VLINE_xxx */
);

/*		PCL_VLINE_xxx				vitc line mode */
#define PCL_VLINE_2LINES	0x00 /* 2-lines mode */
#define PCL_VLINE_BLOCK		0x01 /* block mode */

/*	Note: pclVitcSetup is not supported by PCL PCI(e) firmware version less or
		equal V1.D */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclMixedEnable(	/* @11,	$20, mixed register: reader on/off */
	int hdl,					/* in:	handle from pclOpen() */
	int rmask					/* in:	reader mask, like PCL_RMASK_xxx */
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclPriority(	/* @12,	$21, priority: mixed reg. tc priority */
	int hdl,				/* in:	handle from pclOpen() */
	int prior				/* in:	priority, like PCL_PRIOR_xxx */
);

/*		PCL_PRIOR_xxx				timecode priority */
#define PCL_PRIOR_VLC		0x01 /* vitc-ltc-ctl */
#define PCL_PRIOR_LVC		0x02 /* ltc-vitc-ctl */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclTcFrames(	/* @13,	$30, tc frame rate */
	int hdl,				/* in:	handle from pclOpen() */
	int tcfr				/* in:	frame rate, like PCL_TCFR_xxx */
);

/*		PCL_TCFR_xxx				tc frame rate */
#define PCL_TCFR_AUTO			0x00 /* automatic */
#define PCL_TCFR_24				0x01 /* 24 frames */
#define PCL_TCFR_25				0x02 /* 25 frames */
#define PCL_TCFR_30				0x03 /* 30 frames non drop mode */
#define PCL_TCFR_30DF			0x04 /* 30 frames drop mode */
#define PCL_TCFR_INIT_AUTO		0x05 /* automatic, initialize with unknown */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclLtcFrames(	/* @76,	$31, set ltc frame rate */
	int hdl,				/* in:	handle from pclOpen() */
	int ltcfr				/* in:	ltc frame rate, like PCL_TCFR_xxx */
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclVitcFrames(	/* @77,	$32, set vitc frame rate */
	int hdl,					/* in:	handle from pclOpen() */
	int vitcfr					/* in:	vitc frame rate, like PCL_TCFR_xxx */
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclVitcLevelControl(	/* @72,	$16, vitc level control */
	int hdl,						/* in:	handle from pclOpen() */
	int vlmode,						/* in:	vitc level control mode,
											like PCL_VLMODE_xxx */
	int vlev,						/* in:	vitc level, binary,
											like PCL_VLEX_xxx */
	int _PCL_OUT *vlret				/* out:	vitc level return, binary,
											like PCL_VLEX_xxx */
);

/*		PCL_VLMODE_xxx					vitc level control mode */
#define PCL_VLMODE_AUTO			0x00 /* automatic (default) */
#define PCL_VLMODE_AUTO_SET		0x01 /* automatic with preset value */
#define PCL_VLMODE_AUTO_INIT	0x02 /* automatic with initialization */
#define PCL_VLMODE_OFF			0x03 /* switch automatic off */
#define PCL_VLMODE_OFF_SET		0x04 /* switch automatic off, preset value */

/*		PCL_VLEV_xxx					vitc level range */
#define PCL_VLEV_MIN			0x00 /* minimum value */
#define PCL_VLEV_MAX			0x45 /* maximum value */

/* Note:  Set 'out' parameter to NULL if you don't need it */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclLtcSequence(	/* @63,	$11, ltc sequence */
	int hdl,					/* in:	handle from pclOpen() */
	int lseq					/* in:	ltc sequence, like PCL_LSEQ_xxx */
);

/*		PCL_LSEQ_xxx				ltc sequence */
#define PCL_LSEQ_CHECK		0x00 /* sequence check on */
#define PCL_LSEQ_NOCHECK	0x01 /* sequence check off */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclLtcError(	/* @73,	$17, get ltc error counter */
	int hdl,				/* in:	handle from pclOpen() */
	int ecmode,				/* in:	error counter mode, like PCL_ECMODE_xxx */
	int _PCL_OUT *ecval		/* out:	error counter, 8 bit binary */
);

/*		PCL_ECMODE_xxx					error counter mode */
#define PCL_ECMODE_INITIALIZE	0x00 /* get error counter, initialize counter afterwards */
#define PCL_ECMODE_GET			0x01 /* get error counter, leave it unchanged */
#define PCL_ECMODE_GET_RESET	0x02 /* get error counter, reset it afterwards */

/* Note:  Set 'out' parameter to NULL if you don't need it */

/* Note:  To use PCL_ECMODE_GET_RESET, you need PCL PCI(e) Firmware Version 4
		  or higher */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclVitcError(	/* @74,	$18, get vitc error counter */
	int hdl,				/* in:	handle from pclOpen() */
	int ecmode,				/* in:	error counter mode, like PCL_EMODE_xxx */
	int _PCL_OUT *ecval		/* out:	error counter, 8 bit binary */
);

/* Note:  Set 'out' parameter to NULL if you don't need it */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclFlyWheel(	/* @75,	$19, set flywheel mode */
	int hdl,				/* in:	handle from pclOpen() */
	int fwmode				/* in:	fly wheel mode, like PCL_FWMODE_xxx */
);

/*		PCL_FWMODE_xxx				fly wheel mode bit mask */
#define PCL_FWMODE_OFF		0x00 /* disable all fly wheel counters */
#define PCL_FWMODE_LTC		0x02 /* enable ltc fly wheel counter */
#define PCL_FWMODE_VITC		0x04 /* enable vitc fly wheel counter */
#define PCL_FWMODE_ALL		0x06 /* enable both fly wheel counters */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclAtcLtcError(	/* @85,	$22, get atc ltc error counter */
	int hdl,					/* in:	handle from pclOpen() */
	int ecmode,					/* in:	error counter mode, like PCL_ECMODE_xxx */
	int _PCL_OUT *ecval			/* out:	error counter, 8 bit binary */
);

/* Note:  Set 'out' parameter to NULL if you don't need it */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclAtcVitcError(	/* @86,	$23, get atc vitc1 error counter */
	int hdl,					/* in:	handle from pclOpen() */
	int ecmode,					/* in:	error counter mode, like PCL_EMODE_xxx */
	int _PCL_OUT *ecval			/* out:	error counter, 8 bit binary */
);

/* Note:  Set 'out' parameter to NULL if you don't need it */

#define pclAtcVitc1Error	pclAtcVitcError	/* Obsolete, for backward compatibility only */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclAtcVitc2Error(	/* @87,	$24, obsolete: get atc vitc2 error counter */
	int hdl,					/* in:	handle from pclOpen() */
	int ecmode,					/* in:	error counter mode, like PCL_EMODE_xxx */
	int _PCL_OUT *ecval			/* out:	error counter, 8 bit binary */
);

/* Obsolete, for backward compatibility only */
/* Note:  Set 'out' parameter to NULL if you don't need it */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclHancLtcError(	/* @88,	$25, get hanc ltc error counter */
	int hdl,					/* in:	handle from pclOpen() */
	int ecmode,					/* in:	error counter mode, like PCL_ECMODE_xxx */
	int _PCL_OUT *ecval			/* out:	error counter, 8 bit binary */
);

/* Note:  Set 'out' parameter to NULL if you don't need it */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclHancVitcError(	/* @89,	$26, get hanc vitc error counter */
	int hdl,					/* in:	handle from pclOpen() */
	int ecmode,					/* in:	error counter mode, like PCL_EMODE_xxx */
	int _PCL_OUT *progressive	/* out:	error counter, 8 bit binary */
);

/* Note:  Set 'out' parameter to NULL if you don't need it */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetDID(	/* @92,	$49, set DID of anc register */
	int hdl,				/* in:	handle from pclOpen() */
	int _PCL_IN ancreg,		/* in:	anc register, like PCL_ANCREG_xxx */
	int _PCL_IN did			/* in:	DID */
);

#define		PCL_ANCREG_ATC			0x00	/* ATC LTC and ATC VITC */
#define		PCL_ANCREG_HANC_LTC		0x01	/* HANC LTC */
#define		PCL_ANCREG_HANC_VITC	0x02	/* HANC VITC */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetSDID(	/* @93,	$4A, set SDID of anc register */
	int hdl,				/* in:	handle from pclOpen() */
	int _PCL_IN ancreg,		/* in:	anc register, like PCL_ANCREG_xxx */
	int _PCL_IN sdid		/* in:	SDID */
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetVideoStandard(	/* @91,	$4B, get video standard */
	int hdl,						/* in:	handle from pclOpen() */
	int _PCL_OUT *system,			/* out:	video system, like PCL_VSYSTEM_xxx */
	int _PCL_OUT *progressive		/* out:	0 = interlaced, else progressive video mode */
);

/* Note:  Set 'out' parameter to NULL if you don't need it */

			/* Analog (black+burst) or digital (SDI) SD video standards */

#define		PCL_VSYSTEM_UNKNOWN_SD		 0		/* Unknown SD video */
#define		PCL_VSYSTEM_625I_50			 1		/*  625i50 */
#define		PCL_VSYSTEM_525I_59			 2		/*  525i59.94 */

			/* Digital (SDI) HD video standards */

#define		PCL_VSYSTEM_1080I_PSF_50	 3		/* 1080i50 or psF 25 */
#define		PCL_VSYSTEM_1080I_PSF_59_60	 4		/* 1080i59.94 or 60 or psF 29.97 or 30 */
#define		PCL_VSYSTEM_1080PSF_23_24	 5		/* 1080psF23.98 or 24 */
#define		PCL_VSYSTEM_720P_50			 6		/*  720p50 */
#define		PCL_VSYSTEM_720P_59_60		 7		/*  720p59.94 or 60 */
#define		PCL_VSYSTEM_1080P_25		 8		/* 1080p25 */
#define		PCL_VSYSTEM_1080P_29_30		 9		/* 1080p29.97 or 30 */
#define		PCL_VSYSTEM_1080P_23_24		10		/* 1080p23.98 or 24 */
#define		PCL_VSYSTEM_720P_23_24		11		/*  720p23.98 or 24 */
#define		PCL_VSYSTEM_720P_25			12		/*  720p25 */
#define		PCL_VSYSTEM_720P_29_30		13		/*  720p29.97 or 30 */
#define		PCL_VSYSTEM_1035I_59_60		14		/* 1035i59.94 or 60 */
#define		PCL_VSYSTEM_UNKNOWN_HD		15		/* Unknown HD video */

			/* Digital (3G-SDI) HD video standards */

#define		PCL_VSYSTEM_UNKNOWN_3G		16		/* Unknown 3G video */
#define		PCL_VSYSTEM_1080P_50		17		/* 1080p50, Level A */
#define		PCL_VSYSTEM_1080P_59_60		18		/* 1080p59.94 or 60, Level A */
#define		PCL_VSYSTEM_DL_1080P_50		39		/* 1080p50, Level B */
#define		PCL_VSYSTEM_DL_1080P_59		40		/* 1080p59.94, Level B */
#define		PCL_VSYSTEM_DL_1080P_60		41		/* 1080p60, Level B */
#define		PCL_VSYSTEM_1080I_PSF_59	42		/* 1080i59.94 or psF 29.97 */
#define		PCL_VSYSTEM_1080I_PSF_60	43		/* 1080i 60 or psF 30 */
#define		PCL_VSYSTEM_1035I_59		44		/* 1035i59.94 */
#define		PCL_VSYSTEM_1035I_60		45		/* 1035i60 */
#define		PCL_VSYSTEM_1080P_47_48		48		/* 1080p47.95 or 48, Level A */
#define		PCL_VSYSTEM_DL_1080P_47_48	51		/* 1080p47.95 or 48, Level B */

			/* Analog SD (black+burst) video standards */

#define		PCL_VSYSTEM_625P_50			52		/*  625p50 */
#define		PCL_VSYSTEM_525P_59			53		/*  525p59.94 */
#define		PCL_VSYSTEM_UNKNOWN			54		/* Unknown video */

			/* Analog HD (tri-level) video standards */

#define		PCL_VSYSTEM_1080I_47_48_T	55		/* 1080i47.95 or 48 or psF 23.98 or 24, Tri-Level */
#define		PCL_VSYSTEM_1080P_47_48_T	56		/* 1080p47.95 or 48, Tri-Level */
#define		PCL_VSYSTEM_1080P_50_T		57		/* 1080p50, Tri-Level */
#define		PCL_VSYSTEM_1080P_59_60_T	58		/* 1080p59.94 or 60, Tri-Level */

			/* The following video standards can only be detected */
			/* with the help of SMPTE 352M Payload Identifier packets. */

#define		PCL_VSYSTEM_720P_23			19		/*  720p23.98 */
#define		PCL_VSYSTEM_720P_24			20		/*  720p24 */
#define		PCL_VSYSTEM_720P_29			21		/*  720p29.97 */
#define		PCL_VSYSTEM_720P_30			22		/*  720p30 */
#define		PCL_VSYSTEM_720P_59			23		/*  720p59.94 */
#define		PCL_VSYSTEM_720P_60			24		/*  720p60 */
#define		PCL_VSYSTEM_1080I_50		25		/* 1080i50 */
#define		PCL_VSYSTEM_1080I_59		26		/* 1080i59.94 */
#define		PCL_VSYSTEM_1080I_60		27		/* 1080i60 */
#define		PCL_VSYSTEM_1080PSF_23		28		/* 1080psF23.98 */
#define		PCL_VSYSTEM_1080PSF_24		29		/* 1080psF24 */
#define		PCL_VSYSTEM_1080PSF_25		30		/* 1080psF25 */
#define		PCL_VSYSTEM_1080PSF_29		31		/* 1080psF29.97 */
#define		PCL_VSYSTEM_1080PSF_30		32		/* 1080psF30 */
#define		PCL_VSYSTEM_1080P_23		33		/* 1080p23.98 */
#define		PCL_VSYSTEM_1080P_24		34		/* 1080p24 */
#define		PCL_VSYSTEM_1080P_29		35		/* 1080p29.97 */
#define		PCL_VSYSTEM_1080P_30		36		/* 1080p30 */
#define		PCL_VSYSTEM_1080P_59		37		/* 1080p59.94, Level A */
#define		PCL_VSYSTEM_1080P_60		38		/* 1080p60, Level A */
#define		PCL_VSYSTEM_1080P_47		46		/* 1080p47.95, Level A */
#define		PCL_VSYSTEM_1080P_48		47		/* 1080p48, Level A */
#define		PCL_VSYSTEM_DL_1080P_47		49		/* 1080p47.95, Level B */
#define		PCL_VSYSTEM_DL_1080P_48		50		/* 1080p48, Level B */

			/* The following definitions are obsolete */

#define		PCL_VSYSTEM_525I_29			PCL_VSYSTEM_525I_59
#define		PCL_VSYSTEM_625I_25			PCL_VSYSTEM_625I_50
#define		PCL_VSYSTEM_1035I_29_30		PCL_VSYSTEM_1035I_59_60
#define		PCL_VSYSTEM_1080I_PSF_25	PCL_VSYSTEM_1080I_PSF_50
#define		PCL_VSYSTEM_1080I_PSF_29_30	PCL_VSYSTEM_1080I_PSF_59_60
#define		PCL_VSYSTEM_1080I_25		PCL_VSYSTEM_1080I_50
#define		PCL_VSYSTEM_1080I_29		PCL_VSYSTEM_1080I_59
#define		PCL_VSYSTEM_1080I_30		PCL_VSYSTEM_1080I_60

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetVideoRaster(	/* @94,	$4C, get video raster informations */
	int hdl,						/* in:	handle from pclOpen() */
	int _PCL_OUT *activeWords,		/* out:	active words / line */
	int _PCL_OUT *totalWords,		/* out:	total words / line */
	int _PCL_OUT *activeLines,		/* out:	active lines / frame */
	int _PCL_OUT *totalLines		/* out:	total lines / frame */
);

/* Note:  Set 'out' parameter to NULL if you don't need it */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetPayloadIdentifier(	/* @105, $50, get payload identifier (SMPTE 352M) */
	int hdl,							/* in:	handle from pclOpen() */
	BYTE _PCL_OUT *pbPayloadIdentifier,	/* out:	payload identifier, 4 bytes */
	int _PCL_OUT *flags					/* out:	S352M flags, like PCL_GET_S352M_xxx */
);

/* Note:  Set 'out' parameter to NULL if you don't need it */

#define		PCL_GET_S352M_VALID_M		0x01	/* a valid S352M packed is present in the video */
#define		PCL_GET_S352M_STD_ERR_M		0x02	/* the S352M packet conflicts with the detected video geometry */
#define		PCL_GET_S352M_USE_M			0x04	/* use S252M packets to detect the video standard */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetUsePayloadIdentifier(	/* @106, $51, get payload identifier (SMPTE 352M) */
	int hdl,								/* in:	handle from pclOpen() */
	BYTE _PCL_IN flags						/* in:	S352M flags, like PCL_SET_xxx */
);

/* Note:  Set 'out' parameter to NULL if you don't need it */

#define		PCL_SET_S352M_USE_M			0x04	/* use S252M packets to detect the video standard */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetMultiMode(	/* @101, get multi operating mode */
	int hdl,					/* in:	handle from pclOpen() */
	int _PCL_OUT *multi			/* out:	multi operating mode, like PCL_MULTI_MODE_xxx */
);

#define		PCL_MULTI_MODE_NONE		0	/* not in multi mode */
#define		PCL_MULTI_MODE_SLAVE	1	/* multi slave mode */
#define		PCL_MULTI_MODE_MASTER	2	/* multi master mode */

/***************************************************************************\
*																			*
* Internal functions														*
*																			*
\***************************************************************************/

/* The following functions are internal to AVPCL32.DLL.  Not data is
	transfered to or from PCL PCI(e) card. */

/*-------------------------------------------------------------------------*/

PCL_TC _PCL_FDEF pclTc2Bcd(	/* @50,	convert binary timecode to packed bcd */
	int hours,				/* in:	hours of binary timecode */
	int minutes,			/* in:	minutes of binary timecode */
	int seconds,			/* in:	seconds of binary timecode */
	int frames				/* in:	frames of binary timecode */
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclBcd2Tc(	/* @49,	convert packed bcd timecode to binary */
	PCL_TC time,			/* in:	packed bcd timecode */
	int _PCL_OUT *hours,	/* out:	hours of binary timecode */
	int _PCL_OUT *minutes,	/* out:	minutes of binary timecode */
	int _PCL_OUT *seconds,	/* out:	seconds of binary timecode */
	int _PCL_OUT *frames	/* out:	frames of binary timecode */
);

/* Note:  Set several 'out' parameters to NULL if you don't need them */

/*-------------------------------------------------------------------------*/

long _PCL_FDEF pclTc2Frames(	/* @60,	convert packed bcd timecode to long */
	PCL_TC time,				/* in:	packed bcd timecode */
	int frate					/* in:	frame rate (25 or 30) */
);

/*-------------------------------------------------------------------------*/

const char * _PCL_FDEF pclGetErrorString(	/* @68, get error string */
	void									/* no parameter needed */
);

/*	Note: The function returns a pointer to a error string, if the error
	happens in the low-level part of the device driver.  Otherwise
	it returns a pointer to an empty string. */

/*-------------------------------------------------------------------------*/

#ifdef LINUX
	typedef struct timespec PCL_TIME_STAMP;
#else
	typedef LARGE_INTEGER PCL_TIME_STAMP;
#endif // LINUX

/*-------------------------------------------------------------------------*/

#if defined(_Windows) || defined(_WIN32)

typedef struct {
	int size;
	HANDLE hWD;
	struct
	{
		DWORD dwVer;
		CHAR cVer[100];
	} sWdVersion;
	WORD wVendorID;
	WORD wDeviceID;
	WORD wSubsystemVendorID;
	WORD wSubsystemDeviceID;
	void *intHandler;
	int chipVer;
	struct {
		PCL_TIME_STAMP abs;
		PCL_TIME_STAMP freq;
	} perf;
} PCL_DRIVER_INFORMATION;

int _PCL_FDEF pclGetDriverInformation(		/* @78, get driver information */
	int hdl,								/* in:	handle from pclOpen() */
	PCL_DRIVER_INFORMATION _PCL_OUT *info	/* out:	information data structure */
);

/*	Note: The function returns internal informations about the driver.
	For normal operation, this function is not needed.
	However, you may use the field wSubsystemDeviceID to detect to
	what type of card you're connected.
*/

#endif

#define AVPCL_VENDOR_ID			0x10B5u				/* wVendorID */

													/* wDeviceID */
#define AVPCL_DEVICE_ID_5V		0x9050u				/* PCI 5V */
#define AVPCL_DEVICE_ID_3V		0x9030u				/* PCI 3.3V */
#define AVPCL_DEVICE_ID_E		0x9056u				/* PCIe */

#define AVPCL_SUBVENDOR_ID		AVPCL_VENDOR_ID		/* wSubsystemVendorID */

													/* wSubsystemDeviceID */
#define AVPCL_SUBSYSTEM_DEVICE_ID_L_3V		0x3025u	/* PCL PCI L (3.3V/5V) */
#define AVPCL_SUBSYSTEM_DEVICE_ID_L_V2_3V	0x3591u	/* PCL PCI L (v2) (3.3V/5V) */
#define AVPCL_SUBSYSTEM_DEVICE_ID_L_PCIE	0x3355u	/* PCL PCIe L */

#define AVPCL_SUBSYSTEM_DEVICE_ID_LV_5V		0x2221u	/* PCL PCI LV */
#define AVPCL_SUBSYSTEM_DEVICE_ID_LV_3V		0x2862u	/* PCL PCI LV (3.3V/5V) */
#define AVPCL_SUBSYSTEM_DEVICE_ID_LV_PCIE	0x3354u	/* PCL PCIe LV */

#define AVPCL_SUBSYSTEM_DEVICE_ID_D_5V		0x2431u	/* PCL PCI D */
#define AVPCL_SUBSYSTEM_DEVICE_ID_D_3V		0x2940u	/* PCI D (3.3V/5V) */
#define AVPCL_SUBSYSTEM_DEVICE_ID_D_V2_3V	0x3463u	/* PCL PCI D (v2) (3.3V/5V) */
#define AVPCL_SUBSYSTEM_DEVICE_ID_D_PCIE	0x3353u	/* PCL PCIe D */

#define AVPCL_SUBSYSTEM_DEVICE_ID_HD_3V		0x3068u	/* PCL PCI HD (3.3V/5V) */
#define AVPCL_SUBSYSTEM_DEVICE_ID_HD_PCIE	0x3352u	/* PCL PCIe HD */

#define AVPCL_SUBSYSTEM_DEVICE_ID_3G_PCIE	0x3493u	/* PCL PCIe 3G */

#define AVPCL_SUBSYSTEM_DEVICE_ID_TS_5V		0x2905u	/* PCL PCI TS */
#define AVPCL_SUBSYSTEM_DEVICE_ID_TS_3V		0x2906u	/* PCL PCI TS (3.3V/5V) */
#define AVPCL_SUBSYSTEM_DEVICE_ID_TS_PCIE	0x3415u	/* PCL PCIe TS */

#define AVPCL_REVISION_PCIE					0xBAu	/* PCL PCIe */
#define AVPCL_REVISION_PCIE_V2				0x80u	/* PCL PCIe (v2) */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetIdata(	/* @79, $38, get idata */
	int hdl,				/* in:	handle from pclOpen() */
	int addr,				/* in:	idata address */
	int _PCL_OUT *idata		/* out:	byte read from idata */
);

/*	Note: The function returns internal informations about the driver.
	For normal operation, this function is never needed. */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetPdata(	/* @80, $3A, get pdata */
	int hdl,				/* in:	handle from pclOpen() */
	int addr,				/* in:	pdata address */
	int _PCL_OUT *pdata		/* out:	byte read from pdata */
);

/*	Note: The function returns internal informations about the driver.
	For normal operation, this function is never needed. */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetXdata(	/* @83, $3C, get xdata */
	int hdl,				/* in:	handle from pclOpen() */
	int addr,				/* in:	xdata address */
	int _PCL_OUT *xdata		/* out:	byte read from xdata */
);

/*	Note: The function returns internal informations about the driver.
	For normal operation, this function is never needed. */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclScanCards(		/* @104, scan pcl cards */
	int maxCards,				/* max. number of entries in *cards */
	int _PCL_OUT *cardsFound,	/* number of cards found in system */
	int _PCL_OUT *cards			/* table of id's of cards found like PCL_ID_xxx */
);

#define PCL_ID_PCIE_L_V2	'e'		/* PCL PCIe L (v2) */
#define PCL_ID_PCIE_L		'l'		/* PCL PCIe L */
#define PCL_ID_PCI_L_V2		'C'		/* PCL PCI L (v2) */
#define PCL_ID_PCI_L		'L'		/* PCL PCI L */

#define PCL_ID_PCI_V		'V'		/* PCL PCI V */

#define PCL_ID_PCIE_LV_V2	'b'		/* PCL PCIe LV (v2) */
#define PCL_ID_PCIE_LV		'v'		/* PCL PCIe LV */
#define PCL_ID_PCI_LV		'\1'	/* PCL PCI LV */
#define PCL_ID_PCI_LV_D		'\0'	/* PCL PCI LV or D */

#define PCL_ID_PCIE_D		'D'		/* PCL PCIe D */
#define PCL_ID_PCI_D		'\2'	/* PCL PCI D */
#define PCL_ID_PCI_D_V2		'd'		/* PCL PCI D (v2) */

#define PCL_ID_PCIE_HD		'H'		/* PCL PCIe HD */
#define PCL_ID_PCI_HD		'h'		/* PCL PCI HD */

#define PCL_ID_PCIE_3G_V2	'g'		/* PCL PCIe 3G (v2) */
#define PCL_ID_PCIE_3G		'x'		/* PCL PCIe 3G */

#define PCL_ID_PCIE_TS		't'		/* PCL PCIe TS */
#define PCL_ID_PCIE_TS_V2	'f'		/* PCL PCIe TS (v2) */
#define PCL_ID_PCI_TS		'T'		/* PCL PCI TS */

/***************************************************************************/

#ifdef PCL_OPT_VBI				/* read vbi lines */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetVbiLines(	/* @95, setup vbi reader */
	int hdl,					/* in:	handle from pclOpen() */
	int line1,					/* in:	first line */
	int lines					/* in:	number of lines */
);

/*	Note: To switch off VBI reading, set first line to 0 */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetVbiData(	/* @96, get vbi data */
	int hdl,					/* in:	handle from pclOpen() */
	int lines,					/* in:	number of lines to read */
	int pixels,					/* in:	number of pixels per line */
	BYTE _PCL_OUT *vector[]		/* out:	pointer to array of pointers to pixel data */
);

/*-------------------------------------------------------------------------*/

#endif /* PCL_OPT_VBI */

/***************************************************************************/

typedef struct {
	BYTE did;		/* did */
	BYTE sdid;		/* sdid */
	BYTE dc;		/* data count */
	BYTE error;		/* error flags, like PCL_ANC_ERROR_xxx */
} PCL_ANC_INFO;

/*-------------------------------------------------------------------------*/

#ifdef PCL_OPT_ANC				/* read user defined ancillary data */

/*-------------------------------------------------------------------------*/

/*		PCL_ANC_ERROR_xxx		error flags */
#define PCL_ANC_ERROR_DID		0x01	/* Bit 0	DID  parity error */
#define PCL_ANC_ERROR_SDID		0x02	/* Bit 1	SDID  parity error */
#define PCL_ANC_ERROR_DC		0x04	/* Bit 2	DC  parity error */
#define PCL_ANC_ERROR_DATA		0x38	/* Bit 5:3	Data  parity error count */
#define PCL_ANC_ERROR_DATA_MAX	0x40	/* Bit 6	Data  parity error count > 7 */
#define PCL_ANC_ERROR_CHK		0x80	/* Bit 7	checkword error */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetAncDID(		/* @97, $4D, setup anc reader */
	int hdl,					/* in:	handle from pclOpen() */
	int did,					/* in:	did */
	int sdid					/* in:	sdid */
);

/*	Note: To switch off ANC reading, set both did and sdid to 0 */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetAncData(	/* @98, get anc data */
	int hdl,					/* in:	handle from pclOpen() */
	int size,					/* in:	size of anc data buffer (vector) */
	BYTE _PCL_OUT *did,			/* out:	pointer to did */
	BYTE _PCL_OUT *sdid,		/* out:	pointer to sdid */
	BYTE _PCL_OUT *dc,			/* out:	pointer to data count */
	BYTE _PCL_OUT *vector,		/* out:	pointer to anc data */
	BYTE _PCL_OUT *chk,			/* out:	pointer to checksum */
	BYTE _PCL_OUT *error		/* out:	pointer to error flags, like PCL_ANC_ERROR_xxx */
);

/* Note:  Set several 'out' parameters to NULL if you don't need them */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetAncDataTab(		/* @107, get anc data, extended version */
	int hdl,						/* in:	handle from pclOpen() */
	int size,						/* in:	size of anc data buffer (vector) */
	int count,						/* in:	number of packets in buffer */
	int _PCL_OUT *first,			/* out:	pointer to number of first packet */
	int _PCL_OUT *packets,			/* out:	pointer to packet counter */
	BYTE _PCL_OUT *vector[],		/* out:	pointer to array of pointers to anc data */
	PCL_ANC_INFO _PCL_OUT *info		/* out:	pointer to array of info struct */
);

/* Note:  Set several 'out' parameters to NULL if you don't need them */

/*-------------------------------------------------------------------------*/

#endif /* PCL_OPT_ANC */

/***************************************************************************\
*																			*
* Program options															*
*																			*
\***************************************************************************/

/*	This options enable special features of AVPCL.  To use any of these,
	a special PCL PCI(e) firmware is needed. */

/***************************************************************************/

#ifdef PCL_OPT_GPI_IN			/* gpi inputs */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetGpi(	/* @99, $4F, get status of gpi in/out */
	int hdl,				/* in:	handle from pclOpen() */
	WORD _PCL_OUT *gpii,	/* out:	status of gpi inputs like PCL_GPI_xxx */
	WORD _PCL_OUT *gpio		/* out:	status of gpi outputs like PCL_GPI_xxx */
);

/* Note:  Set several 'out' parameters to NULL if you don't need them */

/*		PCL_GPI_xxx					gpi bit masks */
#define PCL_GPI_1			0x0001	/* bit mask of gpi 1 */
#define PCL_GPI_2			0x0002	/* bit mask of gpi 2 */

/*-------------------------------------------------------------------------*/

#endif /* PCL_OPT_GPI_IN */

/***************************************************************************/

#ifdef PCL_OPT_GPI_IN_EVNT		/* gpi input events */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetGpiEvent(	/* @108, $53, set event handling of gpi inputs */
	int hdl,					/* in:	handle from pclOpen() */
	int evnt,					/* in:	gpi event like PCL_GPI_EVNT_xxx */
	int reg						/* in:	registers number like PCL_REG_xxx */
);

/*		PCL_GPI_EVNT_xxx				gpi event flags */
#define PCL_GPI_EVNT_DISABLE_1		0x00 /* disable all gpi 1 events */
#define PCL_GPI_EVNT_ENABLE_HL_1	0x01 /* enable gpi 1 high to low event */
#define PCL_GPI_EVNT_ENABLE_LH_1	0x02 /* enable gpi 1 low to high event */
#define PCL_GPI_EVNT_DISABLE_2		0x10 /* disable all gpi 2 events */
#define PCL_GPI_EVNT_ENABLE_HL_2	0x11 /* enable gpi 2 high to low event */
#define PCL_GPI_EVNT_ENABLE_LH_2	0x12 /* enable gpi 2 low to high event */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetGpiEvent(	/* @109, $54, get event of gpi input */
	int hdl,					/* in:	handle from pclOpen() */
	int gpi,					/* in:	gpi input like PCL_GPI_xxx */
	PCL_TC _PCL_OUT *time,		/* out:	timecode from register, bcd */
	int _PCL_OUT *bits,			/* out:	misc. bits of register,
										like PCL_BITS_xxx_xxx */
	PCL_TC _PCL_OUT *user,		/* out:	user-bits from register, bcd */
	int _PCL_OUT *reg			/* out:	registers number like PCL_REG_xxx */
);

/* Note:  Set several 'out' parameters to NULL if you don't need them */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetGpiEvents(	/* @110, $55, get events of gpi inputs */
	int hdl,					/* in:	handle from pclOpen() */
	WORD _PCL_OUT *gpii,		/* out:	status of gpi inputs like PCL_GPI_xxx */
	WORD _PCL_OUT *change		/* out:	gpi change status like PCL_GPI_xxx */
);

/* Note:  Set several 'out' parameters to NULL if you don't need them */

/*-------------------------------------------------------------------------*/

#endif /* PCL_OPT_GPI_IN_EVNT */

/***************************************************************************/

#ifdef PCL_OPT_GPI_OUT			/* event driven gpi outputs */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetGpiEventTime(	/* @81, $28/$2C, set gpi event time */
	int hdl,						/* in:	handle from pclOpen() */
	int gpi,						/* in:	number of gpi output (1 or 2) */
	int reg,						/* in:	registers number like PCL_REG_xxx */
	int mode,						/* in:	registers number like PCL_GPIM_xxx */
	PCL_TC time						/* in:	timecode of event, bcd */
);

/*		PCL_GPIM_xxx					   gpi pulse mode */

														/* CMDD0: Mode */
														/* - Bits 0..2: Signal Output */
#define PCL_GPIM_OUT_STATIC_HIGH			(0 << 0)	/*	static High */
#define PCL_GPIM_OUT_STATIC_LOW				(1 << 0)	/*	Low */
#define PCL_GPIM_OUT_LOW_PULSE_RECURRENT	(2 << 0)	/*	Low pulse, recurrent */
#define PCL_GPIM_OUT_HIGH_PULSE_RECURRENT	(3 << 0)	/*	High pulse, recurrent */
#define PCL_GPIM_OUT_LOW_PULSE_ONE_SHOT		(4 << 0)	/*	Low pulse, one shot */
#define PCL_GPIM_OUT_HIGH_PULSE_ONE_SHOT	(5 << 0)	/*	High pulse, one shot */
#define PCL_GPIM_OUT_FALLING_EDGE_ONE_SHOT	(6 << 0)	/*	risiung edge, one shot */
#define PCL_GPIM_OUT_RISING_EDGE_ONE_SHOT	(7 << 0)	/*	falling edge, one shot */
														/* - Bit 3: Time Comparator */
#define PCL_GPIM_WITHOUT_FRAMES				(0 << 3)	/*	without Frames */
#define PCL_GPIM_WITH_FRAMES				(1 << 3)	/*	with Frames */
														/* - Bits 4..5: Time Comparator */
#define PCL_GPIM_COMPARE_EQ_FORWARD			(0 << 4)	/*	"=" (LTC forward only) */
#define PCL_GPIM_COMPARE_GE_FORWARD			(1 << 4)	/*	">=" (LTC forward only) */
#define PCL_GPIM_COMPARE_EQ					(2 << 4)	/*	"=" */
#define PCL_GPIM_COMPARE_GE					(3 << 4)	/*	">=" */
														/* - Bits 6..7: Pulse Width */
#define PCL_GPIM_PULSE_WIDTH_10_MS			(0 << 6)	/*	10 ms approx. */
#define PCL_GPIM_PULSE_WIDTH_40_MS			(1 << 6)	/*	40 ms approx. */
#define PCL_GPIM_PULSE_WIDTH_500_MS			(2 << 6)	/*	500 ms approx. */
#define PCL_GPIM_PULSE_WIDTH_1000_MS		(3 << 6)	/*	1000 ms approx. */
														/* CMDD1: Initialisation */
#define PCL_GPIM_INIT						(0 << 8)	/* - = 0x00: GPI_1 initialisation,
															i.e. the output returns
															to its initial state and the
															next event triggers the chosen
															function. */
#define PCL_GPIM_INIT_AND_TRANSFER_TIME		(1 << 8)	/* - <> 0x00: Initialisation plus
															the time transferred by commands
															0x28 and 0x29 is
															set to the event register. */

/* Example of a timer mode definition */
#define PCL_GPIM_PULSE_NEG					(	PCL_GPIM_OUT_LOW_PULSE_RECURRENT \
											|	PCL_GPIM_WITH_FRAMES \
											|	PCL_GPIM_COMPARE_EQ_FORWARD \
											|	PCL_GPIM_PULSE_WIDTH_40_MS \
											|	PCL_GPIM_INIT_AND_TRANSFER_TIME \
											)	/* set comparator to negative pulse */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetGpiEventStatus(	/* @82, $28/$2C, get gpi event status */
	int hdl,						/* in:	handle from pclOpen() */
	int gpi,						/* in:	number of gpi output (1 or 2) */
	int _PCL_OUT *status,			/* out:	status like PCL_GPIS_xxx */
	int _PCL_OUT *count				/* out:	pulse counter */
);

/*		PCL_GPIS_xxx					   gpi pulse mode */
#define PCL_GPIS_ACTIVE			0x01	/* gpi event active */

/*-------------------------------------------------------------------------*/

#endif /* PCL_OPT_GPI_OUT */		/* event driven gpi outputs */

/***************************************************************************/

#ifdef PCL_OPT_MTD	/* read mtd times */

int _PCL_FDEF pclMtdGetStr(	/* @101,	get mtd time as string */
	int hdl,				/* in:	handle from pclOpen() */
	int mtd,				/* in:	mtd register like PCL_MTDREG_xxx */
	char _PCL_OUT *str,		/* out:	asciiz string, max. 10 chars */
	BYTE _PCL_OUT *state	/* out:	state of time like PCL_MTDSTAT_xxx */
);

/* Note:  Set several 'out' parameters to NULL if you don't need them */

#define PCL_MTD_STR			11	/* size of *str parameter */

/*		PCL_MTDREG_xxx				mtd time */
#define PCL_MTDREG_A		0x0A /* mtd time a (display mode a) */
#define PCL_MTDREG_B		0x0B /* mtd time b (display mode b) */
#define PCL_MTDREG_C		0x0C /* mtd time c (display mode c) */
#define PCL_MTDREG_D		0x0D /* mtd time d (display mode d) */
#define PCL_MTDREG_E		0x0E /* mtd time e (display mode e) */
#define PCL_MTDREG_F		0x0F /* mtd time f (display mode f, maz tc) */
#define PCL_MTDREG_H		0x02 /* mtd time h (display mode 2, time) */
#define PCL_MTDREG_2		PCL_MTDREG_H
#define PCL_MTDREG_I		0x03 /* mtd time i (display mode 3, date) */
#define PCL_MTDREG_3		PCL_MTDREG_I
#define PCL_MTDREG_T		0x06 /* mtd time   (display mode 6) */
#define PCL_MTDREG_6		PCL_MTDREG_T

/*		PCL_MTDSTAT_xxx				mtd status */
#define PCL_MTDSTAT_BLINK	0x01 /* mtd time is blinking */
#define PCL_MTDSTAT_OFF		0x02 /* mtd time is off */
#define PCL_MTDSTAT_ERRLTC	0x04 /* no mtd time available */

#endif /* PCL_OPT_MTD */

/***************************************************************************/

#ifdef PCL_OPT_REALTIME

typedef struct {
	struct {			/* time from pcl card */
		WORD valid;		/* validation flags like PCL_RTVAL_xxx */
		WORD micro;		/* microseconds */
		WORD milli;		/* milliseconds */
		WORD sec;		/* seconds */
		WORD min;		/* minutes */
		WORD hour;		/* hours (0-23) */
		WORD day;		/* day of month (1-31) */
		WORD month;		/* month (1-12) */
		WORD year;		/* year (1990-2089) */
		WORD tzone;		/* time zone like PCL_RTZONE_xxx */
		WORD frame;		/* frame rate like PCL_FRAME_xxx */
		WORD _1;		/* reserved */
	} ltc;
	struct {			/* time from system clock (gmt) */
		WORD milli;		/* milliseconds */
		WORD sec;		/* seconds */
		WORD min;		/* minutes */
		WORD hour;		/* hours (0-23) */
		WORD day;		/* day of month (1-31) */
		WORD month;		/* month (0-11) */
		WORD year;		/* year (minus 1900) */
		WORD _0, _1, _2;/* reserved */
	} system;
} PCL_REALTIME;

/*		PCL_RTVAL_xxx				realtime validation */
#define PCL_RTVAL_INVALID	0x00 /* nothing valid */
#define PCL_RTVAL_TIME		0x01 /* time is valid (sync bit in user) */
#define PCL_RTVAL_DATE		0x02 /* date is valid (plausible and no wrap) */
#define PCL_RTVAL_VALID		0x03 /* everything is valid */

/*		PCL_RTZONE_xxx				time zone */
#define PCL_RTZONE_UTC		0x00 /* utc (from tz bits in user) */
#define PCL_RTZONE_MEZ		0x01 /* mez (from tz bits in user) */
#define PCL_RTZONE_MESZ		0x02 /* mesz (from tz bits in user) */
#define PCL_RTZONE_UNKNOWN	0x03 /* timezone unknown */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetRealtime(	/* @102,	$15, get realtime */
	int hdl,					/* in:	handle from pclOpen() */
	PCL_REALTIME _PCL_OUT *tm	/* out:	realtime from pcl */
);

/*
	return values:

	PCL_NO_NEWD:		PCL PCI(e) didn't read a ltc within the last 40ms.
	PCL_ERR_TIMEOUT:	PCL PCI(e) didn't respond to command.
*/

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetRealtimeMode(	/* @103, $1F, set realtime mode */
	int hdl,						/* in: handle from pclOpen() */
	int mode						/* in: mode for realtime reader,
											like PCL_RTMODE_xxx */
);

/*		PCL_RTMODE_xxx					   realtime reader mode */
#define PCL_RTMODE_DATE_NONE		0x00	/* no date, no status */
#define PCL_RTMODE_MTD				0x01	/* ltc (mtd) */
#define PCL_RTMODE_DATE				0x02	/* User = XX DD MM YY ("date") */
#define PCL_RTMODE_DATE_STATUS		0x03	/* User = SS DD MM YY ("status") */
#define PCL_RTMODE_DATE_BBC			0x04	/* User = EBU Tech. I29-1995 (BBC) */
#define PCL_RTMODE_DATE_DDMMYYYY	0x05	/* User = DD MM YYYY */
#define PCL_RTMODE_DATE_YYMMDDXX	0x06	/* User = YY MM DD XX */
#define PCL_RTMODE_DATE_XXYYMMDD	0x07	/* User = XX YY MM DD */
#define PCL_RTMODE_DATE_XYYMMDDX	0x08	/* User = X YY MM DD X */
#define PCL_RTMODE_DATE_DDMMYYXX	0x09	/* User = DD MM YY XX */

#endif /* PCL_OPT_REALTIME */

/***************************************************************************\
*																			*
*	Obsolete Functions														*
*																			*
\***************************************************************************/

#ifdef PCL_OBSOLETE_VER2

/*	The following declaration are only for backward campatibility to the
	ISA version 2.xx of the DLL.  With this PCI version, any of these
	function will return PCL_ERR_NOT_SUPP. */

/*-------------------------------------------------------------------------*/

/*	Typedefs
	-------- */

typedef DWORD PCL_OPRT;
typedef WORD  PCL_IPRT;
typedef DWORD PCL_SHOU;
typedef DWORD PCL_SHIN;
typedef BYTE  PCL_TDAT[7];
typedef BYTE  PCL_TSTR[8];
typedef BYTE  PCL_CDAT[8];
typedef BYTE  PCL_SIND[12];
typedef BYTE  PCL_STAT[10];
typedef BYTE  PCL_TRAN[15];

/*-------------------------------------------------------------------------*/

/*	Function Prototypes
	------------------- */

/*-------------------------------------------------------------------------*/
/* pclOpen(), @2 */

/*		PCL_BASE_xxx		pcl base address			A5	A6	A7	A8	*/
/*												SW1		.1	.2	.3	.4	*/
#define PCL_BASE_0			0x0200 /* 512				on	on	on	on	*/
#define PCL_BASE_1			0x0220 /* 544				off	on	on	on	*/
#define PCL_BASE_2			0x0240 /* 576				on	off	on	on	*/
#define PCL_BASE_FACTORY	0x0240 /* factory setting	on	off	on	on	*/
#define PCL_BASE_3			0x0260 /* 608				off	off	on	on	*/
#define PCL_BASE_4			0x0280 /* 640				on	on	off	on	*/
#define PCL_BASE_5			0x02A0 /* 672				off	on	off	on	*/
#define PCL_BASE_6			0x02C0 /* 704				on	off	off	on	*/
#define PCL_BASE_7			0x02E0 /* 736				off	off	off	on	*/
#define PCL_BASE_8			0x0300 /* 768				on	on	on	off	*/
#define PCL_BASE_9			0x0320 /* 800				off	on	on	off	*/
#define PCL_BASE_A			0x0340 /* 832				on	off	on	off	*/
#define PCL_BASE_B			0x0360 /* 864				off	off	on	off	*/
#define PCL_BASE_C			0x0380 /* 896				on	on	off	off	*/
#define PCL_BASE_D			0x03A0 /* 928				off	on	off	off	*/
#define PCL_BASE_E			0x03C0 /* 960				on	off	off	off	*/
#define PCL_BASE_F			0x03E0 /* 992				off	off	off	off	*/

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclRawCommand(
	int hdl,
	int cmd,
	const BYTE _PCL_IN *cdat
);

/*-------------------------------------------------------------------------*/
/* pclReset(), @5 */

#define PCL_IMODE_SONY9S	0x01
#define PCL_IMODE_JVC9S		0x02
#define PCL_IMODE_JVC45		0x03
#define PCL_IMODE_PANA34	0x04
#define PCL_IMODE_PANA25	0x05
#define PCL_IMODE_SONY9		0x06
#define PCL_IMODE_JVC9		0x07
#define PCL_IMODE_SONY33	0x08
#define PCL_IMODE_JVC25		0x09
#define PCL_IMODE_PANAIA	0x0A
#define PCL_IMODE_PANATC	0x0B
#define PCL_IMODE_SOSIM		0x0C

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclDataReq(
	int hdl,
	int reg
);

/*-------------------------------------------------------------------------*/
/* pclIntMask(), @7 */

#define PCL_IMASK_CTL		0x02
#define PCL_IMASK_ETIME		0x10
#define PCL_IMASK_EUSER		0x20

/*-------------------------------------------------------------------------*/
/* pclReaderEnable(), @10 */

#define PCL_RMASK_CTL		0x01

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclCtlRate(
	int hdl,
	int ctlrate
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclCtlCount(
	int hdl,
	int frps
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclCtlMode(
	int hdl,
	int cmode
);

#define PCL_CMODE_UNSIGNED	0x00
#define PCL_CMODE_SIGNED	0x01

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclCtlReset(
	int hdl
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclCtlPreset(
	int hdl,
	PCL_TC time
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclCtlDir(
	int hdl,
	int cdir
);

#define PCL_CDIR_REV		0x00 /* h = reverse, l = forward */
#define PCL_CDIR_FWD		0x01 /* l = reverse, h = forward */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclEventTime(
	int hdl,
	PCL_TC time
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclEventUser(
	int hdl,
	PCL_USER user
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclTranBank(
	int hdl,
	int bank,
	const BYTE _PCL_IN *tstr
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSioCmd(
	int hdl,
	int scmd
);

#define PCL_SCMD_RDIS		0x00
#define PCL_SCMD_RENA		0x01
#define PCL_SCMD_NONE		0x00
#define PCL_SCMD_ODD		0x20
#define PCL_SCMD_EVEN		0x60
#define PCL_SCMD_MARK		0xA0
#define PCL_SCMD_SPACE		0xE0

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSioCtl(
	int hdl,
	int sctl
);

#define PCL_SCTL_38400		0x00
#define PCL_SCTL_50			0x01
#define PCL_SCTL_75			0x02
#define PCL_SCTL_110		0x03
#define PCL_SCTL_135		0x04
#define PCL_SCTL_150		0x05
#define PCL_SCTL_300		0x06
#define PCL_SCTL_600		0x07
#define PCL_SCTL_1200		0x08
#define PCL_SCTL_1800		0x09
#define PCL_SCTL_2400		0x0A
#define PCL_SCTL_3600		0x0B
#define PCL_SCTL_4800		0x0C
#define PCL_SCTL_7200		0x0D
#define PCL_SCTL_9600		0x0E
#define PCL_SCTL_19200		0x0F
#define PCL_SCTL_RS232		0x00
#define PCL_SCTL_RS422		0x10
#define PCL_SCTL_B8			0x00
#define PCL_SCTL_B7			0x20
#define PCL_SCTL_B6			0x40
#define PCL_SCTL_B5			0x60
#define PCL_SCTL_S1			0x00
#define PCL_SCTL_S2			0x80

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclRecMode(
	int hdl,
	int rmode
);

#define PCL_RMODE_NORMAL	0x00
#define PCL_RMODE_SONY		0x01
#define PCL_RMODE_PANA		0x00

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclTranStr(
	int hdl,
	int num
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclTranData(
	int hdl,
	int num,
	const BYTE _PCL_IN *tdat
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclModeOut(
	int hdl,
	int omode,
	WORD freq
);

#define PCL_OMODE_NORMAL	0x00
#define PCL_OMODE_FREQ		0x01

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclOutPort(
	int hdl,
	PCL_OPRT oprt
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclPortHigh(
	int hdl,
	int oprt
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclPortLow(
	int hdl,
	int oprt
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclPosPulse(
	int hdl,
	PCL_OPRT oprt,
	int pmode,
	int pwidth
);

#define PCL_PMODE_INDIV		0x00
#define PCL_PMODE_SAME		0x01

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclNegPulse(
	int hdl,
	PCL_OPRT oprt,
	int pmode,
	int pwidth
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclPosPulseOut(
	int hdl,
	int oprt,
	int pwidth
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclNegPulseOut(
	int hdl,
	int oprt,
	int pwidth
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclPulseWidth(
	int hdl,
	PCL_OPRT oprt,
	int pwidth
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclAnalogOut(
	int hdl,
	int value
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclShiftData(
	int hdl,
	PCL_SHOU shou
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclShiftWidth(
	int hdl,
	int swid
);

#define PCL_SWID_8			0x00
#define PCL_SWID_16			0x01
#define PCL_SWID_24			0x02
#define PCL_SWID_32			0x03

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclShiftMode(
	int hdl,
	int smode
);

#define PCL_SMODE_DIS		0x80
#define PCL_SMODE_ONE		0x00
#define PCL_SMODE_CONT		0x01

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclRecCtl(
	int hdl,
	int rctl
);

#define PCL_RCTL_STOP		0x01
#define PCL_RCTL_PLAY		0x02
#define PCL_RCTL_REC		0x03
#define PCL_RCTL_STILL		0x04
#define PCL_RCTL_FFWD		0x05
#define PCL_RCTL_ADV		0x06
#define PCL_RCTL_FREV		0x07
#define PCL_RCTL_BACK		0x08
#define PCL_RCTL_TRIMF		0x10
#define PCL_RCTL_TRIMR		0x11
#define PCL_RCTL_EJECT		0x21
#define PCL_RCTL_EDON		0x24
#define PCL_RCTL_EDOFF		0x25

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclRecCtlData(
	int hdl,
	int rctl,
	int data
);

#define PCL_RCTD_VFWD		0x12
#define PCL_RCTD_VREV		0x13
#define PCL_RCTD_JFWD		0x14
#define PCL_RCTD_JREV		0x15
#define PCL_RCTD_SFWD		0x1F
#define PCL_RCTD_SREV		0x20
#define PCL_RCTD_EDPR		0x26

#define PCL_EDPR_AUDIO1		0x01
#define PCL_EDPR_AUDIO2		0x02
#define PCL_EDPR_AUDIO3		0x04
#define PCL_EDPR_TIMECODE	0x04
#define PCL_EDPR_AUDIO4		0x08
#define PCL_EDPR_VIDEO		0x10
#define PCL_EDPR_ASSEMBLE	0x20
#define PCL_EDPR_INSERT		0x40

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclRecCtlTc(
	int hdl,
	int rctl,
	PCL_TC time
);

#define PCL_RCTC_CUEUP		0x16
#define PCL_RCTC_GOTO		0x16

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclRecStat(
	int hdl,
	BYTE _PCL_OUT *stat,
	int _PCL_OUT *ident
);

/*-------------------------------------------------------------------------*/

#define PCL_REG_CTL			0x01
#define PCL_REG_MIXED2		0x04
#define PCL_REG_CTL2		0x05
#define PCL_REG_LTC2		0x06
#define PCL_REG_VITC2		0x07
#define PCL_REG_EVENT		0x08
#define PCL_REG_EVENT2		0x09
#define PCL_REG_TRAN		0x0A
#define PCL_REG_STATUS		0x0B
#define PCL_REG_IO			0x0C
#define PCL_REG_SERIN		0x0D
#define PCL_REG_SHIFT		0x0E
#define PCL_REG_CMD			0x0F
#define PCL_REG_BANK		0x10
#define PCL_NEWD_CTL		0x20
#define PCL_BITS_CTL_SIGN	0x40
#define PCL_BITS_CTL_DIR	0x80

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetEvent(
	int hdl,
	PCL_TC _PCL_OUT *dist1,
	PCL_TC _PCL_OUT *dist2,
	PCL_TC _PCL_OUT *dist3,
	int _PCL_OUT *evtc,
	int _PCL_OUT *evub,
	int _PCL_OUT *eflg
);

#define PCL_EVTC_FRU		0x01
#define PCL_EVTC_FRT		0x02
#define PCL_EVTC_SEU		0x04
#define PCL_EVTC_SET		0x08
#define PCL_EVTC_MIN		0x10
#define PCL_EVTC_HOU		0x20
#define PCL_EVTC_SIG12		0x40
#define PCL_EVTC_SIG		0x80

#define PCL_EVUB_UB1		0x01
#define PCL_EVUB_UB2		0x02
#define PCL_EVUB_UB3		0x04
#define PCL_EVUB_UB4		0x08
#define PCL_EVUB_UB5		0x10
#define PCL_EVUB_UB6		0x20
#define PCL_EVUB_UB7		0x40
#define PCL_EVUB_UB8		0x80

#define PCL_EFLG_UMASK		0x0F
#define PCL_EFLG_TMASK		0xF0

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetTransfer(
	int hdl,
	BYTE _PCL_OUT *tran
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetIo(
	int hdl,
	PCL_OPRT _PCL_OUT *oprt,
	int _PCL_OUT *ana,
	PCL_IPRT _PCL_OUT *iprt
);

#define PCL_OPRT_1			0x000001l
#define PCL_OPRT_2			0x000002l
#define PCL_OPRT_3			0x000004l
#define PCL_OPRT_4			0x000008l
#define PCL_OPRT_5			0x000010l
#define PCL_OPRT_6			0x000020l
#define PCL_OPRT_7			0x000040l
#define PCL_OPRT_8			0x000080l
#define PCL_OPRT_9			0x000100l
#define PCL_OPRT_10			0x000200l
#define PCL_OPRT_11			0x000400l
#define PCL_OPRT_12			0x000800l
#define PCL_OPRT_13			0x001000l
#define PCL_OPRT_14			0x002000l
#define PCL_OPRT_15			0x004000l
#define PCL_OPRT_16			0x008000l
#define PCL_OPRT_17			0x010000l
#define PCL_OPRT_18			0x020000l
#define PCL_OPRT_19			0x040000l
#define PCL_OPRT_20			0x080000l
#define PCL_OPRT_21			0x100000l
#define PCL_OPRT_22			0x200000l
#define PCL_OPRT_23			0x400000l
#define PCL_OPRT_24			0x800000l

#define PCL_IPRT_1			0x0001
#define PCL_IPRT_2			0x0002
#define PCL_IPRT_3			0x0004
#define PCL_IPRT_4			0x0008
#define PCL_IPRT_5			0x0010
#define PCL_IPRT_6			0x0020
#define PCL_IPRT_7			0x0040
#define PCL_IPRT_8			0x0080
#define PCL_IPRT_9			0x0100
#define PCL_IPRT_10			0x0200
#define PCL_IPRT_11			0x0400
#define PCL_IPRT_12			0x0800
#define PCL_IPRT_13			0x1000
#define PCL_IPRT_14			0x2000
#define PCL_IPRT_15			0x4000

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetSio(
	int hdl,
	int _PCL_OUT *ssta,
	int _PCL_OUT *scmd,
	int _PCL_OUT *sctl
);

#define PCL_SSTA_PARITY		0x01
#define PCL_SSTA_FRAMING	0x02
#define PCL_SSTA_OVERRUN	0x04
#define PCL_SSTA_RFULL		0x08
#define PCL_SSTA_TEMPTY		0x10
#define PCL_SSTA_RS232		0x00
#define PCL_SSTA_RS422		0x20
#define PCL_SSTA_DSR		0x40
#define PCL_SSTA_IRQ		0x80

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclResetTrans(
	int hdl
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetSerIn(
	int hdl,
	BYTE _PCL_OUT *sind,
	int _PCL_OUT *srcv,
	int _PCL_OUT *count
);

#define PCL_SRCV_PARITY		0x01
#define PCL_SRCV_FRAMING	0x02
#define PCL_SRCV_OVERRUN	0x04
#define PCL_SRCV_IBUFOV		0x08
#define PCL_SRCV_MBUFOV		0x10

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetShift(
	int hdl,
	PCL_SHOU _PCL_OUT *shou,
	PCL_SHIN _PCL_OUT *shin,
	int _PCL_OUT *shst
);

#define PCL_SHST_INEWD		0x01
#define PCL_SHST_OENA		0x80

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetTrans(
	int hdl,
	int _PCL_OUT *strf
);

/*-------------------------------------------------------------------------*/

DWORD _PCL_FDEF pclIoctlError(
	void
);

/*-------------------------------------------------------------------------*/

unsigned _PCL_FDEF pclGetBaseAddress(
	int index
);

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclReaderSource(
	int hdl,
	int rsrc
);

#define PCL_RSRC_1			0x01
#define PCL_RSRC_2			0x02

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetMixed(
	int hdl,
	PCL_TC time,
	PCL_USER user,
	int status,
	int newd
);

/*-------------------------------------------------------------------------*/

#ifdef PCL_OPT_GETCMD

int _PCL_FDEF pclOptGetcmd(
	int hdl,
	int _PCL_OUT *rctl,
	BYTE _PCL_OUT *full
);

#define PCL_RCTF_STATS		0xEF
#define PCL_RCTF_NONE		0xFE
#define PCL_RCTF_UNKNOWN	0xFF

#endif /* PCL_OPT_GETCMD */

/***************************************************************************/

#endif /* PCL_OBSOLETE_VER2 */

/***************************************************************************/

#ifdef __cplusplus
}
#endif

#endif /****************************************************** _AV_AVPCL_H */
