/**********************************************\
*                                              *
*  Linux Kernel Driver for PCL PCI(e) Devices  *
*                                              *
*        Part of the shared library            *
*                                              *
\**********************************************/

// $Id: avpcl.c 2123 2023-10-05 14:32:37Z J.Loh $
// Version: 5.45 (c) 1995-2023 PLURA Europe GmbH

#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#ifndef __USE_BSD
#	define __USE_BSD
#	include <sys/time.h>
#	undef __USE_BSD
#else
#	include <sys/time.h>
#endif

#include <stdio.h>
#include <assert.h>

#include "params.h"

#include "avpcl.h"
#include "avpcldrv.h"
#include "funcs.h"
#include "pci_lib_linux.h"
#include "global.h"

#include <sys/ipc.h>
#ifdef PCL_DEBUG_LOG
#	include <sys/utsname.h>
#endif

#include <sys/errno.h>	// errno
#if defined(PCL_USE_SHM_MUTEX)
#	include <sys/shm.h>
#endif

#ifndef PATHTODEVICEFILES
#	error Path not defined. Please do so in the Makefile.
#endif

#ifdef PCL_OPT_FLASH
TPclBridges gPclBridges;
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

BOOL gbAllowPclOpen = FALSE; // allow pclOpen when flash flag is set

//----------------------------------------------------------------------

#ifdef PCL_USE_SHM_MUTEX
	int gAvPclShmid = -1;			// Shared memory handle
	TMutex *gAvPclMutex = NULL;	// Pointer to global access mutex
	char *gpszError = "";
#endif

//----------------------------------------------------------------------
#ifdef PCL_USE_SHM_MUTEX
int drvCount(BOOL bInc)
{
	int cnt = -1;
	int fd = open((bInc ? "/proc/"PROC_FILENAME_INC : "/proc/"PROC_FILENAME_DEC), O_RDONLY);
	if (fd != -1) {
		char buf[20];
		memset(buf, 0, sizeof(buf));
		ssize_t size = read(fd, buf, sizeof(buf));
		if (size > 0) {
			buf[sizeof(buf)-1] = 0;
			cnt = atoi(buf);
		} else {
			cnt = -1;
		} // end else
		close(fd);
	} // end if
	return cnt;
} // end DrvCount()
#endif // PCL_USE_SHM_MUTEX

//----------------------------------------------------------------------

#ifdef PCL_USE_SHM_MUTEX
int incDrvCount() { return drvCount(TRUE ); }
int decDrvCount() { return drvCount(FALSE); }
#endif

//----------------------------------------------------------------------

#ifdef PCL_USE_SHM_MUTEX
int lockTimeout(pthread_mutex_t *pMutex, unsigned int toutMs)
{
	struct timeval tstart, tstop, tdiff;
	int result = 0;
	gettimeofday(&tstart, NULL);	

	while ((result = pthread_mutex_trylock(pMutex))) {
		if (result == EOWNERDEAD || result == -EOWNERDEAD) {
			APPEND_DEBUG_LOG(-1, "Mutex owner has died. Recovering.");
			result = pthread_mutex_consistent(pMutex);
			break;
		} // end if
		gettimeofday(&tstop, NULL);
		timersub(&tstop, &tstart, &tdiff);
		unsigned int msec = tdiff.tv_sec * 1000 + tdiff.tv_usec;
		if (msec >= toutMs) break;
	} // end while
	return result;
} // end lockTimeout()
#endif // PCL_USE_SHM_MUTEX

//----------------------------------------------------------------------
// some helper functions
#ifdef PCL_USE_SHM_MUTEX
BOOL allocSharedMem(int *shmid, key_t key) // alloc shared memory
{
	*shmid = shmget(key, sizeof(TMutex), 0666 | IPC_CREAT);
	if (*shmid == -1) {
		decDrvCount();
		return FALSE;
	} // end if
	return TRUE;
} // end allocSharedMem()

BOOL initMutex(TMutex *pMutex) // initialize mutex
{
	BOOL ok = TRUE;
	if (!pMutex -> bMutexValid) {
		pthread_mutexattr_t attr;
		ok = ok && !pthread_mutexattr_init(&attr);
		ok = ok && !pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
		ok = ok && !pthread_mutexattr_setpshared(&attr, PTHREAD_PROCESS_SHARED);
		ok = ok && !pthread_mutexattr_setrobust(&attr, PTHREAD_MUTEX_ROBUST);
//		pMutex -> count = 0;
		
		ok = ok && !pthread_mutex_init(&pMutex -> mutex, &attr);
		
		if (ok) pMutex -> bMutexValid = TRUE;
		else {
			decDrvCount();
		} // end else
	} // end if
	return ok;
} // end initMutex()
#endif

//----------------------------------------------------------------------

__attribute__((constructor)) void loadLib()
{
#ifndef NDEBUG
//	printf("*** load lib.\n");
#endif
#ifdef PCL_USE_SHM_MUTEX
	int count = incDrvCount();
#endif

	key_t key = ftok(PATHTODEVICEFILES"avpcl0", 0xFF);
	if (key == -1) {
#ifdef PCL_USE_SHM_MUTEX
		decDrvCount();
#endif
#ifndef NDEBUG
		perror("ftok");
#endif
		gpszError = "Can't access avpcl driver. Please check if it's running (./avpcl status).";
		return;
	} // end if
	
	// Init bridge struct
	{
		int n;
		for (n = 0; n < NELT(gPclBridges.v); n++) {
			gPclBridges.v[n].hPCI = -1;
		}
	} // end for
	
#ifdef PCL_USE_SHM_MUTEX
	int shmid = -1;

	// Allocate shared memory
	if (!allocSharedMem(&shmid, key)) {
#ifndef NDEBUG
		perror("allocSharedMem");
#endif
		gpszError = "Allocation of global shared memory not successful.";
		return;
	} // end if
	
	// attach to shared memory
	TMutex *pMutex = shmat(shmid, NULL, 0);
	if (pMutex == (void*)-1) {
		shmctl(shmid, IPC_RMID, 0);
		decDrvCount();
#ifndef NDEBUG
		perror("shmat");
#endif
		gpszError = "Could not attach to global shared memory.";
		return;
	} // end if

	// In case of concurrent start of multiple processes:
	// Wait up to 1 second for other lib if mutex is not yet ready
	if (!pMutex -> bMutexValid && (count > 1)) {
		int tout = 100;
		while (tout > 0 && !pMutex -> bMutexValid) {
			tout--;
			usleep(10000);
		} // end while
	} // end if

	// init mutex
	if (!initMutex(pMutex)) {
#ifndef NDEBUG
		perror("initMutex");
#endif
		gpszError = "Could not initialize global mutex.";
		return;
	} // end if

	gAvPclMutex = pMutex;
	gAvPclShmid = shmid;
#endif // PCL_USE_SHM_MUTEX
} // loadLib()

//----------------------------------------------------------------------

void pclDestroyAllSharedMemory()
{
	// Open all available cards
	int n, res, hdl;	
	
	for (n = 0; n <= PCL_BASE_PCI_LAST; n++) {
		res = pclOpen(&hdl, PCL_BASE_PCI_OFS + n); // Attach
		if (res == PCL_OK) pclClose(hdl);		   // Detach
	} // end for
	
	// Global shared Memory will be cleaned by unloadLib()
	
} // end pclDestroyAllSharedMemory()

//----------------------------------------------------------------------

__attribute__((destructor)) void unloadLib()
{
#ifdef PCL_USE_SHM_MUTEX
#ifndef NDEBUG
//	printf("*** unload lib.\n");
#endif
	// detach from global shared memory
	if (gAvPclMutex != NULL) {
		shmdt(gAvPclMutex);
		gAvPclMutex = NULL;
	} // end if
	
	// get number of attached processes
	struct shmid_ds ds;
	if (gAvPclShmid != -1) {
		shmctl(gAvPclShmid, IPC_STAT, &ds);
		
		// delete shared memory if no process left
		if (ds.shm_nattch == 0) {
			shmctl(gAvPclShmid, IPC_RMID, 0);
		} // end if
	} // end if

	// clean up shared memory for each card if no process is attached
	int card = 0;
	for (; card < PCL_CARDS; card++) {
		if (gPciCardLocal.v[card].shmid <= 0) continue;
		shmctl(gPciCardLocal.v[card].shmid, IPC_STAT, &ds);
#ifndef NDEBUG
//		printf(" card %i: nattch:%i\n", card, (int)ds.shm_nattch);
#endif
		if (ds.shm_nattch == 0) {
			shmctl(gPciCardLocal.v[card].shmid, IPC_RMID, 0);
		} // end if
	} // end for
#endif // PCL_USE_SHM_MUTEX
} // end unloadLib()

//----------------------------------------------------------------------

int pclGetActiveProcessCount()
{
	if (gAvPclShmid < 0) {
		PCI_ERR("%s", gpszError);
		pclErrorString = PCI_ErrorString;
		return 0;
	} // end if
	
	// get number of attached processes
	struct shmid_ds ds;
	shmctl(gAvPclShmid, IPC_STAT, &ds);
	
	return ds.shm_nattch;
} // end pclGetActiveProcesses()

//----------------------------------------------------------------------

BOOL pclIsFlashActive() {
	if (gAvPclMutex == NULL) return FALSE;
	return (pclGetActiveProcessCount() > 1) && gAvPclMutex -> flags.flash;
} // end pclIsFlashActive()

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
BOOL pclSetFlashActive(BOOL bActive)
{
	if (gAvPclMutex == NULL) {
		PCI_ERR("%s", gpszError);
		pclErrorString = PCI_ErrorString;
		return FALSE;
	} // end if
	gAvPclMutex -> flags.flash = bActive;
	gbAllowPclOpen = bActive;
	return TRUE;
} // end pclSetFlashActive()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_USE_SHM_MUTEX
int createSharedCardMutex(int card) {
	key_t key = ftok(PATHTODEVICEFILES"avpcl0", card + 1);
	assert(key != -1); // Does the device file exist?
	if (key == -1) {
		return 2; // failed
	} // end if
	
	int shmid = -1;

	// alloc shared memory
	shmid = shmget(key, sizeof(TMutex), 0666 | IPC_CREAT);
	if (shmid == -1) {
		return 3; // failed
	} // end if
	
	// attach to shared memory
	TMutex *pMutex = shmat(shmid, NULL, 0);
	if (pMutex == (void*)-1) {
		shmctl(shmid, IPC_RMID, 0);
		return 4; // failed
	} // end if
	
	// init mutex
	if (!pMutex -> bMutexValid) {
		pthread_mutexattr_t attr;
		BOOL ok = TRUE;
		ok = ok && !pthread_mutexattr_init(&attr);
		ok = ok && !pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
		ok = ok && !pthread_mutexattr_setpshared(&attr, PTHREAD_PROCESS_SHARED);
		ok = ok && !pthread_mutexattr_setrobust(&attr, PTHREAD_MUTEX_ROBUST);
		ok = ok && !pthread_mutex_init(&pMutex -> mutex, &attr);
		
		if (ok) pMutex -> bMutexValid = TRUE;
		else {
			return 5; // failed
		} // end else
	} // end if

	// Test Mutex
	/*
	int suc = lockTimeout(&pMutex -> mutex, 2000);
	if (suc != 0) { // lock failed
		incDrvCount();
		int libCount = decDrvCount();
		if (libCount == 1) {
			shmctl(shmid, IPC_RMID, 0);
			return 6; // failed
		} // end if
	} // end if
	pthread_mutex_unlock(&pMutex -> mutex);*/
	
	gPciCardLocal.v[card].shmid  = shmid;
	gPciCardLocal.v[card].pMutex = pMutex;
	return 0/*TRUE*/;
} // end createSharedCardMutex()
#endif // PCL_USE_SHM_MUTEX

//----------------------------------------------------------------------

#ifdef PCL_USE_SHM_MUTEX
void removeSharedCardMutex(int card) {
	shmdt(gPciCardLocal.v[card].pMutex); // Detach
	gPciCardLocal.v[card].pMutex = NULL;
} // end removeSharedCardMutex()
#endif // PCL_USE_SHM_MUTEX

//----------------------------------------------------------------------

#ifdef PCL_DEBUG_LOG

void openDebugLog(int card, int hdl) {
	char sz[MAX_PATH];
	sprintf(sz, /* sizeof(sz), */ "logavpcl-%d-%d-%d.log", getpid(), card, hdl);
	assert(pgPciHdl);
	pgPciHdl->v[hdl].file = fopen(sz, "wt");
	assert(pgPciHdl->v[hdl].file);
	APPEND_DEBUG_LOG(hdl, "  -----------------------------------------------------");
	APPEND_DEBUG_LOG(hdl, "  openDebugLog() logging opened");

	{
		struct utsname buf;
		uname(&buf);
		APPEND_DEBUG_LOG(hdl, "  OS: %s Version:%s Release:%s Node:%s"
			, buf.sysname, buf.version, buf.release, buf.nodename);
	}
	APPEND_DEBUG_LOG(hdl, "  Avpcl32 DLL Version %d.%d, " __DATE__ ", " __TIME__, PCL_VER_DLL / 0x100, PCL_VER_DLL % 0x100);
}

#endif // PCL_DEBUG_LOG

//----------------------------------------------------------------------

int pclOpen( int *hdl, unsigned base )
{
	int fileDescriptor = -1;
	const char *devicefilename;
	
	// Check for global mutex
	if (gAvPclMutex == NULL || gAvPclShmid == -1) {
		PCI_ERR("%s", gpszError);
		pclErrorString = PCI_ErrorString;
		return PCL_ERR_NOTOPEN;
	} // end if

	// Check for PCLFlash
	if (!gbAllowPclOpen && pclIsFlashActive()) {
		PCI_ERR("*Cannot open card while PCLFlash is running.*\n");
		pclErrorString = PCI_ErrorString;
		fprintf(stderr, "%s", pclErrorString);
		return PCL_ERR_NOTOPEN;
	} // end if

	if (hdl == NULL) {
		pclErrorString = "pclOpen(): Parameter 'hdl' is NULL\n";
		return(PCL_ERR_NULL);
	} // end if
	
	*hdl = -1;
	
	pclErrorString = pclPciGetErrorString();

	int card;
	if (PCL_BASE_PCI_1 <= base && base <= PCL_BASE_PCI_10) {
		card = base - PCL_BASE_PCI_OFS - 1;
	} else {
		PCI_ERR("Card %i(0x%x) not found.", base, base);
		return PCL_ERR_NOT_FOUND;
	} // end else
	
	// open card
	devicefilename = PCL_CardFilename(card);
#ifdef PCL_USE_SHM_MUTEX
	int suc = lockTimeout(&gAvPclMutex -> mutex, 5000);
#endif // PCL_USE_SHM_MUTEX
	if (suc != 0) {
		PCI_ERR("pclOpen: Mutex lock failed (%i).\n", suc);
#ifndef NDEBUG
		perror("pclOpen");
		assert(FALSE);
#endif
		return PCL_ERR_NOTOPEN;
	} // end if
	
	if (!gPciCard.v[card].open)
		fileDescriptor = open(devicefilename, O_RDWR);

	if (fileDescriptor != -1 || gPciCard.v[card].open) {
		
		assert(pgPciHdl);
		assert(PCL_HANDLES == NELT(pgPciHdl->v));
		assert(PCL_CARDS == NELT(gPciCard.v));

		// search free handle
		int i;
		for (i = 0; i < PCL_HANDLES; i++) {
			if (pgPciHdl->v[i].type == TYPE_UNUSED) {
				*hdl = i;
				assert(card >= 0 && card < NELT(gPciCard.v));

				pgPciHdl->v[i].type = TYPE_PCI;
				pgPciHdl->v[i].card = card;
				if (!gPciCard.v[card].open) {

					gPciCard.v[card].hPCI = fileDescriptor;
					
#ifdef PCL_USE_SHM_MUTEX
					int iResult1 = createSharedCardMutex(card);
					// if not successful, try again
					if (iResult1) {
						int iResult2 = createSharedCardMutex(card);
						if (iResult2) {
//							pclErrorString = "Could not create shared Mutex.";
							PCI_ERR("Could not create shared Mutex (%d, %d)", iResult1, iResult2);

							close(fileDescriptor);
							gPciCard.v[card].hPCI = -1;
							pgPciHdl->v[i].type = TYPE_UNUSED;
							removeSharedCardMutex(card);
							pthread_mutex_unlock(&gAvPclMutex -> mutex);
							return PCL_ERR_NOTOPEN;
						} // end if
					} // end if
#endif // PCL_USE_SHM_MUTEX

					gPciCard.v[card].wVendorID = PCI_DEFAULT_VENDOR_ID;
					PCI_GetDeviceId(fileDescriptor, &gPciCard.v[card].wDeviceID);

					gPciCard.v[card].wSubsystemVendorID = PCI_DEFAULT_VENDOR_ID;
					if (PCI_GetSubsystemDeviceId(fileDescriptor, &gPciCard.v[card].wSubsystemDeviceID) 
						!= PCL_OK)
					{
						PCI_ERR("Could not retrieve Subsystem device ID.");
						close(fileDescriptor);
						gPciCard.v[card].hPCI = -1;
						pgPciHdl->v[i].type = TYPE_UNUSED;
						removeSharedCardMutex(card);
						pthread_mutex_unlock(&gAvPclMutex -> mutex);
						return PCL_ERR_NOTOPEN;
					} // end if

					assert(	gPciCard.v[card].wDeviceID == AVPCL_DEVICE_ID_E  ||
							gPciCard.v[card].wDeviceID == AVPCL_DEVICE_ID_5V ||
							gPciCard.v[card].wDeviceID == AVPCL_DEVICE_ID_3V
						  );

					// Check irq16
					switch (gPciCard.v[card].wSubsystemDeviceID) {
						case AVPCL_SUBSYSTEM_DEVICE_ID_HD_3V:
						case AVPCL_SUBSYSTEM_DEVICE_ID_D_V2_3V:
						case AVPCL_SUBSYSTEM_DEVICE_ID_HD_PCIE:
						case AVPCL_SUBSYSTEM_DEVICE_ID_D_PCIE:
						case AVPCL_SUBSYSTEM_DEVICE_ID_3G_PCIE:
#ifdef PCL_OPT_GPI_IN_EVNT		/* gpi input events */
						case AVPCL_SUBSYSTEM_DEVICE_ID_LV_PCIE:
						case AVPCL_SUBSYSTEM_DEVICE_ID_L_PCIE:
#endif
							gPciCard.v[card].irq16 = TRUE;
						break;
						default:
							gPciCard.v[card].irq16 = FALSE;
						break;
					} // end switch

				} // end if
				gPciCard.v[card].open++;
				break;
			} // end if
		} // end for
		if (i >= PCL_HANDLES) {
			PCI_ERR("pclOpen(): All handles used, no more handles available.");
#ifdef PCL_USE_SHM_MUTEX
			pthread_mutex_unlock(&gAvPclMutex -> mutex);
#endif // PCL_USE_SHM_MUTEX

			return PCL_ERR_NO_HANDLE;
		} // end if
	} else {
		PCI_ERR("Could not open device file %s. Please ensure that the driver is started correctly.", devicefilename);
#ifdef PCL_USE_SHM_MUTEX
		pthread_mutex_unlock(&gAvPclMutex -> mutex);
#endif // PCL_USE_SHM_MUTEX
		return PCL_ERR_NOTOPEN;
	} // end else
	
#ifdef PCL_USE_SHM_MUTEX
	pthread_mutex_unlock(&gAvPclMutex -> mutex);
#endif // PCL_USE_SHM_MUTEX

#ifdef PCL_DEBUG_LOG
	openDebugLog(card, *hdl);
	APPEND_DEBUG_LOG(*hdl, "> pclPciOpen(card:%d)", card);

	GetTimeStampFreq(&gPciCard.v[card].perf.freq);
	GetTimeStamp(&gPciCard.v[card].perf.abs);
#endif // PCL_DEBUG_LOG

#ifndef NDEBUG
//	printf("*** pclOpen(%p, %u) -> hdl=%i.\n", hdl, base, *hdl);
#endif
	return PCL_OK;
} // end pclOpen()

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
BOOL IsBridge(int hdl) {
	return (PCL_HANDLES <= hdl) && (hdl < (PCL_HANDLES + PCL_CARDS));
} // end IsBridge()

int ConvertHandle(int hdl) {
	if (!IsBridge(hdl)) return hdl;
	return hdl - PCL_HANDLES;
} // end ConvertHdl()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

int pclClose( int hdl )
{
	int result = PCL_OK;
#ifndef NDEBUG
//	printf("*** pclClose(%i)%s.\n", hdl, IsBridge(hdl) ? " bridge" : "");
#endif

#ifdef PCL_OPT_FLASH
	if (IsBridge(hdl)) {
		hdl = ConvertHandle(hdl);
		result = close(gPclBridges.v[hdl].hPCI);
		gPclBridges.v[hdl].hPCI = -1;
		return result;
	} // end if
#endif // PCL_OPT_FLASH

	if (hdl < 0 || hdl > PCL_HANDLES) {
		pclErrorString = "pclClose(): Parameter 'hdl' is out of range.";
#ifndef NDEBUG
		printf("%s", pclErrorString);
#endif
		return(PCL_ERR_INV_HDL);
	} // end if

#ifdef PCL_USE_SHM_MUTEX
	int suc = lockTimeout(&gAvPclMutex -> mutex, 1000);
#endif // PCL_USE_SHM_MUTEX
	if (suc != 0) {
		PCI_ERR("pclClose: mutex locked (%i).\n", suc);
		pclErrorString = PCI_ErrorString;
		result = PCL_ERR_TIMEOUT;
		assert(FALSE);
	} // end if
	
	assert(pgPciHdl);
	int card = pgPciHdl->v[hdl].card;
	assert(card >= 0 && card < NELT(gPciCard.v));
	PCI_HANDLE hPCI = gPciCard.v[card].hPCI;

#ifdef PCL_DEBUG_LOG
	{
		APPEND_DEBUG_LOG(hdl, "  pclClose() logging closed");
		fclose(pgPciHdl->v[hdl].file);
	}
#endif // PCL_DEBUG_LOG

	if (pgPciHdl->v[hdl].imask != 0) {
		pclIntMask(hdl, 0);	// free own eigene interrupts
		Sleep(100);	// give own interrupts time to finish
	}

	switch (pgPciHdl->v[hdl].type) {
	case TYPE_PCI:
		pclErrorString = pclPciGetErrorString();
		if (gPciCard.v[card].open == 1) {
			result = close(hPCI);
			removeSharedCardMutex(card); // Detach
		} // end if
		pgPciHdl->v[hdl].type = TYPE_UNUSED;
		gPciCard.v[card].open--;
		break;
	default:
		pclErrorString = "pclClose(): Type of PCL card unknown\n";
#ifndef NDEBUG
		printf("%s", pclErrorString);
#endif
		result = PCL_ERR_NOTOPEN;
		break;
	} // end switch

#ifdef PCL_USE_SHM_MUTEX
	pthread_mutex_unlock(&gAvPclMutex -> mutex);
#endif // PCL_USE_SHM_MUTEX

	return(result);
} // end pclClose()

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int pclGetDeviceInfo(int hdl, TDeviceInfo _PCL_OUT *deviceInfo) {
	BOOL bBridge = IsBridge(hdl);
	CHECK_PTR(deviceInfo);
	PCI_GET_HANDLE_LOCK_BRIDGE(hdl, pclGetDeviceInfo, (), hPCI, card, bBridge);
	if (hPCI < 0) UNLOCK_RETURN_BRIDGE(hdl, pclGetDeviceInfo, PCL_ERR_INV_HDL, bBridge);

	int res;
	if (bBridge) 	res = PCI_GetDeviceInfoBridge(hPCI, deviceInfo);
	else 		 	res = PCI_GetDeviceInfo(hPCI, deviceInfo);
	
	if (res != PCL_OK)
		UNLOCK_RETURN_BRIDGE(hdl, pclGetDeviceInfo, PCL_ERR_IOCTL, bBridge);
	
	UNLOCK_RETURN_BRIDGE(hdl, pclGetDeviceInfo, PCL_OK, bBridge);
} // end pclGetDeviceInfo()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int pclGetDeviceId(int hdl, WORD _PCL_OUT *deviceId) {
	BOOL bBridge = IsBridge(hdl);
	CHECK_PTR(deviceId);
	PCI_GET_HANDLE_LOCK_BRIDGE(hdl, pclGetDeviceId, (), hPCI, card, bBridge);
	if (hPCI < 0) UNLOCK_RETURN_BRIDGE(hdl, pclGetDeviceId, PCL_ERR_INV_HDL, bBridge);

	int res;
	if (bBridge) 	res = PCI_GetDeviceIdBridge(hPCI, deviceId);
	else 		 	res = PCI_GetDeviceId(hPCI, deviceId);
	
	if (res != PCL_OK)
		UNLOCK_RETURN_BRIDGE(hdl, pclGetDeviceId, PCL_ERR_IOCTL, bBridge);
	
	UNLOCK_RETURN_BRIDGE(hdl, pclGetDeviceId, PCL_OK, bBridge);
} // end pclGetDeviceId()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int pclGetBus(int hdl, BYTE _PCL_OUT *bus) {
	CHECK_PTR(bus);
	if (IsBridge(hdl)) return PCL_ERR_INV_HDL;
	PCI_GET_HANDLE_LOCK(hdl, pclGetBus, (), hPCI, card);
	if (hPCI < 0) UNLOCK_RETURN(hdl, pclGetBus, PCL_ERR_INV_HDL);

	int res = PCI_GetBus(hPCI, bus);
	
	if (res != PCL_OK)
		UNLOCK_RETURN(hdl, pclGetBus, PCL_ERR_IOCTL);
	
	UNLOCK_RETURN(hdl, pclGetBus, PCL_OK);
} // pclGetBus()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int pclGetSubsystemDeviceId(int hdl, WORD _PCL_OUT *deviceId) {
	CHECK_PTR(deviceId);
	PCI_GET_HANDLE_LOCK(hdl, pclGetSubsystemDeviceId, (), hPCI, card);
	if (hPCI < 0) UNLOCK_RETURN(hdl, pclGetSubsystemDeviceId, PCL_ERR_INV_HDL);
	
	if (PCI_GetSubsystemDeviceId(hPCI, deviceId) != PCL_OK)
		UNLOCK_RETURN(hdl, pclGetSubsystemDeviceId, PCL_ERR_IOCTL);
	
	UNLOCK_RETURN(hdl, pclGetSubsystemDeviceId, PCL_OK);
} // end pclGetSubsystemDeviceId()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int pclConfigDump(int hdl, UINT offset, UINT size, void _PCL_OUT *result) {
	BOOL bBridge = IsBridge(hdl);
	CHECK_PTR(result);
	PCI_GET_HANDLE_LOCK_BRIDGE(hdl, pclConfigDump, (), hPCI, card, bBridge);
	if (hPCI < 0) 
		UNLOCK_RETURN_BRIDGE(hdl, pclConfigDump, PCL_ERR_INV_HDL, bBridge);
	
	int res;
	if (bBridge) 	res = PCI_ConfigDumpBridge(hPCI, offset, size, result);
	else 		 	res = PCI_ConfigDump(hPCI, offset, size, result);

	if (res != PCL_OK)
		UNLOCK_RETURN_BRIDGE(hdl, pclConfigDump, PCL_ERR_IOCTL, bBridge);

	UNLOCK_RETURN_BRIDGE(hdl, pclConfigDump, PCL_OK, bBridge);
} // end pclConfigDump()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int pclReadOffset8 (int hdl, BYTE bar, UINT offset, BYTE _PCL_OUT *result) {
	BOOL bBridge = IsBridge(hdl);
	CHECK_PTR(result);
	PCI_GET_HANDLE_LOCK_BRIDGE(hdl, pclReadOffset8, (), hPCI, card, bBridge);
	if (hPCI < 0) UNLOCK_RETURN_BRIDGE(hdl, pclReadOffset8, PCL_ERR_INV_HDL, bBridge);
	
	int res;
	if (bBridge) 	res = PCI_ReadOffset8Bridge(hPCI, bar, offset, result);
	else 		 	res = PCI_ReadOffset8(hPCI, bar, offset, result);
	if (res != PCL_OK)
		UNLOCK_RETURN_BRIDGE(hdl, pclReadOffset8, PCL_ERR_IOCTL, bBridge);
	
	UNLOCK_RETURN_BRIDGE(hdl, pclReadOffset8, PCL_OK, bBridge);
} // end pclReadOffset8()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int pclReadOffset32(int hdl, BYTE bar, UINT offset, DWORD _PCL_OUT *result) {
	BOOL bBridge = IsBridge(hdl);
	CHECK_PTR(result);
	PCI_GET_HANDLE_LOCK_BRIDGE(hdl, pclReadOffset32, (), hPCI, card, bBridge);
	if (hPCI < 0) UNLOCK_RETURN(hdl, pclReadOffset32, PCL_ERR_INV_HDL);

	int res;
	if (bBridge)	res = PCI_ReadOffset32Bridge(hPCI, bar, offset, result);
	else			res = PCI_ReadOffset32(hPCI, bar, offset, result);

	if (res != PCL_OK)
		UNLOCK_RETURN_BRIDGE(hdl, pclReadOffset32, PCL_ERR_IOCTL, bBridge);
	
	UNLOCK_RETURN_BRIDGE(hdl, pclReadOffset32, PCL_OK, bBridge);
} // end pclReadOffset32()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int pclWriteOffset8 (int hdl, BYTE bar, UINT offset, BYTE data) {
	BOOL bBridge = IsBridge(hdl);
	PCI_GET_HANDLE_LOCK_BRIDGE(hdl, pclWriteOffset8, (), hPCI, card, bBridge);
	if (hPCI < 0) UNLOCK_RETURN_BRIDGE(hdl, pclWriteOffset8, PCL_ERR_INV_HDL, bBridge);
	
	int res;
	if (bBridge)	res = PCI_WriteOffset8Bridge(hPCI, bar, offset, data);
	else 			res = PCI_WriteOffset8(hPCI, bar, offset, data);
	
	if (res != PCL_OK)
		UNLOCK_RETURN_BRIDGE(hdl, pclWriteOffset8, PCL_ERR_IOCTL, bBridge);
	
	UNLOCK_RETURN_BRIDGE(hdl, pclWriteOffset8, PCL_OK, bBridge);
} // end pclWriteOffset8()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int pclWriteOffset32(int hdl, BYTE bar, UINT offset, DWORD data) {
	BOOL bBridge = IsBridge(hdl);
	PCI_GET_HANDLE_LOCK_BRIDGE(hdl, pclWriteOffset8, (), hPCI, card, bBridge);
	if (hPCI < 0) UNLOCK_RETURN_BRIDGE(hdl, pclWriteOffset32, PCL_ERR_INV_HDL, bBridge);

	int res;
	if (bBridge)	res = PCI_WriteOffset32Bridge(hPCI, bar, offset, data);
	else 			res = PCI_WriteOffset32(hPCI, bar, offset, data);
	
	if (res != PCL_OK)
		UNLOCK_RETURN_BRIDGE(hdl, pclWriteOffset32, PCL_ERR_IOCTL, bBridge);
	
	UNLOCK_RETURN_BRIDGE(hdl, pclWriteOffset32, PCL_OK, bBridge);
} // end pclWriteOffset32()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
BYTE pclReadVersion(int hdl) {
	PCI_GET_HANDLE_LOCK(hdl, pclReadVersion, (), hPCI, card);
	if (hPCI < 0) UNLOCK_RETURN(hdl, pclReadVersion, PCL_ERR_INV_HDL);
	
	BYTE res = PCI_ReadVERSION(hPCI);
	
	UNLOCK_RETURN(hdl, pclReadVersion, res);
} // end pclReadVersion()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int pclOpenBridge(int *hdl, UINT base) {
	int fileDescriptor = -1;
	const char *devicefilename;

	// Check for global mutex
	if (gAvPclMutex == NULL || gAvPclShmid == -1) {
		PCI_ERR("%s", gpszError);
		pclErrorString = PCI_ErrorString;
		return PCL_ERR_NOTOPEN;
	} // end if
	
	if (hdl == NULL) {
		pclErrorString = "pclOpenBridge(): Parameter 'hdl' is NULL\n";
		return(PCL_ERR_NULL);
	} // end if
	
	*hdl = -1;
	
	pclErrorString = pclPciGetErrorString();

	int card;
	if (PCL_BASE_PCI_1 <= base && base <= PCL_BASE_PCI_10) {
		card = base - PCL_BASE_PCI_OFS - 1;
	} else {
		PCI_ERR("Card %i(0x%x) not found.", base, base);
		return PCL_ERR_NOT_FOUND;
	} // end else
	
	// open card
	devicefilename = PCL_CardFilename(card);
#ifdef PCL_USE_SHM_MUTEX
	int suc = lockTimeout(&gAvPclMutex -> mutex, 5000);
#endif // PCL_USE_SHM_MUTEX
	if (suc != 0) {
		PCI_ERR("pclOpenBridge: Mutex lock failed (%i).\n", suc);
		assert(FALSE);
		return PCL_ERR_NOTOPEN;
	} // end if
	
	fileDescriptor = open(devicefilename, O_RDWR);

	if (fileDescriptor != -1) {
		if (!PCI_IsBridge(fileDescriptor)) {
			PCI_ERR("No bridge found on position %i.", card);
			close(fileDescriptor);
			pthread_mutex_unlock(&gAvPclMutex -> mutex);
			return PCL_ERR_NOTOPEN;
		} // end if
		
		// Find free handle
		int n;
		for (n = 0; n < NELT(gPclBridges.v); n++) {
			if (gPclBridges.v[n].hPCI < 0) {
				gPclBridges.v[n].hPCI = fileDescriptor;
				gPclBridges.v[n].card = card;
				*hdl = n;
				break;
			} // end if
		} // end for

		if (n == NELT(gPclBridges.v)) {
			return PCL_ERR_NO_HANDLE;
		} // end if
		
/*#ifdef PCL_USE_SHM_MUTEX
		int iResult1 = createSharedCardMutex(card);
		// if not successful, try again
		if (iResult1) {
			int iResult2 = createSharedCardMutex(card);
			if (iResult2) {
				PCI_ERR("Could not create shared Mutex (%d, %d)", iResult1, iResult2);
				close(fileDescriptor);
				pthread_mutex_unlock(&gAvPclMutex -> mutex);
				return PCL_ERR_NOTOPEN;
			} // end if
		} // end if
#endif // PCL_USE_SHM_MUTEX*/

	} else {
		PCI_ERR("Could not open device file %s. Please ensure that the driver is started correctly.", devicefilename);
#ifdef PCL_USE_SHM_MUTEX
		pthread_mutex_unlock(&gAvPclMutex -> mutex);
#endif // PCL_USE_SHM_MUTEX
		return PCL_ERR_NOTOPEN;
	} // end else
	
#ifdef PCL_USE_SHM_MUTEX
	pthread_mutex_unlock(&gAvPclMutex -> mutex);
#endif // PCL_USE_SHM_MUTEX

	*hdl += PCL_HANDLES;
#ifndef NDEBUG
//	printf("*** pclOpenBridge(%i, %u).\n", *hdl, base);				
#endif // NDEBUG
	return PCL_OK;
} // end pclOpenBridge()
#endif // PCL_OPT_FLASH

//----------------------------------------------------------------------
