/**********************************************\
*                                              *
*  Linux Kernel Driver for PCL PCI(e) Devices  *
*                                              *
\**********************************************/

// $Id: avpcldrv.c 2123 2023-10-05 14:32:37Z J.Loh $
// Version: 5.45 (c) 1995-2023 PLURA Europe GmbH

#define pr_fmt(fmt) "avpcl: " fmt

#include "params.h"
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/cdev.h>
#include <linux/sort.h>
#include <linux/sched.h>
#include <linux/version.h>
#include <linux/uaccess.h>
#include <asm/signal.h>

#if 0 //def PCL_USE_SHM_MUTEX
#include <linux/proc_fs.h>	// for the proc filesystem
#include <linux/seq_file.h>	// proc: for sequence files
#endif

#include "avpcl.h"
#include "avpcldrv.h"
#include "driver.h"
#include "pci_lib_linux.h"

#ifndef NDEBUG
#	define DEBUG	// activate pr_debug
#	define DEBUG_VERBOSE
#endif

#define USE_CDEV 		// Recommended unless a legacy kernel requires otherwise
#define USE_UL_IOCTL	// Recommended unless a legacy kernel requires otherwise

#define DRIVER_NAME "AV PCL-PCI"

MODULE_AUTHOR(VER_LEGAL_COPYRIGHT);
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("AVPCL Driver Ver " AVPCL_VER_S "." AVPCL_SUB_S);

// -----------------------------------------------------------------------------
// These definitions were removed in Kernel 3.8

#ifndef __devinit
#define __devinit
#endif

#ifndef __devexit
#define __devexit
#endif

#ifndef __devexit_p
#define __devexit_p(x) x
#endif

// see:
//	http://lwn.net/Articles/537562/
//	http://elinux.org/Kernel_3.4_to_3.10_porting_guide
//	https://git.kernel.org/cgit/linux/kernel/git/torvalds/linux.git/commit/?id=54b956b903607
// -----------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
typedef struct {
	const struct pci_dev	*pdev;  // NULL: Card is inactive or not present
	unsigned long 		bar0Start;
	unsigned long 		bar0Length;
	unsigned long 		bar0Base;
} pclbridge_t;
static pclbridge_t pclBridges[PCI_BRIDGES];		// Array to hold bridges
#endif

typedef struct {
	const struct pci_dev	*pdev;  // NULL: Card is inactive or not present
	unsigned long 		bar0Start;
	unsigned long 		bar0Length;
	unsigned long 		bar0Base;
	unsigned long 		bar2Start;
	unsigned long 		bar2Length;
	unsigned long 		bar2Base;
	int 				iCounter;
	wait_queue_head_t	waitQueue;
} pclcard_t;
static pclcard_t pclCards[PCL_CARDS];		// Array to hold cards
static pclcard_t *pclCard[PCL_CARDS];		// Pointer for sorted access to cards
#if 0 //def PCL_USE_SHM_MUTEX
DEFINE_MUTEX(countMutex);
static int iCounter = 0; // Counter for access via proc filesystem
#endif

#ifdef USE_UL_IOCTL
DEFINE_MUTEX(ioctlMutex);
#endif

#ifdef USE_CDEV
typedef struct {
	struct cdev *cdev;
	uint8_t major;
	uint8_t minor;
} avpcldriver_t;
  
static avpcldriver_t avpcldriver = {
	.cdev = 0,
	.major = 0,
	.minor = 0,
};
#else
static int avpcl_major_number;
#endif // USE_CDEV

#ifdef USE_SIGNAL
struct task_struct signalRcvTask; // Task to receive the interrupt signal
#else
#endif

// -----------------------------------------------------------------------------

// PCI
static int avpcl_ioctl(struct inode *inode, struct file *instanz,
			   unsigned int cmd, unsigned long arg)
{
	int this_minor;

	if (_IOC_TYPE(cmd) != AVPCL_IOC_MAGIC) return -ENOTTY;
	this_minor = MINOR(inode -> i_rdev);

	if ((pclCards[this_minor].pdev == NULL && !IS_BRIDGE_COMMAND(cmd)) ||
		(pclBridges[this_minor].pdev == NULL &&  IS_BRIDGE_COMMAND(cmd)))
	{
//		if (IS_BRIDGE_COMMAND(cmd)) pr_debug("Nope %i (%X).\n", this_minor, cmd);
		return -EFAULT;
	} // end if
//	if (IS_BRIDGE_COMMAND(cmd)) pr_debug("Ok %i (%X).\n", this_minor, cmd);

//	pr_debug("ioctl cmd:%u card:%i\n", cmd, this_minor);

	switch (cmd) {

#ifdef USE_SIGNAL
	case (AVPCL_IOC_REGISTER_INT):{ // register signal receive task
			signalRcvTask = *current;
			return 0;
		} // end case
	case (AVPCL_IOC_UNREGISTER_INT):{ // revoke signal registration
			memset(&signalRcvTask, 0, sizeof(signalRcvTask));
			return 0;
		} // end case
#endif
	case (AVPCL_IOC_READDATA0):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATA0_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READDATA1):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATA1_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READDATA2):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATA2_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READDATA3):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATA3_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READDATA4):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATA4_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READDATA5):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATA5_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READDATA6):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATA6_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READDATA7):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATA7_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READDATA8):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATA8_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READDATA9):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATA9_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READDATAA):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_DATAA_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READINTFLGL):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_INTFLGL_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READINTFLGH):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_INTFLGH_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READCMDR):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_CMDR_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case (AVPCL_IOC_READACK):{
			u8 data = ioread8((u8*)(pclCard[this_minor] -> bar2Base + PCI_ACK_OFFSET));
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
	case AVPCL_IOC_READVERSION:{
			u8 version = ioread8((u8*) pclCard[this_minor] -> bar2Base + PCI_VERSION_OFFSET);
			if (copy_to_user((void *)arg, &version, sizeof(version))) return -EFAULT;
			return 0;
		} // end case
	case AVPCL_IOC_WRITECMD:{
			u8 comm;
			if (copy_from_user(&comm, (void*) arg, sizeof(comm))) return -EFAULT;
			iowrite8(comm, (u8*) pclCard[this_minor] -> bar2Base + PCI_CMDW_OFFSET);
			return 0;
		} // end case
	case AVPCL_IOC_WRITECMDD0:{
			u8 comm;
			if (copy_from_user(&comm, (void*) arg, sizeof(comm))) return -EFAULT;
			iowrite8(comm, (u8*) pclCard[this_minor] -> bar2Base + PCI_CMDD0_OFFSET);
			return 0;
		} // end case
	case AVPCL_IOC_WRITECMDD1:{
			u8 comm;
			if (copy_from_user(&comm, (void*) arg, sizeof(comm))) return -EFAULT;
			iowrite8(comm, (u8*) pclCard[this_minor] -> bar2Base + PCI_CMDD1_OFFSET);
			return 0;
		} // end case
	case AVPCL_IOC_WRITECFG:{
			u8 comm;
			if (copy_from_user(&comm, (void*) arg, sizeof(comm))) return -EFAULT;
			iowrite8(comm, (u8*) pclCard[this_minor] -> bar2Base + PCI_CFG_OFFSET);
			return 0;
		} // end case
	case AVPCL_IOC_WRITEINTCSR:{
			DWORD data;
			if (copy_from_user(&data, (void*) arg, sizeof(data))) return -EFAULT;
			switch (pclCard[this_minor] -> pdev -> device) {
				case AVPCL_DEVICE_ID_5V: 
				case AVPCL_DEVICE_ID_3V: 
					iowrite32(data, (u8*) pclCard[this_minor] -> bar0Base + PCI_INTCSR_9030_OFFSET); 
					break;
				case AVPCL_DEVICE_ID_E:  
					iowrite32(data, (u8*) pclCard[this_minor] -> bar0Base + PCI_INTCSR_9056_OFFSET); 
					break;
				default:
//					pr_debug("ioctl: AVPCL_IOC_WRITEINTCSR %i %lu\n", this_minor, data);
					return -EFAULT;
			} // end switch
			return 0;
		} // end case
	case AVPCL_IOC_WRITEINTACK:{
			switch (pclCard[this_minor] -> pdev -> device) {
				case AVPCL_DEVICE_ID_5V: 
				case AVPCL_DEVICE_ID_3V: 
					iowrite32(INTCSR_ACK_9030, (u8*) pclCard[this_minor] -> bar0Base + PCI_INTCSR_9030_OFFSET);
					break;
				case AVPCL_DEVICE_ID_E:  
					iowrite8(PCI_CFG_ACK_M, (u8*) pclCard[this_minor] -> bar2Base + PCI_CFG_OFFSET);
					break;
				default:
					return -EFAULT;
			} // end switch
			return 0;
		} // end case
#if 0 //def PCL_USE_SHM_MUTEX
		case AVPCL_COUNTER_GET:{
			mutex_lock(&countMutex);
			if (copy_to_user((void *)arg, &pclCard[this_minor] -> iCounter, sizeof(pclCard[this_minor] -> iCounter))) return -EFAULT;
			mutex_unlock(&countMutex);
			return 0;
		} // end case
		case AVPCL_COUNTER_INC:{
			mutex_lock(&countMutex);
			pclCards[this_minor].iCounter++;
			if (copy_to_user((void *)arg, &pclCard[this_minor] -> iCounter, sizeof(pclCard[this_minor] -> iCounter))) return -EFAULT;
			mutex_unlock(&countMutex);
			return 0;
		} // end case
		case AVPCL_COUNTER_DEC:{
			mutex_lock(&countMutex);
			pclCards[this_minor].iCounter--;
			if (copy_to_user((void *)arg, &pclCard[this_minor] -> iCounter, sizeof(pclCard[this_minor] -> iCounter))) return -EFAULT;
			mutex_unlock(&countMutex);
			return 0;
		} // end case
#endif // PCL_USE_SHM_MUTEX
		case AVPCL_BRIDGE_RETR_DEVID:
		case AVPCL_RETR_DEVID:
		{
			if (cmd == AVPCL_BRIDGE_RETR_DEVID) {
				if (copy_to_user((void *)arg, &(pclBridges[this_minor].pdev -> device), sizeof(pclBridges[this_minor].pdev -> device))) {
					return -EFAULT;
				} // end if
			} else {
				if (copy_to_user((void *)arg, &(pclCard[this_minor] -> pdev -> device), sizeof(pclCard[this_minor] -> pdev -> device))) {
					return -EFAULT;
				} // end if
			} // end else
			return 0;
		} // end case
		case AVPCL_RETR_SUBSYSDEVID:
		{
			u16 data = pclCard[this_minor] -> pdev -> subsystem_device;
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
#ifdef PCL_OPT_FLASH
		case AVPCL_BRIDGE_RETR_DEVINFO:
		case AVPCL_RETR_DEVINFO:
		{
			TDeviceInfo deviceInfo;
			memset(&deviceInfo, 0, sizeof(deviceInfo));
			deviceInfo.deviceId = (cmd == AVPCL_BRIDGE_RETR_DEVINFO) ? pclBridges[this_minor].pdev -> device :
									pclCard[this_minor] -> pdev -> device;
			deviceInfo.bus = (cmd == AVPCL_BRIDGE_RETR_DEVINFO) ? pclBridges[this_minor].pdev -> bus -> number :
									pclCard[this_minor] -> pdev -> bus -> number;

			deviceInfo.bar0Start = (cmd == AVPCL_BRIDGE_RETR_DEVINFO) ? pclBridges[this_minor].bar0Start :
									pclCard[this_minor] -> bar0Start;
			deviceInfo.bar0Length = (cmd == AVPCL_BRIDGE_RETR_DEVINFO) ? pclBridges[this_minor].bar0Length :
									pclCard[this_minor] -> bar0Length;
			deviceInfo.bar2Start = (cmd == AVPCL_BRIDGE_RETR_DEVINFO) ? 0 :
									pclCard[this_minor] -> bar2Start;
			deviceInfo.bar2Length = (cmd == AVPCL_BRIDGE_RETR_DEVINFO) ? 0 :
									pclCard[this_minor] -> bar2Length;

			if (copy_to_user((void *)arg, &deviceInfo, sizeof(deviceInfo))) return -EFAULT;
			return 0;
		}
#endif // PCL_OPT_FLASH
#ifdef PCL_OPT_FLASH
		case AVPCL_RETR_BUS:
		{
			u8 data = pclCard[this_minor] -> pdev -> bus -> number;
			if (copy_to_user((void *)arg, &data, sizeof(data))) return -EFAULT;
			return 0;
		} // end case
#endif // PCL_OPT_FLASH
#ifdef PCL_OPT_FLASH
		case AVPCL_BRIDGE_READ_CONFIG:
		case AVPCL_READ_CONFIG:
		{
			struct pci_dev* pdev;
			avpcl_read_config_t readConfig;
			if (cmd == AVPCL_BRIDGE_READ_CONFIG) pdev = (struct pci_dev*)pclBridges[this_minor].pdev;
			else pdev = (struct pci_dev*)pclCard[this_minor] -> pdev;
			if (pdev == NULL) return -ENXIO;
			if (copy_from_user(&readConfig, (void*) arg, sizeof(readConfig))) return -EFAULT;
			switch (readConfig.size) {
				case 0: return -EFAULT;
				case 1:
				{
					u8 val;
					if (pci_read_config_byte(pdev, readConfig.offset, &val)) {
						return -EFAULT;
					} // end if
					if (copy_to_user((void *)readConfig.result, &val, sizeof(val))) return -EFAULT;
					break;
				} // end case
				case 2:
				{
					u16 val;
					if (pci_read_config_word(pdev, readConfig.offset, &val)) {
						return -EFAULT;
					} // end if
					if (copy_to_user((void *)readConfig.result, &val, sizeof(val))) return -EFAULT;
					break;
				} // end case
				case 4:
				{
					u32 val;
					if (pci_read_config_dword(pdev, readConfig.offset, &val)) {
						return -EFAULT;
					} // end if
					if (copy_to_user((void *)readConfig.result, &val, sizeof(val))) {
						return -EFAULT;
					} // end if
					break;
				} // end case
				default:
				{
					return -EFAULT;
/*					u8 val;
					int n;
					for (n = 0; n < readConfig.size; n++) {
						if (pci_read_config_byte((struct pci_dev*)pclCard[this_minor] -> pdev, readConfig.offset + n, &val)) {
							return -EFAULT;
						} // end if
						if (copy_to_user((void *)(readConfig.result + n), &val, sizeof(val))) return -EFAULT;
					} // end for
					break;
*/				} // end case
			} // end switch
			return 0;
		} // end case
		case AVPCL_READ_BAR:
		case AVPCL_BRIDGE_READ_BAR:
		{
			avpcl_read_bar_t readBar;
			unsigned long base;
			if (copy_from_user(&readBar, (void*) arg, sizeof(readBar))) return -EFAULT;

			if (readBar.bar != AVPCL_BAR_0 && readBar.bar != AVPCL_BAR_2) return -EFAULT;

			if (cmd == AVPCL_BRIDGE_READ_BAR) {
				if (readBar.bar == AVPCL_BAR_0) {
					base = pclBridges[this_minor].bar0Base;
				} else {
					return -EFAULT;
				} // end else
			} else {
				base = (readBar.bar == AVPCL_BAR_0) ? 
						pclCard[this_minor] -> bar0Base :
						pclCard[this_minor] -> bar2Base;
			} // end else

			switch (readBar.size) {
				case 0: return -EFAULT;
				case 1:
				{
					u8 data = ioread8((u8*)(base + readBar.offset));
					if (copy_to_user((void *)readBar.result, &data, sizeof(data))) return -EFAULT;
					return 0;
				} // end case
				case 2:
				{
					u16 data = ioread16((u16*)(base + readBar.offset));
					if (copy_to_user((void *)readBar.result, &data, sizeof(data))) return -EFAULT;
					return 0;
				} // end case
				case 4:
				{
					u32 data = ioread32((u32*)(base + readBar.offset));
					if (copy_to_user((void *)readBar.result, &data, sizeof(data))) return -EFAULT;
					return 0;
				} // end case
				default:
				{
					return -EFAULT;
/*					int n;
					for (n = 0; n < readBar.size; n++) {
						u8 data = ioread8((u8*)(base + readBar.offset + n));
						if (copy_to_user((void *)(readBar.result + n), &data, sizeof(data))) return -EFAULT;
					} // end for
					return 0;
*/				} // end case
			} // end switch
			return -EIO;
		} // end case
		case AVPCL_WRITE_BAR:
		case AVPCL_BRIDGE_WRITE_BAR:
		{
			avpcl_write_bar_t writeBar;
			unsigned long base;
			if (copy_from_user(&writeBar, (void*) arg, sizeof(writeBar))) return -EFAULT;

			if (writeBar.bar != AVPCL_BAR_0 && writeBar.bar != AVPCL_BAR_2) return -EFAULT;

			if (cmd == AVPCL_BRIDGE_WRITE_BAR) {
				if (writeBar.bar == AVPCL_BAR_0) {
					base = pclBridges[this_minor].bar0Base;
				} else {
					return -EFAULT;
				} // end else
			} else {
				base = (writeBar.bar == AVPCL_BAR_0) ? 
						pclCard[this_minor] -> bar0Base :
						pclCard[this_minor] -> bar2Base;
			} // end else

			switch (writeBar.size) {
				case 0: return -EFAULT;
				case 1:
				{
					u8 data;
					if (copy_from_user(&data, (void*)writeBar.data, sizeof(data))) return -EFAULT;
					iowrite8(data, (u8*)(base + writeBar.offset));
					return 0;
				} // end case
				case 2:
				{
					u16 data;
					if (copy_from_user(&data, (void*)writeBar.data, sizeof(data))) return -EFAULT;
					iowrite16(data, (u16*)(base + writeBar.offset));
					return 0;
				} // end case
				case 4:
				{
					u32 data;
					if (copy_from_user(&data, (void*)writeBar.data, sizeof(data))) return -EFAULT;
					iowrite32(data, (u32*)(base + writeBar.offset));
					return 0;
				} // end case
				default:
				{
					return -EFAULT;
/*					u8 data;
					int n;
					for (n = 0; n < writeBar.size; n++) {
						if (copy_from_user(&data, (void*)(writeBar.data + n), sizeof(data))) return -EFAULT;
						iowrite8(data, (u8*)(base + writeBar.offset + n));
					} // end for
					return 0;
*/				} // end case
			} // end switch
			return -EIO;
		} // end case
		case AVPCL_TEST_BRIDGE:
		{
			if (this_minor < 0 || this_minor >= PCL_CARDS) return -ENXIO; // No such device or address
			if (pclBridges[this_minor].pdev == NULL) return -ENXIO;
			return 0;
		} // end case
#endif // PCL_OPT_FLASH
	default: 
		pr_info("avpcl: unknown ioctl cmd: %u\n", cmd);
		return -EINVAL;

	} // end switch
	return -EIO;
}

// -----------------------------------------------------------------------------

#ifdef USE_UL_IOCTL
// PCI
static long avpcl_unlocked_ioctl(struct file *filp, unsigned int cmd, 
							unsigned long arg)
{
	long err;

	mutex_lock(&ioctlMutex);
#if LINUX_VERSION_CODE <= KERNEL_VERSION(3,13,0)
	err = avpcl_ioctl(filp->f_dentry->d_inode, filp, cmd, arg);	// deprecated
#else
	err = avpcl_ioctl(file_inode(filp), filp, cmd, arg);
#endif
	mutex_unlock(&ioctlMutex);
	return err;
}
#endif

// -----------------------------------------------------------------------------

// PCI
static irqreturn_t pci_isr(int irq, void *dev_id)
{
	struct pci_dev *pdev = dev_id;
	int handled = 0;
	int cardIdx = 0;
//	u8 data0, data1, data2, data3;
#ifdef DEBUG_VERBOSE
//	pr_debug("pci_isr(irq:%d, dev_id:%p).\n", irq, dev_id);
#endif

	// Find card
	for (; cardIdx < PCL_CARDS; cardIdx++) {
		if (!memcmp(pclCards[cardIdx].pdev, pdev, sizeof(struct pci_dev))) {
			break;
		} // end if
	} // end for

	if (cardIdx == PCL_CARDS) {
		// error!
		pr_alert("  avpcl IRQ:%i - no card found!\n", irq);
		return IRQ_HANDLED;
	} // end if

	handled = 1;

	// Send signal
#ifdef USE_SIGNAL
//	if (signalRcvTask.pid != 0) kill(signalRcvTask.pid, SIGUSR1);
	if ((signalRcvTask.pid > 0) && (AV_SIGNAL > SIGRTMIN) && (AV_SIGNAL < SIGRTMAX)) {
		int ret = -1;
		struct siginfo info;
#ifdef DEBUG_VERBOSE
		pr_debug("Send signal %i to PID %i.\n", AV_SIGNAL, signalRcvTask.pid);
#endif
//		ret = send_sig_info(AV_SIGNAL, NULL, &signalRcvTask);
//		ret = send_sig(AV_SIGNAL, &signalRcvTask, 0);
		info.si_signo = AV_SIGNAL;
		info.si_errno = 0;
		info.si_code = SI_USER;
		info.si_pid = signalRcvTask.pid;
//		info.si_uid = signalRcvTask.uid;
//		kill_proc_info(AV_SIGNAL, &info, signalRcvTask.pid);
#ifdef DEBUG_VERBOSE
		if (!ret) pr_debug("Signal sent successfully.\n");
		else pr_debug("Signal not sent (%i)\n", ret);
#endif
	} else {
#ifdef DEBUG_VERBOSE
		pr_debug("Signal %i not sent (PID:%i).\n", AV_SIGNAL, signalRcvTask.pid);
#endif
	} // end else
#else
	// Wake waiting tasks
#ifdef DEBUG_VERBOSE
//	pr_debug("Wake waiting tasks.\n");
#endif
	wake_up_interruptible(&pclCards[cardIdx].waitQueue);
#endif

	// Set acknowledge in card
	switch (pdev -> device) {
		case PCI_DEFAULT_DEVICE_ID_PCIE:
			iowrite8(PCI_CFG_ACK_M, (u8*) pclCards[cardIdx].bar2Base + PCI_CFG_OFFSET);
//			pr_info("  avpcl IRQ:%i, Card:%i (PCIe)\n", irq, cardIdx);

			// Dummy Read from the CFG register
			ioread32((u8*) pclCards[cardIdx].bar2Base + PCI_VERSION_OFFSET);
			break;
		case PCI_DEFAULT_DEVICE_ID_5V:
		case PCI_DEFAULT_DEVICE_ID_3V:
			iowrite32(INTCSR_ACK_9030, (u8*) pclCards[cardIdx].bar0Base + PCI_INTCSR_9030_OFFSET);
//			pr_info("  avpcl IRQ:%i, Card:%i (PCI)\n", irq, cardIdx);
			break;
		default:
#ifdef DEBUG_VERBOSE
			pr_debug("  avpcl IRQ:%i - device %i unknown!!\n", irq, pdev -> device);
#endif
			break;
	} // end switch

	return IRQ_RETVAL(handled);
}

// -----------------------------------------------------------------------------
// PCI
static int device_compare(const void *elem1, const void *elem2) {
	pclcard_t *pcard1;
	pclcard_t *pcard2;
	int ret = 0; // equal per default
	int id1, id2;
	int bus1, bus2;
	int slot1, slot2;

	if (!elem1 || !elem2) {
		pr_alert("avpcldrv:device_compare: NULL pointer parameter.\n");
		return 0;
	} // end if

	pcard1 = *(pclcard_t **)elem1;
	pcard2 = *(pclcard_t **)elem2;

	if (pcard1 -> pdev == NULL && pcard2 -> pdev == NULL) { return  0; }
	if (pcard1 -> pdev == NULL) { return  1; }
	if (pcard2 -> pdev == NULL) { return -1; }

	// map device id
	switch (pcard1 -> pdev -> device) {
		case AVPCL_DEVICE_ID_5V: id1 = 0; break;
		case AVPCL_DEVICE_ID_3V: id1 = 1; break;
		case AVPCL_DEVICE_ID_E:  id1 = 2; break;
		default: id1 = 3;
	} // end switch
	switch (pcard2 -> pdev -> device) {
		case AVPCL_DEVICE_ID_5V: id2 = 0; break;
		case AVPCL_DEVICE_ID_3V: id2 = 1; break;
		case AVPCL_DEVICE_ID_E:  id2 = 2; break;
		default: id2 = 3;
	} // end switch

	// set bus
	if (pcard1 -> pdev -> bus  == NULL) { bus1 = 0; }
	else bus1 = pcard1 -> pdev -> bus  -> number;
	if (pcard2 -> pdev -> bus  == NULL) { bus2 = 0; }
	else bus2 = pcard2 -> pdev -> bus  -> number;

	// set slot
	if (pcard1 -> pdev -> slot == NULL) { slot1 = 0; }
	else slot1 = pcard1 -> pdev -> slot -> number;
	if (pcard2 -> pdev -> slot == NULL) { slot2 = 0; }
	else slot2 = pcard2 -> pdev -> slot -> number;

	ret = id1 - id2; // device id
	if (ret == 0) ret = bus1 - bus2;   // bus number
	if (ret == 0) ret = slot1 - slot2; // slot number

	return ret;
} // end device_compare()

// -----------------------------------------------------------------------------

// PCI
static int __devinit device_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
	int newIndex;
	int irq = pdev->irq;
/*
	bool bPcieV2 = FALSE;
	if (pdev->vendor == AVPCL_VENDOR_ID
	&&	pdev->device == AVPCL_DEVICE_ID_E
	&&	pdev->subsystem_vendor == AVPCL_SUBVENDOR_ID
	) {
		if (pdev->revision != AVPCL_REVISION_PCIE) {
			bPcieV2 = TRUE;
		}
	}
*/
#ifdef PCL_OPT_FLASH
	if (pdev -> device == PCI_DEFAULT_DEVICE_ID_PCIE_BRIDGE) {
//		pr_debug("PCI Bridge: 0x%x\n", pdev -> device);
		// find next free spot
		for (newIndex = 0; newIndex < PCL_CARDS; newIndex++) {
			if (pclBridges[newIndex].pdev == NULL) break;
		} // end for
		if (newIndex >= PCL_CARDS) {
			pr_err( DRIVER_NAME " : Maximum number of bridge devices reached.\n");
			return -EIO;
		} // end if
		// insert
		pclBridges[newIndex].pdev = pdev;
		pclBridges[newIndex].bar0Start  = pci_resource_start(pdev, 0);
		pclBridges[newIndex].bar0Length = pci_resource_len(  pdev, 0);

		if (request_mem_region(pclBridges[newIndex].bar0Start, pclBridges[newIndex].bar0Length, pdev->dev.kobj.name) == NULL 
			) {
#if 1 //def DEBUG_VERBOSE
			pr_err( DRIVER_NAME ": Memory address conflict for device \"%s\"\n",
				pdev -> dev.kobj.name);
#endif
			return -EIO;
		} // end if

		// map memory
#ifdef DEBUG_VERBOSE
		pr_debug("  Map mem\n");
#endif
		pclBridges[newIndex].bar0Base = (unsigned long) ioremap(pclBridges[newIndex].bar0Start, pclBridges[newIndex].bar0Length);

		return 0;
	} // end if
//	pr_debug("PCI Card: 0x%x\n", pdev -> device);
#endif

	// find next free spot
	for (newIndex = 0; newIndex < PCL_CARDS; newIndex++) {
		if (pclCards[newIndex].pdev == NULL) break;
	} // end for

	if (newIndex >= PCL_CARDS) {
		pr_err( DRIVER_NAME " : Maximum number of devices reached.\n");
		return -EIO;
	} // end if

	// insert card
	pclCards[newIndex].pdev = pdev;

	sort(pclCard, PCL_CARDS, sizeof(pclCard[0]), device_compare, NULL);

#if 0 //def DEBUG_VERBOSE
	for (sortedSlot = 0; sortedSlot < PCL_CARDS; sortedSlot++) {
		if (pclCard[sortedSlot] && pclCard[sortedSlot] -> pdev == pdev) break;
	} // end for
	pr_debug("  Inserted new card %x into slot %i\n", pdev -> device, sortedSlot);
#endif

#ifdef DEBUG_VERBOSE
	pr_debug("  PCI enable\n");
#endif
	if (pci_enable_device(pdev)) return -EFAULT;

	// allocate IRQ
#ifdef DEBUG_VERBOSE
	pr_debug("  Allocate IRQ\n");
#endif
	if (pci_alloc_irq_vectors(pdev, 1, 1, PCI_IRQ_ALL_TYPES) != 1) {
		pr_err( DRIVER_NAME ": Unable to allocate IRQ.\n");
		return -EIO;
	} // end if

	irq = pci_irq_vector(pdev, 0);
	if (irq < 0) {
		pr_err( DRIVER_NAME ": Unable to get IRQ number (%d).\n", irq);
		return -EIO;
	} // end if

	// request IRQ
#ifdef DEBUG_VERBOSE
	pr_debug("  Request IRQ %i\n", irq);
#endif
	if (request_irq(pdev->irq, pci_isr, IRQF_SHARED, "avpcl_drv", pdev)) {
		pr_err( DRIVER_NAME ": Unable to request IRQ %d.\n", irq);
		return -EIO;
	} // end if

	// Reserve memory
#ifdef DEBUG_VERBOSE
	pr_debug("  Reserve mem\n");
#endif
	pclCards[newIndex].bar0Start  = pci_resource_start(pdev, 0);
	pclCards[newIndex].bar0Length = pci_resource_len(  pdev, 0);
	pclCards[newIndex].bar2Start  = pci_resource_start(pdev, 2);
	pclCards[newIndex].bar2Length = pci_resource_len(  pdev, 2);
	if (request_mem_region(pclCards[newIndex].bar2Start, pclCards[newIndex].bar2Length, pdev->dev.kobj.name) == NULL
	||	request_mem_region(pclCards[newIndex].bar0Start, pclCards[newIndex].bar0Length, pdev->dev.kobj.name) == NULL 
	) {
#if 1 //def DEBUG_VERBOSE
		pr_err( DRIVER_NAME ": Memory address conflict for device \"%s\"\n",
			pdev->dev.kobj.name);
#endif
		free_irq(pdev->irq, pdev);
		pci_free_irq_vectors(pdev);

		return -EIO;
	} // end if

	// map memory
#ifdef DEBUG_VERBOSE
	pr_debug("  Map mem\n");
#endif
	pclCards[newIndex].bar2Base = (unsigned long) ioremap(pclCards[newIndex].bar2Start, pclCards[newIndex].bar2Length);
	pclCards[newIndex].bar0Base = (unsigned long) ioremap(pclCards[newIndex].bar0Start, pclCards[newIndex].bar0Length);

	// reset counter
	pclCards[newIndex].iCounter = 0;
//	pr_info("   >>>>>Counter reset: %i %i", newIndex, pclCards[newIndex].iCounter);

	return 0;
}

// -----------------------------------------------------------------------------

// PCI
static void __devexit device_remove(struct pci_dev *pdev)
{
	int removeIndex;
#ifdef DEBUG_VERBOSE
	pr_debug("Remove\n");
	pr_debug("  PCI disable\n");
#endif
	pci_disable_device(pdev);

#ifdef PCL_OPT_FLASH
	if (pdev->device == PCI_DEFAULT_DEVICE_ID_PCIE_BRIDGE) {
		// find
		for (removeIndex = 0; removeIndex < PCL_CARDS; removeIndex++) {
			if (pclBridges[removeIndex].pdev == pdev) break;
		} // end for
		if (removeIndex >= PCL_CARDS) {
			pr_alert("  Bridge not found. Invalid index: %i.\n", removeIndex);
			return;
		} // end if
		pr_debug("Remove PCI Bridge: 0x%x\n", pdev -> device);

		if (pclBridges[removeIndex].bar0Start)
			release_mem_region(pclBridges[removeIndex].bar0Start, pclBridges[removeIndex].bar0Length);
		if (pclBridges[removeIndex].bar0Base)
			iounmap((void *) pclBridges[removeIndex].bar0Base);

		return;
	} // end if
#endif

	for (removeIndex = 0; removeIndex < PCL_CARDS; removeIndex++) {
		if (pclCards[removeIndex].pdev == pdev) break;
	} // end for

	if (removeIndex >= PCL_CARDS) {
		pr_alert("  Card not found. Invalid index: %i.\n", removeIndex);
		return;
	} // end if

#ifdef DEBUG_VERBOSE
	pr_debug("  Removing card %x from slot %i\n", pdev -> device, removeIndex);
#endif

#ifdef DEBUG_VERBOSE
	pr_debug("  Free IRQ %i\n", pdev -> irq);
#endif
	free_irq(pdev->irq, pdev);
	pci_free_irq_vectors(pdev);

#ifdef DEBUG_VERBOSE
	pr_debug("  Release mem\n");
#endif
	if (pclCards[removeIndex].bar0Start)
		release_mem_region(pclCards[removeIndex].bar0Start, pclCards[removeIndex].bar0Length);
	if (pclCards[removeIndex].bar2Start)
		release_mem_region(pclCards[removeIndex].bar2Start, pclCards[removeIndex].bar2Length);
#ifdef DEBUG_VERBOSE
		pr_debug("  Unmap mem\n");
#endif
	if (pclCards[removeIndex].bar0Base)
		iounmap((void *) pclCards[removeIndex].bar0Base);
	if (pclCards[removeIndex].bar2Base)
		iounmap((void *) pclCards[removeIndex].bar2Base);
}

// -----------------------------------------------------------------------------

// PCI
#define AVPCL_DEVICE(_device, _subdevice) \
	.device = _device, .subdevice = _subdevice, \
	.vendor = AVPCL_VENDOR_ID, .subvendor = AVPCL_SUBVENDOR_ID, \
	.class = 0, .class_mask = 0, .driver_data = 0

static const struct pci_device_id pci_drv_tbl[] =
{
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_5V, AVPCL_SUBSYSTEM_DEVICE_ID_LV_5V)		},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_5V, AVPCL_SUBSYSTEM_DEVICE_ID_D_5V)		},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_5V, AVPCL_SUBSYSTEM_DEVICE_ID_TS_5V)		},

	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_3V, AVPCL_SUBSYSTEM_DEVICE_ID_LV_3V)		},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_3V, AVPCL_SUBSYSTEM_DEVICE_ID_TS_3V)		},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_3V, AVPCL_SUBSYSTEM_DEVICE_ID_D_3V)		},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_3V, AVPCL_SUBSYSTEM_DEVICE_ID_L_3V)		},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_3V, AVPCL_SUBSYSTEM_DEVICE_ID_L_V2_3V)	},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_3V, AVPCL_SUBSYSTEM_DEVICE_ID_HD_3V)		},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_3V, AVPCL_SUBSYSTEM_DEVICE_ID_D_V2_3V)	},

	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_E,  AVPCL_SUBSYSTEM_DEVICE_ID_HD_PCIE)	},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_E,  AVPCL_SUBSYSTEM_DEVICE_ID_D_PCIE)	},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_E,  AVPCL_SUBSYSTEM_DEVICE_ID_LV_PCIE)	},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_E,  AVPCL_SUBSYSTEM_DEVICE_ID_L_PCIE)	},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_E,  AVPCL_SUBSYSTEM_DEVICE_ID_TS_PCIE)	},
	{ AVPCL_DEVICE(AVPCL_DEVICE_ID_E,  AVPCL_SUBSYSTEM_DEVICE_ID_3G_PCIE)	},

#ifdef PCL_OPT_FLASH
	// PCI bridge
	{ .device = PCI_DEFAULT_DEVICE_ID_PCIE_BRIDGE, .vendor = AVPCL_VENDOR_ID,
	  .subdevice = 0, .subvendor = 0, .class = 0, .class_mask = 0, .driver_data = 0 },
#endif

	{ 0, }
};

MODULE_DEVICE_TABLE(pci, pci_drv_tbl);

// -----------------------------------------------------------------------------

// PCI
static struct pci_driver pci_drv_driver = {
	.name     = DRIVER_NAME,
	.id_table = pci_drv_tbl,
	.probe    = device_probe,
	.remove   = __devexit_p(device_remove),
};

// -----------------------------------------------------------------------------

static void avpcl_interruptible_sleep_on(wait_queue_head_t* waitQueue)
{
	DEFINE_WAIT(wait);
	prepare_to_wait(waitQueue, &wait, TASK_INTERRUPTIBLE);
	schedule_timeout(HZ); // wait for up to one second; alternatively, just: schedule();
	finish_wait(waitQueue, &wait);
}

// -----------------------------------------------------------------------------

// DEV
static ssize_t avpcl_read(struct file *filp, char __user *buffer, size_t length, loff_t *offset)
{
	long iMinor = (long)filp -> private_data;
	if (iMinor < PCL_CARDS) {
//		interruptible_sleep_on(&pclCard[iMinor] -> waitQueue);	// deprecated
//		wait_event_interruptible(pclCard[iMinor] -> waitQueue, TRUE);
		avpcl_interruptible_sleep_on(&pclCard[iMinor] -> waitQueue);
	} else {
		pr_alert("avpcl_read: Invalid minor number (%li)", iMinor);
	} // end else
  	return 0;
}
// -----------------------------------------------------------------------------

static int avpcl_open (struct inode *inode, struct file *filp) {
	long rdev = MINOR(inode -> i_rdev);
	filp -> private_data = (void *)rdev;
	return 0;
}

// -----------------------------------------------------------------------------
/*
// DEV
static ssize_t avpcl_write(struct file *filp, const char __user *buff, size_t len, loff_t *off)
{
	if (len > 199)
		return -EINVAL;
	copy_from_user(msg, buff, len);
	msg[len] = '\0';
	wake_up_interruptible(&wait_queue);
	return len;
}
*/
// -----------------------------------------------------------------------------

// PCI
static struct file_operations driver_fops = {
	.owner = THIS_MODULE,
#ifdef USE_UL_IOCTL
	.unlocked_ioctl = avpcl_unlocked_ioctl,
#else
	.ioctl = avpcl_ioctl,
#endif
	.read = avpcl_read,
	.open = avpcl_open,
};

// ---------------------------------------------------------------------

// PROC
#if 0 //def PCL_USE_SHM_MUTEX
static int counterInc(struct seq_file *m, void *v) {
	mutex_lock(&countMutex);
	seq_printf(m, "%i\n", ++iCounter);
	mutex_unlock(&countMutex);
	return 0;
} // end counterShow()
#endif // PCL_USE_SHM_MUTEX

// ---------------------------------------------------------------------

// PROC
#if 0 //def PCL_USE_SHM_MUTEX
static int counterDec(struct seq_file *m, void *v) {
	mutex_lock(&countMutex);
	seq_printf(m, "%i\n", --iCounter);
	mutex_unlock(&countMutex);
	return 0;
} // end counterShow()
#endif // PCL_USE_SHM_MUTEX

// ---------------------------------------------------------------------

// PROC
#if 0 //def PCL_USE_SHM_MUTEX
static int counterOpenInc(struct inode *inode, struct file *file) {
	return single_open(file, counterInc, NULL);
} // end couterOpen()
#endif

// ---------------------------------------------------------------------

// PROC
#if 0 //def PCL_USE_SHM_MUTEX
static int counterOpenDec(struct inode *inode, struct file *file) {
	return single_open(file, counterDec, NULL);
} // end couterOpen()
#endif
// ---------------------------------------------------------------------

// PROC
#if 0 //def PCL_USE_SHM_MUTEX
static const struct file_operations proc_fopsInc = {
	.owner		= THIS_MODULE,
	.open		= counterOpenInc,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};
#endif

// ---------------------------------------------------------------------

// PROC
#if 0 //def PCL_USE_SHM_MUTEX
static const struct file_operations proc_fopsDec = {
	.owner		= THIS_MODULE,
	.open		= counterOpenDec,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};
#endif

// -----------------------------------------------------------------------------

static int __init avpcl_init(void)
{

#ifdef USE_CDEV
	avpcldriver_t *driver = &avpcldriver;
	struct cdev *cdev;
	dev_t devt;
	int rc;
#endif
	int n;

#ifdef DEBUG_VERBOSE
	pr_info("**********\n");
	pr_info("avpcl_init\n");
#endif

	// Create proc file
#if 0 //def PCL_USE_SHM_MUTEX
#ifdef DEBUG_VERBOSE
	pr_info("  create /proc/" PROC_FILENAME_INC " and " PROC_FILENAME_DEC "\n");
#endif
	if (!(proc_create(PROC_FILENAME_INC, 0666, NULL, &proc_fopsInc) &&
		  proc_create(PROC_FILENAME_DEC, 0666, NULL, &proc_fopsDec))) {
		return -ENOMEM;
	} // end if
#endif // PCL_USE_SHM_MUTEX

	// Initialize arrays
	memset(pclBridges, 0, sizeof(pclBridges));
	memset(pclCards, 0, sizeof(pclCards));
	for (n = 0; n < PCL_CARDS; n++) {
		pclCard[n] = &pclCards[n];
		init_waitqueue_head(&pclCards[n].waitQueue);
	} // end for

#ifdef USE_CDEV
#ifdef DEBUG_VERBOSE
	pr_info("  pci_register_driver\n");
#endif
	if (pci_register_driver(&pci_drv_driver)) {
		pr_err( "Registration of PCI driver failed.\n");
	} // end if

#ifdef DEBUG_VERBOSE
	pr_info("  alloc_chrdev_region\n");
#endif
	rc = alloc_chrdev_region(&devt, avpcldriver.minor, PCL_CARDS, DRIVER_NAME);
	driver -> major = MAJOR(devt);

	if (rc) {
		pr_err( "Driver registration failed (Major:%d)\n", driver -> major);
		unregister_chrdev_region(devt, 1);
#if 0 //def PCL_USE_SHM_MUTEX
		remove_proc_entry(PROC_FILENAME_INC, NULL);
		remove_proc_entry(PROC_FILENAME_DEC, NULL);
#endif
		return -EIO;
	} // end if

#ifdef DEBUG_VERBOSE
	pr_info("  cdev_alloc\n");
#endif
	rc = -ENOMEM;
	cdev = cdev_alloc();

	if (!cdev) {
		pr_err( "Driver cdev alloc failed (Major:%d)\n", driver -> major);
		cdev_del(cdev);
		unregister_chrdev_region(devt, 1);
#if 0 //def PCL_USE_SHM_MUTEX
		remove_proc_entry(PROC_FILENAME_INC, NULL);
		remove_proc_entry(PROC_FILENAME_DEC, NULL);
#endif
		return -EIO;
	} // end if

	cdev -> owner = THIS_MODULE;
	cdev -> ops = &driver_fops;

#ifdef DEBUG_VERBOSE
	pr_info("  cdev_add\n");
#endif
	rc = cdev_add(cdev, devt, PCL_CARDS);
	if (rc) {
		pr_err( "Driver registration failed (Major:%d)\n", driver -> major);
		cdev_del(cdev);
		unregister_chrdev_region(devt, 1);
#if 0 //def PCL_USE_SHM_MUTEX
		remove_proc_entry(PROC_FILENAME_INC, NULL);
		remove_proc_entry(PROC_FILENAME_DEC, NULL);
#endif
		return -EIO;
	} // end if

	driver -> cdev = cdev;

	pr_info("Driver " DRIVER_NAME " started at major number %i\n", driver -> major);

	return 0;

#else
	// register device
	avpcl_major_number = register_chrdev(0 /*automatic Majornumber*/, DRIVER_NAME, &driver_fops);
	if (avpcl_major_number != 0) {
		if (pci_register_driver(&pci_drv_driver) != 0) {
			pr_err( "PCI init failed\n");
			unregister_chrdev(avpcl_major_number, DRIVER_NAME);
#if 0 //def PCL_USE_SHM_MUTEX
			remove_proc_entry(PROC_FILENAME_INC, NULL);
			remove_proc_entry(PROC_FILENAME_DEC, NULL);
#endif
			return -EIO;
		} // end if
#ifdef DEBUG
		pr_debug("Driver registered (Major number %i)\n", avpcl_major_number);
#endif
		return 0;
	} // end if
	pr_err( "Init failed\n");
#if 0 //def PCL_USE_SHM_MUTEX
	remove_proc_entry(PROC_FILENAME_INC, NULL);
	remove_proc_entry(PROC_FILENAME_DEC, NULL);
#endif
	return -EIO;
#endif // USE_CDEV
}

// -----------------------------------------------------------------------------

static void __exit avpcl_exit(void)
{

#ifdef USE_CDEV
	avpcldriver_t *driver = &avpcldriver;

#ifdef DEBUG_VERBOSE
	pr_info("**********\n");
	pr_info("avpcl_exit\n");
#endif

	if (driver->cdev) {
#ifdef DEBUG_VERBOSE
		pr_info("  unregister_chrdev_region\n");
#endif
		unregister_chrdev_region(driver->cdev->dev, PCL_CARDS);
#ifdef DEBUG_VERBOSE
		pr_info("  cdev_del\n");
#endif
		cdev_del(driver->cdev);
#ifdef DEBUG_VERBOSE
		pr_info("  pci_unregister_driver\n");
#endif
		pci_unregister_driver(&pci_drv_driver);
	} // end if
#else
	unregister_chrdev(avpcl_major_number, DRIVER_NAME);
	pci_unregister_driver(&pci_drv_driver);

#ifdef DEBUG
	pr_debug("Driver unloaded.\n");
#endif // DEBUG
#endif // USE_CDEF

	// Remove proc file
#if 0 //def PCL_USE_SHM_MUTEX
#ifdef DEBUG_VERBOSE
	pr_info("  remove /proc/" PROC_FILENAME_INC " and " PROC_FILENAME_DEC "\n");
#endif
	remove_proc_entry(PROC_FILENAME_INC, NULL);
	remove_proc_entry(PROC_FILENAME_DEC, NULL);
#endif

	pr_info("Driver " DRIVER_NAME " stopped.\n");
}

// -----------------------------------------------------------------------------

module_init(avpcl_init);
module_exit(avpcl_exit);
