/**********************************************\
*                                              *
*  Linux Kernel Driver for PCL PCI(e) Devices  *
*                                              *
*        Part of the shared library            *
*                                              *
\**********************************************/

// $Id: funcs.cpp 2123 2023-10-05 14:32:37Z J.Loh $
// Version: 5.45 (c) 1995-2023 PLURA Europe GmbH

//---------------------------------------------------------------------------
#include "params.h"
//---------------------------------------------------------------------------
//#include <stdio.h>
#include <stddef.h>
//#ifndef LINUX
#include <string.h>
#include <assert.h>
//#endif
//---------------------------------------------------------------------------
#	include "avpcl.h"
#ifdef _WIN32
#	include "pci_lib.h"
#endif
#ifdef LINUX
#	include <stdio.h>
#	include <stdarg.h>
#	include <sys/errno.h>	// errno
#	include "pci_lib_linux.h"
#	ifndef __USE_BSD
#		define __USE_BSD
#		include <sys/time.h>
#		undef __USE_BSD
#	else
#		include <sys/time.h>
#	endif
#	define vsprintf_s vsnprintf
#endif // LINUX
#include "global.h"
//---------------------------------------------------------------------------
#undef	_PCL_EXTERN
#define _PCL_EXTERN
#undef	_PCL_CONST
#define _PCL_CONST
#include "funcs.h"
//---------------------------------------------------------------------------

TPciCard gPciCard = { -1 };

TPciCardLocal gPciCardLocal = { -1 };

//---------------------------------------------------------------------------

#ifdef LINUX

TPciShared gPciShared = {
	{ -1 }, // PciCardShared
	{ -1 }, // PciHdl
};

TPciCardShared *pgPciCardShared = &gPciShared.PciCardShared;
TPciHdl *pgPciHdl = &gPciShared.PciHdl;

#endif

//---------------------------------------------------------------------------

#ifdef LINUX
#	define PCI_ErrorStringOffset 0
#endif

const char *_PCL_FDEF pclPciGetErrorString(
	void
) {
	PCI_ErrorString[PCI_ErrorStringOffset] = '\0';
	return(PCI_ErrorString);
}

//---------------------------------------------------------------------------

const char * _PCL_FDEF pclGetErrorString(	// @68, get error string
	void									// no parameter needed
) {
	ASSERT_STATIC(sizeof(BYTE)  == 1);
	ASSERT_STATIC(sizeof(WORD)  == 2);
	ASSERT_STATIC(sizeof(DWORD) == 4);

	ASSERT_STATIC((PCL_VER_DLL >> 8)   == AVPCL_VER);
	ASSERT_STATIC((PCL_VER_DLL & 0xFF) == AVPCL_SUB);
	ASSERT_STATIC((AVPCL_SUB % 10) != 0);	// Null am Ende der Versionsnummer vermeiden
	ASSERT_STATIC((AVPCL_SUB % 10) != (AVPCL_SUB / 10));	// Zwei gleiche Versionsnummern vermeiden

	return(pclErrorString);
}

//---------------------------------------------------------------------------
#ifdef PCL_DEBUG_LOG

#ifdef LINUX

void GetTimeStamp(PCL_TIME_STAMP *stamp)
{
	assert(stamp != NULL);
	if (clock_gettime(CLOCK_MONOTONIC_RAW, stamp) != 0) {
		stamp->tv_sec = stamp->tv_nsec = 0;
	}
}

void GetTimeStampFreq(PCL_TIME_STAMP *freq)
{
	assert(freq != NULL);
	freq->tv_sec = 1;
	freq->tv_nsec = 0;
}

void DiffTimeStamp(PCL_TIME_STAMP *diff, const PCL_TIME_STAMP *s1, const PCL_TIME_STAMP *s2)
{
	assert(diff != NULL);
	assert(s1 != NULL);
	assert(s2 != NULL);

	diff->tv_sec = s1->tv_sec - s2->tv_sec;
	diff->tv_nsec = s1->tv_nsec - s2->tv_nsec;
	if (diff->tv_nsec < 0) {
		diff->tv_nsec += 1000 * 1000 * 1000;
		diff->tv_sec--;
	}
}

BOOL TimeStampIsZero(const PCL_TIME_STAMP *stamp)
{
	assert(stamp != NULL);
	return stamp->tv_sec == 0 && stamp->tv_nsec == 0;
}

int TimeStampToMicro(const PCL_TIME_STAMP *stamp, const PCL_TIME_STAMP *freq)
{
	assert(stamp != NULL);
	assert(freq != NULL);

	return stamp->tv_sec * 1000 * 1000 + stamp->tv_nsec / 1000;
}

#else

void GetTimeStamp(PCL_TIME_STAMP *stamp)
{
	assert(stamp != NULL);
	QueryPerformanceCounter(stamp);
}

void GetTimeStampFreq(PCL_TIME_STAMP *freq)
{
	QueryPerformanceFrequency(freq);
}

void DiffTimeStamp(PCL_TIME_STAMP *diff, const PCL_TIME_STAMP *s1, const PCL_TIME_STAMP *s2)
{
	assert(s1 != NULL);
	assert(s2 != NULL);
	assert(diff != NULL);
	diff->QuadPart = s1->QuadPart - s2->QuadPart;
}

BOOL TimeStampIsZero(const PCL_TIME_STAMP *stamp)
{
	return stamp->QuadPart == 0;
}

int TimeStampToMicro(const PCL_TIME_STAMP *stamp, const PCL_TIME_STAMP *freq)
{
	return (int)(stamp->QuadPart * 1000000i64 / freq->QuadPart);
}

#endif // LINUX

#endif // PCL_DEBUG_LOG
//---------------------------------------------------------------------------
#ifdef PCL_DEBUG_LOG

void appendDebugLog(int myhdl, const char *format, ...)
{
	va_list args;
	char buffer[MAX_PATH];
	int hdl;

	PCL_TIME_STAMP count;
	GetTimeStamp(&count);

	assert(pgPciHdl);
#ifdef PCL_MULTI_MODE
	if (pgPciHdl->v[myhdl].type == TYPE_MULTI) {
		hdl = pgPciHdl->v[myhdl].multi.hdl;
	} else
#endif // PCL_MULTI_MODE
	{
		hdl = myhdl;
	}

	int card = pgPciHdl->v[hdl].card;
	CHECK_CARD(card);

	FILE *file = pgPciHdl->v[hdl].file;
	if (!file) return;

	int mikro;
	PCL_TIME_STAMP diff;
	DiffTimeStamp(&diff, &count, &gPciCard.v[card].perf.abs);

	if (!TimeStampIsZero(&gPciCard.v[card].perf.freq)) {
		mikro = TimeStampToMicro(&diff, &gPciCard.v[card].perf.freq);
	} else {
		mikro = 0;
	}

	PCI_HANDLE hPCI = gPciCard.v[card].hPCI;
	char sCardFlags[] = "................";
	if (hPCI) {
#ifdef LINUX
//		strcpy(sCardFlags, "----------------");
#else
		strcpy_s(sCardFlags, _countof(sCardFlags), "----------------");
#endif // LINUX
		int cardFlags = PCI_ReadINTFLGL(hPCI);
		if (gPciCard.v[card].irq16) cardFlags |= PCI_ReadINTFLGH(hPCI) << 8;
//		sprintf(sCardFlags, "%04X", cardFlags);

		if (cardFlags & PCL_IMASK_MIXED		)	sCardFlags[15] = 'm';	/* 0x0001 mix disable/enable */
		if (cardFlags & 0x0002				)	sCardFlags[14] = 'x';
		if (cardFlags & PCL_IMASK_LTC		)	sCardFlags[13] = 'l';	/* 0x0004 ltc disable/enable */
		if (cardFlags & PCL_IMASK_VITC		)	sCardFlags[12] = 'v';	/* 0x0008 vitc disable/enable */
		if (cardFlags & 0x0010				)	sCardFlags[11] = 'x';
		if (cardFlags & 0x0020				)	sCardFlags[10] = 'x';
		if (cardFlags & PCL_IMASK_V1		)	sCardFlags[9]  = '1';	/* 0x0040 odd video field */
		if (cardFlags & PCL_IMASK_V2		)	sCardFlags[8]  = '2';	/* 0x0080 even video field */

		if (cardFlags & PCL_IMASK_ATC_LTC	)	sCardFlags[7]  = 'L';	/* 0x0100 atc ltc disable/enable */
		if (cardFlags & PCL_IMASK_ATC_VITC	)	sCardFlags[6]  = 'V';	/* 0x0200 atc vitc disable/enable */
		if (cardFlags & 0x0400				)	sCardFlags[5]  = 'x';
		if (cardFlags & PCL_IMASK_HANC_LTC	)	sCardFlags[4]  = 'X';	/* 0x0800 hanc ltc disable/enable */
		if (cardFlags & PCL_IMASK_HANC_VITC	)	sCardFlags[3]  = 'X';	/* 0x1000 hanc vitc disable/enable */
		if (cardFlags & PCL_IMASK_VBI		)	sCardFlags[2]  = 'I';	/* 0x2000 vbi disable/enable */
		if (cardFlags & PCL_IMASK_ANC		)	sCardFlags[1]  = 'A';	/* 0x4000 anc disable/enable */
		if (cardFlags & 0x8000				)	sCardFlags[0]  = 'x';
	}

	sprintf_s(buffer, sizeof(buffer), "%3d:%03d.%03d %d-%d %s ", mikro / 1000000, (mikro / 1000) % 1000, mikro % 1000, hdl, card, sCardFlags);
/*
	SYSTEMTIME localTime;
	GetLocalTime(&localTime);
	sprintf_s(buffer, sizeof(buffer), "%02d.%02d.%04d %2d:%02d:%02d.%03d - ",
		localTime.wDay, localTime.wMonth, localTime.wYear,
		localTime.wHour, localTime.wMinute, localTime.wSecond, localTime.wMilliseconds
	);
*/
	fputs(buffer, file);

	va_start(args, format);
	vsprintf_s(buffer, sizeof(buffer), format, args);
	{
		// replace control chars
		char *p;
		while ((p = strchr(buffer, '\n')) != NULL) {
			*p = ' ';
		}
		while ((p = strchr(buffer, '\r')) != NULL) {
			*p = ' ';
		}
	}
	fputs(buffer, file);
	fputs("\n", file);
}

#endif // PCL_DEBUG_LOG
//---------------------------------------------------------------------------

int doResetPor(int hdl)
{
	APPEND_DEBUG_LOG(hdl, "> doResetPor(hdl:%d, ...)", hdl, PCL_OK);
	int result = PCL_ERR_TIMEOUT;
#	define doResetPorMulti pclMultiNotSupported
	PCI_GET_HANDLE(hdl, doResetPor, (), hPCI, card);
	int k, i;

#ifdef PCL_INTERRUPTS
	PCI_WriteDisable(hdl);	// disable linti1
#endif

	for (k = 0; k < 5; k++) {
		PCI_WriteCFG(hPCI, PCI_CFG_RESET_M);
		Sleep(50);
		PCI_WriteCFG(hPCI, 0);

		Sleep(PCL_TIMEOUT * 2);
		pclPciExec(hdl, CMD_GET_VERSION);	// UNKNOWN CMD
		for (i = 0; i < 3; i++) {
			if ((result = pclPciExec(hdl, CMD_GET_VERSION)) == PCL_OK) {
				break;
			}
			Sleep(PCL_TIMEOUT);
		}
		if (result == PCL_OK) break;
	}

#ifdef PCL_INTERRUPTS
	PCI_WriteAck(hdl);	// acknowledge linti1
#endif

	APPEND_DEBUG_LOG(hdl, "< doResetPor(hdl:%d, ...) = %d", hdl, result);
	return(result);
}

//---------------------------------------------------------------------------

int getPciImask(int card)
{
	int imask = 0;
	int hdl;

	assert(pgPciHdl);
	for (hdl = 0; hdl < PCL_HANDLES; hdl++) {
		if (pgPciHdl->v[hdl].type == TYPE_PCI && pgPciHdl->v[hdl].card == card) {
			imask |= pgPciHdl->v[hdl].imask;
#ifdef PCL_OPT_ANC				/* read user defined ancillary data */
			imask |= pgPciHdl->v[hdl].anc.imask;
#endif
		}
	}
	return(imask);
}

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int getMultiImask(int card)
{
	int imask = 0;

	assert(pgPciHdl);
	for (int hdl = 0; hdl < PCL_HANDLES; hdl++) {
		if (pgPciHdl->v[hdl].type == TYPE_MULTI) {
			imask |= pgPciHdl->v[hdl].imask;
		}
	}
	return(imask);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

int _PCL_FDEF pclReset(	// $01,	reset: initialize pcl card
	int hdl,			// in:	handle from pclOpen()
	int imode			// in:	operation mode of pcl like PCL_IMODE_xxx
) {
	int result = PCL_ERR_TIMEOUT;
#	define pclResetMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclReset, (), hPCI, card);

	if (imode == PCL_IMODE_POR) {
		result = doResetPor(hdl);
	} else {
		PCI_WriteCMDD0(hPCI, imode);
		APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", imode);
		result = pclPciExec(hdl, CMD_RESET);
		if (result != PCL_OK) {
			// Workaround:
			// Older PCL PCI (e.g. PCL PCI LV with firmware V7)
			// run into timeout after reset.
			// Check if the card is accessible again
			result = pclPciExec(hdl, CMD_GET_VERSION);
		}
#ifdef PCL_INTERRUPTS
		PCI_WriteAck(hdl);	// acknowledge linti1
#endif
	}
	UNLOCK_RETURN(hdl, pclReset, result);
}

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int _PCL_FDEF pclPciReopenMulti(
	int hdl							// in:	handle from pclOpen()
) {
	int result = PCL_OK;
	int multiHdl;

	assert(pgPciHdl);
//	THdlType type = pgPciHdl->v[hdl].type;
//	pgPciHdl->v[hdl].type = TYPE_RESERVED;
	if ((result = pclOpen(&multiHdl, pgPciHdl->v[hdl].card + PCL_BASE_PCI_OFS + 1 | PCL_BASE_OPTION_MULTI)) == PCL_OK) {
		assert(pgPciHdl->v[multiHdl].multi.main);
		pgPciHdl->v[multiHdl].multi.main = FALSE;
		for (int i = 0; i < PCL_HANDLES; i++) {
			if (i != multiHdl) {
				if (pgPciHdl->v[i].type == TYPE_MULTI_WAIT) {
					if (pgPciHdl->v[i].card == pgPciHdl->v[multiHdl].card) {
						pgPciHdl->v[i] = pgPciHdl->v[multiHdl];
					}
				}
			}
		}
		pgPciHdl->v[hdl] = pgPciHdl->v[multiHdl];
		pgPciHdl->v[hdl].multi.main = TRUE;
/*
		pgPciHdl->v[hdl].type = type;
		pgPciHdl->v[hdl].multi.hdl = pgPciHdl->v[multiHdl].multi.hdl;
*/
		pgPciHdl->v[multiHdl].type = TYPE_UNUSED;
//		pclClose(multiHdl);
	} else {
//		pgPciHdl->v[hdl].type = type;
		pclClose(hdl);
	}
	return(result);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int _PCL_NDEF pclGetVersionMulti(	// @59, $46, read pcl firmware version number
	int hdl,						// in:	handle from pclOpen()
	int _PCL_OUT *ident,			// out:	software identification
	int _PCL_OUT *firmware,			// out:	firmware revision number
	int _PCL_OUT *dll				// out:	dll revision number
) {
	PCI_GET_HANDLE_MULTI(hdl, pclGetVersion, (hdl, ident, firmware, dll), hPCI, card);

	int card = pgPciHdl->v[hdl].card;
	CHECK_CARD(card);
	if (ident)    *ident    = pgPciCardShared->v[card].multi.pclGetVersion.ident;
	if (firmware) *firmware = pgPciCardShared->v[card].multi.pclGetVersion.firmware;

	return(PCL_OK);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetVersion(	// @59, $46, read pcl firmware version number
	int hdl,					// in:	handle from pclOpen()
	int _PCL_OUT *ident,		// out:	software identification
	int _PCL_OUT *firmware,		// out:	firmware revision number
	int _PCL_OUT *dll			// out:	dll revision number
) {
	APPEND_DEBUG_LOG(hdl, "> pclGetVersion(hdl:%d, ...)", hdl, PCL_OK);
	if (dll != NULL) {
		*dll = (AVPCL_VER << 8) + AVPCL_SUB;
#ifndef LINUX
		assert(PCL_VER_DLL == (AVPCL_VER << 8) + AVPCL_SUB);
#endif
	}

	if (ident != NULL || firmware != NULL) {
		int result;
		PCI_GET_HANDLE_LOCK(hdl, pclGetVersion, (hdl, ident, firmware, dll), hPCI, card);

		if ((result = pclPciExec(hdl, CMD_GET_VERSION)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetVersion, result);
		if (ident    != NULL) *ident    = PCI_ReadDATA0(hPCI);
		if (firmware != NULL) *firmware = PCI_ReadDATA1(hPCI);
		UNLOCK_RETURN(hdl, pclGetVersion, PCL_OK);
	}

	APPEND_DEBUG_LOG(hdl, "< pclGetVersion(hdl:%d, ...) = %d", hdl, PCL_OK);
	return(PCL_OK);
}

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int _PCL_NDEF pclGetVerSpecialMulti(	// @66, $47, read special pcl firmware version
	int hdl,							// in:	handle from pclOpen()
	int _PCL_OUT *ident1,				// out:	identifier from byte $A
	int _PCL_OUT *ident2,				// out:	identifier from byte $B
	int _PCL_OUT *ident3				// out:	identifier from byte $C
) {
	PCI_GET_HANDLE_MULTI(hdl, pclGetVerSpecial, (), hPCI, card);

	int card = pgPciHdl->v[hdl].card;
	CHECK_CARD(card);
	if (ident1) *ident1 = pgPciCardShared->v[card].multi.pclGetVerSpecial.ident1;
	if (ident2) *ident2 = pgPciCardShared->v[card].multi.pclGetVerSpecial.ident2;
	if (ident3) *ident3 = pgPciCardShared->v[card].multi.pclGetVerSpecial.ident3;

	return(PCL_OK);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetVerSpecial(	// @66, $47, read special pcl firmware version
	int hdl,					// in:	handle from pclOpen()
	int _PCL_OUT *ident1,		// out:	identifier from byte $A
	int _PCL_OUT *ident2,		// out:	identifier from byte $B
	int _PCL_OUT *ident3		// out:	identifier from byte $C
) {
	int result;
	PCI_GET_HANDLE_LOCK(hdl, pclGetVerSpecial, (hdl, ident1, ident2, ident3), hPCI, card);

	if ((result = pclPciExec(hdl, CMD_GET_VER_SPECIAL)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetVerSpecial, result);

	if (ident1 != NULL) *ident1 = PCI_ReadDATA0(hPCI);
	if (ident2 != NULL) *ident2 = PCI_ReadDATA1(hPCI);
	if (ident3 != NULL) {
		int bytec = PCI_ReadDATA2(hPCI);
		if (bytec == PCL_ID_PCI_LV_D) {	// LV or D
			switch (gPciCard.v[card].wSubsystemDeviceID) {
			case AVPCL_SUBSYSTEM_DEVICE_ID_LV_5V:
			case AVPCL_SUBSYSTEM_DEVICE_ID_LV_3V:
				bytec = PCL_ID_PCI_LV;
				break;
			case AVPCL_SUBSYSTEM_DEVICE_ID_D_5V:
			case AVPCL_SUBSYSTEM_DEVICE_ID_D_3V:
				bytec = PCL_ID_PCI_D;
				break;
			}
		}
		*ident3 = bytec;
	}

	UNLOCK_RETURN(hdl, pclGetVerSpecial, PCL_OK);
}

//---------------------------------------------------------------------------

int regToImask(
	int reg						// in:	registers number like PCL_REG_xxx
) {
	switch (reg) {
	case PCL_REG_MIXED:		return(PCL_IMASK_MIXED);		/* mixed register */
	case PCL_REG_LTC:		return(PCL_IMASK_LTC);			/* ltc register */
	case PCL_REG_VITC:		return(PCL_IMASK_VITC);			/* vitc register */
	case PCL_REG_ATC_LTC:	return(PCL_IMASK_ATC_LTC);		/* atc ltc register */
	case PCL_REG_ATC_VITC:	return(PCL_IMASK_ATC_VITC);		/* atc vitc register */
	case PCL_REG_HANC_LTC:	return(PCL_IMASK_HANC_LTC);		/* hanc ltc register */
	case PCL_REG_HANC_VITC:	return(PCL_IMASK_HANC_VITC);	/* hanc vitc register */
	default:				assert(FALSE);	return(PCL_IMASK_NONE);
	}
}

//---------------------------------------------------------------------------

int imaskToReg(
	int imask					// in:	interrupt mask, like PCL_IMASK_xxx
) {
	switch (imask) {
	case PCL_IMASK_MIXED:		return(PCL_REG_MIXED);		/* mixed register */
	case PCL_IMASK_LTC:			return(PCL_REG_LTC);		/* ltc register */
	case PCL_IMASK_VITC:		return(PCL_REG_VITC);		/* vitc register */
	case PCL_IMASK_ATC_LTC:		return(PCL_REG_ATC_LTC);	/* atc ltc register */
	case PCL_IMASK_ATC_VITC:	return(PCL_REG_ATC_VITC);	/* atc vitc register */
	case PCL_IMASK_HANC_LTC:	return(PCL_REG_HANC_LTC);	/* hanc ltc register */
	case PCL_IMASK_HANC_VITC:	return(PCL_REG_HANC_VITC);	/* hanc vitc register */
	default:					assert(FALSE);	return(PCL_REG_NONE);
	}
}

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int enableRegMulti(
	int hdl,					// in:	handle from pclOpen()
	int reg						// in:	registers number like PCL_REG_xxx
) {
	int imask = regToImask(reg);
	int result = PCL_OK;

	assert(pgPciHdl);
	pgPciHdl->v[hdl].imask |= imask;

	if (pgPciHdl->v[hdl].multi.main) {
		// wir sind der main-multi-prozess
		int hdlMulti = pgPciHdl->v[hdl].multi.hdl;
		CHECK_HANDLE(hdlMulti);
		int card = pgPciHdl->v[hdlMulti].card;
		CHECK_CARD(card);
		int imask = getMultiImask(card);
		if (imask != pgPciHdl->v[hdlMulti].imask) {
			result = pclIntMask(hdlMulti, imask);
		}
	} else {
		// wir sind ein anderer prozess
	}
	return(result);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

TPciMultiCardTc *getMultiCardTc(int hdl, int imask)
{
	assert(pgPciHdl);
	int card = pgPciHdl->v[hdl].card;
	CHECK_CARD(card);

	switch (imask) {
	case PCL_IMASK_MIXED:		return(&pgPciCardShared->v[card].multi.tc.mixed);
	case PCL_IMASK_LTC:			return(&pgPciCardShared->v[card].multi.tc.ltc);
	case PCL_IMASK_VITC:		return(&pgPciCardShared->v[card].multi.tc.vitc);
	case PCL_IMASK_ATC_LTC:		return(&pgPciCardShared->v[card].multi.tc.atcLtc);
	case PCL_IMASK_ATC_VITC:	return(&pgPciCardShared->v[card].multi.tc.atcVitc);
	case PCL_IMASK_HANC_LTC:	return(&pgPciCardShared->v[card].multi.tc.hancLtc);
	case PCL_IMASK_HANC_VITC:	return(&pgPciCardShared->v[card].multi.tc.hancVitc);
	default:					return(NULL);
	}
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

TPciMultiHdlTc *getMultiHdlTc(int hdl, int imask)
{
	assert(pgPciHdl);
	switch (imask) {
	case PCL_IMASK_MIXED:		return(&pgPciHdl->v[hdl].multi.tc.mixed);
	case PCL_IMASK_LTC:			return(&pgPciHdl->v[hdl].multi.tc.ltc);
	case PCL_IMASK_VITC:		return(&pgPciHdl->v[hdl].multi.tc.vitc);
	case PCL_IMASK_ATC_LTC:		return(&pgPciHdl->v[hdl].multi.tc.atcLtc);
	case PCL_IMASK_ATC_VITC:	return(&pgPciHdl->v[hdl].multi.tc.atcVitc);
	case PCL_IMASK_HANC_LTC:	return(&pgPciHdl->v[hdl].multi.tc.hancLtc);
	case PCL_IMASK_HANC_VITC:	return(&pgPciHdl->v[hdl].multi.tc.hancVitc);
	default:					return(NULL);
	}
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int _PCL_NDEF pclGetTcMulti(	// @45, $41, read mixed-register timecode asynchronous
	int hdl,					// in:	handle from pclOpen()
	PCL_TC _PCL_OUT *time		// out:	timecode from mixed-register, bcd
) {
	PCI_GET_HANDLE_MULTI(hdl, pclGetTc, (hdl, time), hPCI, card);

	int result;
	if ((result = enableRegMulti(hdl, PCL_REG_MIXED)) != PCL_OK) return(result);

	int card = pgPciHdl->v[hdl].card;
	CHECK_CARD(card);

	TPciMultiCardTc *pMultiTc = &pgPciCardShared->v[card].multi.tc.mixed;

	assert(time);
	*time = pMultiTc->time;
	result = pMultiTc->result;

	return(result);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetTc(		// @45, $41, read mixed-register timecode asynchronous
	int hdl,				// in:	handle from pclOpen()
	PCL_TC _PCL_OUT *time	// out:	timecode from mixed-register, bcd
) {
	int result;
	CHECK_PTR(time);

	PCI_GET_HANDLE_LOCK(hdl, pclGetTc, (hdl, time), hPCI, card);

	if ((result = pclPciExec(hdl, CMD_GET_TC)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetTc, result);
	*((BYTE *) time    ) = (BYTE) PCI_ReadDATA0(hPCI);
	*((BYTE *) time + 1) = (BYTE) PCI_ReadDATA1(hPCI);
	*((BYTE *) time + 2) = (BYTE) PCI_ReadDATA2(hPCI);
	*((BYTE *) time + 3) = (BYTE) PCI_ReadDATA3(hPCI);

	UNLOCK_RETURN(hdl, pclGetTc, PCL_OK);
}

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int _PCL_NDEF pclGetUserMulti(	// @46, $42, read mixed-register user-bits asynchronous
	int hdl,					// in:	handle from pclOpen()
	PCL_USER _PCL_OUT *user		// out:	user-bits from mixed-register
) {
	PCI_GET_HANDLE_MULTI(hdl, pclGetUser, (hdl, user), hPCI, card);

	int result;
	if ((result = enableRegMulti(hdl, PCL_REG_MIXED)) != PCL_OK) return(result);

	int card = pgPciHdl->v[hdl].card;
	CHECK_CARD(card);

	TPciMultiCardTc *pMultiTc = &pgPciCardShared->v[card].multi.tc.mixed;

	assert(user);
	*user = pMultiTc->user;
	result = pMultiTc->result;

	return(result);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetUser(	// @46, $42, read mixed-register user-bits asynchronous
	int hdl,				// in:	handle from pclOpen()
	PCL_USER _PCL_OUT *user	// out:	user-bits from mixed-register
) {
	int result;
	CHECK_PTR(user);

	PCI_GET_HANDLE_LOCK(hdl, pclGetUser, (hdl, user), hPCI, card);

	if ((result = pclPciExec(hdl, CMD_GET_USER)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetUser, result);
	*((BYTE *) user    ) = (BYTE) PCI_ReadDATA0(hPCI);
	*((BYTE *) user + 1) = (BYTE) PCI_ReadDATA1(hPCI);
	*((BYTE *) user + 2) = (BYTE) PCI_ReadDATA2(hPCI);
	*((BYTE *) user + 3) = (BYTE) PCI_ReadDATA3(hPCI);

	UNLOCK_RETURN(hdl, pclGetUser, PCL_OK);
}

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int _PCL_NDEF pclGetRegisterMulti(	// @47, $43, read register asynchronous
	int hdl,						// in:	handle from pclOpen()
	int reg,						// in:	registers number like PCL_REG_xxx
	PCL_TC _PCL_OUT *time,			// out:	timecode from register, bcd
	int _PCL_OUT *bits,				// out:	misc. bits of register,
									//		meaning depends of '*newd',
									//		like PCL_BITS_xxx_xxx
	PCL_TC _PCL_OUT *user,			// out:	user-bits from register, bcd
	int _PCL_OUT *frame,			// out:	frame rate as detected by pcl,
									//		like PCL_FRAME_xxx
	int _PCL_OUT *newd				// out:	identifier of readed timecode,
									//		like PCL_NEWD_xxx
) {
	PCI_GET_HANDLE_MULTI(hdl, pclGetRegister, (hdl, reg, time, bits, user, frame, newd), hPCI, card);

	int result;
	if ((result = enableRegMulti(hdl, reg)) != PCL_OK) return(result);

	int hdlMulti = pgPciHdl->v[hdl].multi.hdl;
	CHECK_HANDLE(hdlMulti);

	int imask = regToImask(reg);
	if (imask == PCL_IMASK_NONE) {
		PCI_ERR("pclGetRegisterMulti(): Register 'reg' 0x%X is not available\n", reg);
		return(PCL_ERR_INV_REG);
	}
	TPciMultiCardTc *pMultiTc = getMultiCardTc(hdlMulti, imask);
	if (!pMultiTc) {
		PCI_ERR("pclGetRegisterMulti(): Register 'reg' 0x%X is not available\n", reg);
		return(PCL_ERR_INV_REG);
	}

	if (time)	*time  = pMultiTc->time;
	if (bits)	*bits  = pMultiTc->bits;
	if (user)	*user  = pMultiTc->user;
	if (frame)	*frame = pMultiTc->frame;
	if (newd)	*newd  = pMultiTc->newd;
	result = pMultiTc->result;

	return(result);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetRegister(	// @47, $43, read register asynchronous
	int hdl,					// in:	handle from pclOpen()
	int reg,					// in:	registers number like PCL_REG_xxx
	PCL_TC _PCL_OUT *time,		// out:	timecode from register, bcd
	int _PCL_OUT *bits,			// out:	misc. bits of register,
								//		meaning depends of '*newd',
								//		like PCL_BITS_xxx_xxx
	PCL_TC _PCL_OUT *user,		// out:	user-bits from register, bcd
	int _PCL_OUT *frame,		// out:	frame rate as detected by pcl,
								//		like PCL_FRAME_xxx
	int _PCL_OUT *newd			// out:	identifier of readed timecode,
								//		like PCL_NEWD_xxx
) {
	int result;

	switch (reg) {
	case PCL_REG_LTC:
	case PCL_REG_VITC:
	case PCL_REG_MIXED:
	case PCL_REG_ATC_LTC:
	case PCL_REG_ATC_VITC:
	case PCL_REG_ATC_VITC2:	// obsolete
	case PCL_REG_HANC_LTC:
	case PCL_REG_HANC_VITC:
		break;
	default:
		PCI_ERR("pclGetRegister(): Register 'reg' 0x%X is not available\n", reg);
		return(PCL_ERR_INV_REG);
		break;
	}

	if (
		reg   == PCL_REG_MIXED &&
		time  != NULL &&
		bits  == NULL &&
		user  == NULL &&
		frame == NULL &&
		newd  == NULL
	) return(pclGetTc(hdl, time));

	if (
		reg   == PCL_REG_MIXED &&
		time  == NULL &&
		bits  == NULL &&
		user  != NULL &&
		frame == NULL &&
		newd  == NULL
	) return(pclGetUser(hdl, user));

	PCI_GET_HANDLE_LOCK(hdl, pclGetRegister, (hdl, reg, time, bits, user, frame, newd), hPCI, card);

	PCI_WriteCMDD0(hPCI, reg);
	APPEND_DEBUG_LOG(hdl, "  pclGetRegister(hdl:%d, reg:0x%x)", hdl, reg);
	if ((result = pclPciExec(hdl, CMD_GET_REGISTER)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetRegister, result);
	if (time != NULL) {
		*((BYTE *) time    ) = (BYTE) PCI_ReadDATA0(hPCI);
		*((BYTE *) time + 1) = (BYTE) PCI_ReadDATA1(hPCI);
		*((BYTE *) time + 2) = (BYTE) PCI_ReadDATA2(hPCI);
		*((BYTE *) time + 3) = (BYTE) PCI_ReadDATA3(hPCI);
	}
	if (bits != NULL) *bits = PCI_ReadDATA4(hPCI);
	if (user != NULL) {
		*((BYTE *) user    ) = (BYTE) PCI_ReadDATA5(hPCI);
		*((BYTE *) user + 1) = (BYTE) PCI_ReadDATA6(hPCI);
		*((BYTE *) user + 2) = (BYTE) PCI_ReadDATA7(hPCI);
		*((BYTE *) user + 3) = (BYTE) PCI_ReadDATA8(hPCI);
	}
	if (frame != NULL) *frame = PCI_ReadDATA9(hPCI);
	if (newd  != NULL) *newd  = PCI_ReadDATAA(hPCI);
	APPEND_DEBUG_LOG(hdl, "  pclGetRegister(hdl:%d, reg:0x%x)", hdl, reg);
	UNLOCK_RETURN(hdl, pclGetRegister, PCL_OK);
}

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int _PCL_NDEF pclGetRegSyncMulti(	// @48, $43, read register synchronous
	int hdl,						// in:	handle from pclOpen()
	int reg,						// in:	registers number like PCL_REG_xxx
	PCL_TC _PCL_OUT *time,			// out:	timecode from register, bcd
	int _PCL_OUT *bits,				// out:	misc. bits of register,
									//		meaning depends of '*newd',
									//		like PCL_BITS_xxx_xxx
	PCL_TC _PCL_OUT *user,			// out:	user-bits from register, bcd
	int _PCL_OUT *frame,			// out:	frame rate as detected by pcl,
									//		like PCL_FRAME_xxx
	int _PCL_OUT *newd				// out:	identifier of readed timecode,
									//		like PCL_NEWD_xxx
) {
	PCI_GET_HANDLE_MULTI(hdl, pclGetRegSync, (hdl, reg, time, bits, user, frame, newd), hPCI, card);

	int result;
	if ((result = enableRegMulti(hdl, reg)) != PCL_OK) return(result);
	result = PCL_NO_NEWD;

	int hdlMain = pgPciHdl->v[hdl].multi.hdl;

	int imask = regToImask(reg);
	if (imask == PCL_IMASK_NONE) {
		PCI_ERR("pclGetRegSyncMulti(): Register 'reg' 0x%X is not available (imask)\n", reg);
		return(PCL_ERR_INV_REG);
	}
/*
	TPciMultiCardTc *pMultiTc = getMultiCardTc(hdl, imask);
	if (!pMultiTc) {
		PCI_ERR("pclGetRegSyncMulti(): Register 'reg' 0x%X is not available (pMultiTc)\n", reg);
		return(PCL_ERR_INV_REG);
	}
*/
	TPciMultiCardTc *pMultiTcMain = getMultiCardTc(hdlMain, imask);
	if (!pMultiTcMain) {
		PCI_ERR("pclGetRegSyncMulti(): Register 'reg' 0x%X is not available (pMultiTcMain)\n", reg);
		return(PCL_ERR_INV_REG);
	}
	TPciMultiHdlTc *pMultiHdlTc = getMultiHdlTc(hdl, imask);
	if (!pMultiHdlTc) {
		PCI_ERR("pclGetRegSyncMulti(): Register 'reg' 0x%X is not available (pMultiHdlTc)\n", reg);
		return(PCL_ERR_INV_REG);
	}

	if (pMultiHdlTc->counter != pMultiTcMain->counter) {
		result = pclGetRegisterMulti(hdl, reg, time, bits, user, frame, newd);
		pMultiHdlTc->counter  = pMultiTcMain->counter;
	} else {
		assert(result == PCL_NO_NEWD);
	}
	return(result);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetRegSync(	// @48, $43, read register synchronous
	int hdl,					// in:	handle from pclOpen()
	int reg,					// in:	registers number like PCL_REG_xxx
	PCL_TC _PCL_OUT *time,		// out:	timecode from register, bcd
	int _PCL_OUT *bits,			// out:	misc. bits of register,
								//		meaning depends of '*newd',
								//		like PCL_BITS_xxx_xxx
	PCL_TC _PCL_OUT *user,		// out:	user-bits from register, bcd
	int _PCL_OUT *frame,		// out:	frame rate as detected by pcl,
								//		like PCL_FRAME_xxx
	int _PCL_OUT *newd			// out:	identifier of readed timecode,
								//		like PCL_NEWD_xxx
) {
	PCI_GET_HANDLE(hdl, pclGetRegSync, (hdl, reg, time, bits, user, frame, newd), hPCI, card);
	int intflg = PCI_ReadINTFLGL(hPCI);
	if (gPciCard.v[card].irq16) intflg |= PCI_ReadINTFLGH(hPCI) << 8;

	switch (reg) {
	case PCL_REG_LTC:		if ((intflg & PCL_IMASK_LTC)       == 0) return(PCL_NO_NEWD);  break;
	case PCL_REG_VITC:		if ((intflg & PCL_IMASK_VITC)      == 0) return(PCL_NO_NEWD);  break;
	case PCL_REG_MIXED:		if ((intflg & PCL_IMASK_MIXED)     == 0) return(PCL_NO_NEWD);  break;
	case PCL_REG_ATC_LTC:	if ((intflg & PCL_IMASK_ATC_LTC)   == 0) return(PCL_NO_NEWD);  break;
	case PCL_REG_ATC_VITC:	if ((intflg & PCL_IMASK_ATC_VITC)  == 0) return(PCL_NO_NEWD);  break;
	case PCL_REG_ATC_VITC2:	if ((intflg & PCL_IMASK_ATC_VITC2) == 0) return(PCL_NO_NEWD);  break;	// obsolete
	case PCL_REG_HANC_LTC:	if ((intflg & PCL_IMASK_HANC_LTC)  == 0) return(PCL_NO_NEWD);  break;
	case PCL_REG_HANC_VITC:	if ((intflg & PCL_IMASK_HANC_VITC) == 0) return(PCL_NO_NEWD);  break;
		break;

	default:
		PCI_ERR("pclGetRegSync(): Register 'reg' 0x%X is not available\n", reg);
		return(PCL_ERR_INV_REG);
		break;
	}
	return(pclGetRegister(hdl, reg, time, bits, user, frame, newd));
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetRegStatus(	/* @84,	$44, read register status */
	int hdl,					/* in:	handle from pclOpen() */
	int reg,					/* in:	registers number like PCL_REG_xxx */
	int _PCL_OUT *stat			/* out:	status bytes, like PCI_STAT_xxx */
) {
	int result;
	CHECK_PTR(stat);

#	define pclGetRegStatusMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetRegStatus, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, reg);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", reg);
	if ((result = pclPciExec(hdl, CMD_GET_REG_STATUS)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetRegStatus, result);
	*((BYTE *) stat    ) = (BYTE) PCI_ReadDATA0(hPCI);
	*((BYTE *) stat + 1) = (BYTE) PCI_ReadDATA1(hPCI);
	*((BYTE *) stat + 2) = 0;
	*((BYTE *) stat + 3) = 0;

	UNLOCK_RETURN(hdl, pclGetRegStatus, PCL_OK);
}

//---------------------------------------------------------------------------

#if defined(_Windows) || defined(_WIN32) || defined(LINUX)

int _PCL_FDEF pclIntMask(	// $03, interrupts: set hw interrupt mask
	int hdl,				// in:	handle from pclOpen()
	int imask				// in:	interrupt mask, like PCL_IMASK_xxx
) {
	int result;

	CHECK_RANGE(imask, 0, 0xFFFF, PCL_ERR_RANGE);
#	define pclIntMaskMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclIntMask, (), hPCI, card);
	APPEND_DEBUG_LOG(hdl, "  pclIntMask(hdl:%d, imask:0x%04X)", hdl, imask);

	pgPciHdl->v[hdl].imask = imask;
#ifdef PCL_OPT_ANC				/* read user defined ancillary data */
	if (imask & PCL_IMASK_ANC) {
		pgPciHdl->v[hdl].anc.imask = PCL_IMASK_V1 | PCL_IMASK_V2;
	} else {
		pgPciHdl->v[hdl].anc.imask = 0;
	}
#endif
	int cmask = getPciImask(card);
	APPEND_DEBUG_LOG(hdl, "  pclIntMask() hdl:%d cmask:0x%04X", hdl, cmask);

	PCI_WriteCMDD0(hPCI, cmask);
	if (gPciCard.v[card].irq16) PCI_WriteCMDD1(hPCI, cmask >> 8);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %04X", cmask);
	if ((result = pclPciExec(hdl, CMD_INT_MASK)) != PCL_OK) UNLOCK_RETURN(hdl, pclIntMask, result);

	gPciCard.v[card].imask = PCI_ReadDATA0(hPCI);
	if (gPciCard.v[card].irq16) gPciCard.v[card].imask |= PCI_ReadDATA1(hPCI) << 8;
	APPEND_DEBUG_LOG(hdl, "  pclIntMask() hdl:%d imask:0x%04X", hdl, gPciCard.v[card].imask);

	pclPciUnlock(hdl);
	result = pclIntEnable(hdl, gPciCard.v[card].imask != 0);
	APPEND_DEBUG_LOG(hdl, "< pclIntMask() hdl:%d = %d", hdl, result);
	return(result);
}

#endif

//---------------------------------------------------------------------------

#if defined(_Windows) || defined(_WIN32) || defined(LINUX)

int _PCL_FDEF pclSetIntMask(	// $04, set interrupt bits
	int hdl,					// in:	handle from pclOpen()
	int imask					// in:	interrupt mask, like PCL_IMASK_xxx
) {
	int result;

	CHECK_RANGE(imask, 0, 0xFFFF, PCL_ERR_RANGE);
#	define pclSetIntMaskMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclSetIntMask, (), hPCI, card);
	APPEND_DEBUG_LOG(hdl, "  pclSetIntMask(hdl:%d, imask:0x%04X)", hdl, imask);

	pgPciHdl->v[hdl].imask |= imask;
#ifdef PCL_OPT_ANC				/* read user defined ancillary data */
	if (imask & PCL_IMASK_ANC) {
		pgPciHdl->v[hdl].anc.imask = PCL_IMASK_V1 | PCL_IMASK_V2;
		imask |= PCL_IMASK_V1 | PCL_IMASK_V2;
	}
#endif
	int cmask = getPciImask(card) & imask;
	APPEND_DEBUG_LOG(hdl, "  pclSetIntMask() hdl:%d cmask:0x%04X", hdl, cmask);

	PCI_WriteCMDD0(hPCI, cmask);
	if (gPciCard.v[card].irq16) PCI_WriteCMDD1(hPCI, cmask >> 8);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X");
	if ((result = pclPciExec(hdl, CMD_SET_INT_MASK)) != PCL_OK) UNLOCK_RETURN(hdl, pclSetIntMask, result);

	gPciCard.v[card].imask = PCI_ReadDATA0(hPCI);
	if (gPciCard.v[card].irq16) gPciCard.v[card].imask |= PCI_ReadDATA1(hPCI) << 8;
	APPEND_DEBUG_LOG(hdl, "  pclSetIntMask() hdl:%d imask:0x%04X", hdl, gPciCard.v[card].imask);

	pclPciUnlock(hdl);
	result = pclIntEnable(hdl, gPciCard.v[card].imask != 0);
	APPEND_DEBUG_LOG(hdl, "< pclSetIntMask() hdl:%d = %d", hdl, result);
	return(result);
}

#endif

//---------------------------------------------------------------------------

#if defined(_Windows) || defined(_WIN32) || defined(LINUX)

int _PCL_FDEF pclResetIntMask(	// $05, reset interrupt bits
	int hdl,					// in:	handle from pclOpen()
	int imask					// in:	interrupt mask, like PCL_IMASK_xxx
) {
	int result;

	CHECK_RANGE(imask, 0, 0xFFFF, PCL_ERR_RANGE);
#	define pclResetIntMaskMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclResetIntMask, (), hPCI, card);
	APPEND_DEBUG_LOG(hdl, "  pclResetIntMask(hdl:%d, imask:0x%04X)", hdl, imask);

	pgPciHdl->v[hdl].imask &= ~imask;
#ifdef PCL_OPT_ANC				/* read user defined ancillary data */
	if (imask & PCL_IMASK_ANC) {
		pgPciHdl->v[hdl].anc.imask = 0;
		imask |= ~pgPciHdl->v[hdl].imask & (PCL_IMASK_V1 | PCL_IMASK_V2);
	}
#endif
	int cmask = ~getPciImask(card) & imask;
	APPEND_DEBUG_LOG(hdl, "  pclResetIntMask() hdl:%d cmask:0x%04X", hdl, cmask);

	PCI_WriteCMDD0(hPCI, cmask);
	if (gPciCard.v[card].irq16) PCI_WriteCMDD1(hPCI, cmask >> 8);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %04X", cmask);
	if ((result = pclPciExec(hdl, CMD_RESET_INT_MASK)) != PCL_OK) UNLOCK_RETURN(hdl, pclResetIntMask, result);

	gPciCard.v[card].imask = PCI_ReadDATA0(hPCI);
	if (gPciCard.v[card].irq16) gPciCard.v[card].imask |= PCI_ReadDATA1(hPCI) << 8;
	APPEND_DEBUG_LOG(hdl, "  pclResetIntMask() hdl:%d imask:0x%04X", hdl, gPciCard.v[card].imask);

	pclPciUnlock(hdl);
	result = pclIntEnable(hdl, gPciCard.v[card].imask != 0);
	APPEND_DEBUG_LOG(hdl, "< pclResetIntMask() hdl:%d = %d", hdl, result);
	return(result);
}

#endif

//---------------------------------------------------------------------------

#if defined(_Windows) || defined(_WIN32) || defined(LINUX)

int _PCL_FDEF pclSetIntHandler(	// @69,	set interrupt handler
	int hdl,					// in:	handle from pclOpen()
	PCL_INT_HANDLER intHandler	// in:	address of interrupt handler
) {
#	define pclSetIntHandlerMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclSetIntHandler, (), hPCI, card);

	(void)hPCI;

	pgPciHdl->v[hdl].intHandler.f = intHandler;

	UNLOCK_RETURN(hdl, pclSetIntHandler, PCL_OK);
}

#endif

//---------------------------------------------------------------------------

#if defined(_Windows) || defined(_WIN32) || defined(LINUX)

int _PCL_FDEF pclIntAck(	// @70,	$48, interrupt acknowledge
	int hdl,				// in:	handle from pclOpen()
	int imask				// in:	interrupt mask, like PCL_IMASK_xxx
) {
	int result;
#	define pclIntAckMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclIntAck, (), hPCI, card);
	APPEND_DEBUG_LOG(hdl, "  pclIntAck(hdl:%d, imask:0x%04x)", hdl, imask);

	PCI_WriteCMDD0(hPCI, imask);
	if (gPciCard.v[card].irq16) PCI_WriteCMDD1(hPCI, imask >> 8);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %04X", imask);
	if ((result = pclPciExec(hdl, CMD_INT_ACK)) != PCL_OK) UNLOCK_RETURN(hdl, pclIntAck, result);

	UNLOCK_RETURN(hdl, pclIntAck, PCL_OK);
}

#endif

//---------------------------------------------------------------------------

#if defined(_Windows) || defined(_WIN32)

int _PCL_FDEF pclSetIntPriority(	// @71, set interrupt priority
   int hdl,							// in:	handle from pclOpen()
   int priority						// in:	priority of interrupt handler
) {
#	define pclSetIntPriorityMulti pclMultiNotSupported
	PCI_GET_HANDLE(hdl, pclSetIntPriority, (), hPCI, card);

	if (!PCI_IntIsEnabled(hPCI)) {
		PCI_ERR("pclSetIntPriority(): Interrupt thread is not running\n");
		return(PCL_ERR_IOCTL);
	}

	if (!PCI_IntPriority(hPCI, priority)) {
		PCI_ERR("pclSetIntPriority(): Can't set priority of interrupt thread\n");
		return(PCL_ERR_IOCTL);
	}

	return(PCL_OK);
}

#endif

//---------------------------------------------------------------------------

#if defined(_Windows) || defined(_WIN32)

int _PCL_FDEF pclGetThreadHandle(	// @90, get thread handle
   int hdl,							// in:	handle from pclOpen()
   int thread,						// in:	thread to use, like PCL_THREAD_xxx
   HANDLE *phHandle					// out:	handle of thread
) {
	CHECK_PTR(phHandle);
#	define pclGetThreadHandleMulti pclMultiNotSupported
	PCI_GET_HANDLE(hdl, pclGetThreadHandle, (), hPCI, card);
	HANDLE hThread;

	switch (thread) {
	case PCL_THREAD_INT:
		hThread = PCL_GetIntHandleThread(hPCI);
		break;
	case PCL_THREAD_EVENT:
		hThread = PCL_GetEventHandleThread(hPCI);
		break;
	default:
		PCI_ERR("pclGetThreadHandle(): thread parameter out of range\n");
		return(PCL_ERR_RANGE);
	}

	if (!hThread) {
		PCI_ERR("pclGetThreadHandle(): Can't get thread handle\n");
		return(PCL_ERR_IOCTL);
	}
	*phHandle = hThread;
	return(PCL_OK);
}

#endif

//---------------------------------------------------------------------------

int _PCL_FDEF pclReaderEnable(	// $10, reader on/off
	int hdl,					// in:	handle from pclOpen()
	int rmask					// in:	reader mask, like PCL_RMASK_xxx
) {
	int result;
#	define pclReaderEnableMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclReaderEnable, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, rmask);
	PCI_WriteCMDD1(hPCI, rmask >> 8);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %04X", rmask);
	if ((result = pclPciExec(hdl, CMD_READER_ENABLE)) != PCL_OK) UNLOCK_RETURN(hdl, pclReaderEnable, result);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.readerEnable.use = TRUE;
	gPciCard.v[card].persist.readerEnable.rmask = rmask;
#endif
	UNLOCK_RETURN(hdl, pclReaderEnable, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclVitcSetup(	// @65, $14, vitc setup (pcl5/6)
	int hdl,				// in:	handle from pclOpen()
	int line1,				// in:	1st line, hexadecimal
	int line2,				// in:	2nd line, hexadecimal
	int vline				// in:	vitc line mode, like PCL_VLINE_xxx
) {
	int result;
	BYTE err;
#	define pclVitcSetupMulti pclMultiIgnored
	PCI_GET_HANDLE_LOCK(hdl, pclVitcSetup, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, line1);
	PCI_WriteCMDD1(hPCI, line2);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X %02X", line2, line1);
	if ((result = pclPciExec(hdl, CMD_VITC_SETUP_1)) != PCL_OK) UNLOCK_RETURN(hdl, pclVitcSetup, result);
	if ((err = PCI_ReadDATA0(hPCI)) != 0) {
		PCI_ERR("pclVitcSetup(): VITC lines out of range, reason: 0x%X\n", err);
		UNLOCK_RETURN(hdl, pclVitcSetup, PCL_ERR_RANGE);
	}
	PCI_WriteCMDD0(hPCI, vline);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", vline);
	if ((result = pclPciExec(hdl, CMD_VITC_SETUP_2)) != PCL_OK) UNLOCK_RETURN(hdl, pclVitcSetup, result);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.vitcSetup.use = TRUE;
	gPciCard.v[card].persist.vitcSetup.line1 = line1;
	gPciCard.v[card].persist.vitcSetup.line2 = line2;
	gPciCard.v[card].persist.vitcSetup.vline = vline;
#endif
	UNLOCK_RETURN(hdl, pclVitcSetup, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclMixedEnable(	// $20, mixed register: reader on/off
	int hdl,					// in:	handle from pclOpen()
	int rmask					// in:	reader mask, like PCL_RMASK_xxx
) {
	int result;
#	define pclMixedEnableMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclMixedEnable, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, rmask);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", rmask);
	if ((result = pclPciExec(hdl, CMD_MIXED_ENABLE)) != PCL_OK) UNLOCK_RETURN(hdl, pclMixedEnable, result);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.mixedEnable.use = TRUE;
	gPciCard.v[card].persist.mixedEnable.rmask = rmask;
#endif
	UNLOCK_RETURN(hdl, pclMixedEnable, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclPriority(	// $21, priority: mixed reg. tc priority
	int hdl,				// in:	handle from pclOpen()
	int prior				// in:	priority, like PCL_PRIOR_xxx
) {
	int result;
#	define pclPriorityMulti pclMultiIgnored
	PCI_GET_HANDLE_LOCK(hdl, pclPriority, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, prior);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", prior);
	if ((result = pclPciExec(hdl, CMD_PRIORITY)) != PCL_OK) UNLOCK_RETURN(hdl, pclPriority, result);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.priority.use = TRUE;
	gPciCard.v[card].persist.priority.prior = prior;
#endif
	UNLOCK_RETURN(hdl, pclPriority, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclTcFrames(	// $30, tc frame rate
	int hdl,				// in:	handle from pclOpen()
	int tcfr				// in:	frame rate, like PCL_TCFR_xxx
) {
	int result;
#	define pclTcFramesMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclTcFrames, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, tcfr);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", tcfr);
	if ((result = pclPciExec(hdl, CMD_TC_FRAMES)) != PCL_OK) UNLOCK_RETURN(hdl, pclTcFrames, result);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.tcFrames.use = TRUE;
	gPciCard.v[card].persist.tcFrames.tcfr = tcfr;
#endif
	UNLOCK_RETURN(hdl, pclTcFrames, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclLtcFrames(	// $31, set ltc frame rate
	int hdl,				// in:	handle from pclOpen()
	int ltcfr				// in:	ltc frame rate, like PCL_TCFR_xxx
) {
	int result;
#	define pclLtcFramesMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclLtcFrames, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, ltcfr);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", ltcfr);
	if ((result = pclPciExec(hdl, CMD_LTC_FRAMES)) != PCL_OK) UNLOCK_RETURN(hdl, pclLtcFrames, result);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.ltcFrames.use = TRUE;
	gPciCard.v[card].persist.ltcFrames.ltcfr = ltcfr;
#endif
	UNLOCK_RETURN(hdl, pclLtcFrames, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclVitcFrames(	// $32, set vitc frame rate
	int hdl,					// in:	handle from pclOpen()
	int vitcfr					// in:	vitc frame rate, like PCL_TCFR_xxx
) {
	int result;
#	define pclVitcFramesMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclVitcFrames, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, vitcfr);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", vitcfr);
	if ((result = pclPciExec(hdl, CMD_VITC_FRAMES)) != PCL_OK) UNLOCK_RETURN(hdl, pclVitcFrames, result);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.vitcFrames.use = TRUE;
	gPciCard.v[card].persist.vitcFrames.vitcfr = vitcfr;
#endif
	UNLOCK_RETURN(hdl, pclVitcFrames, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclVitcLevelControl(	// $16, vitc level control
	int hdl,						// in:	handle from pclOpen()
	int vlmode,						// in:	vitc level control mode,
									//		like PCL_VLMODE_xxx
	int vlev,						// in:	vitc level, binary,
									//		like PCL_VLEX_xxx
	int *vlret						// out:	vitc level return, binary,
									//		like PCL_VLEX_xxx
) {
	int result;
#	define pclVitcLevelControlMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclVitcLevelControl, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, vlmode);
	PCI_WriteCMDD1(hPCI, vlev);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X %02X", vlev, vlmode);
	if ((result = pclPciExec(hdl, CMD_VITC_LEVEL)) != PCL_OK) UNLOCK_RETURN(hdl, pclVitcLevelControl, result);
	if (vlret != NULL) *vlret = PCI_ReadDATA0(hPCI);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.vitcLevelControl.use = TRUE;
	gPciCard.v[card].persist.vitcLevelControl.vlmode = vlmode;
	gPciCard.v[card].persist.vitcLevelControl.vlev = vlev;
#endif
	UNLOCK_RETURN(hdl, pclVitcLevelControl, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclLtcError(	// $17, get ltc error counter
	int hdl,				// in:	handle from pclOpen()
	int ecmode,				// in:	error counter mode, like PCL_ECMODE_xxx
	int *ecval				// out:	error counter, 8 bit binary
) {
	int result;
#	define pclLtcErrorMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclLtcError, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, ecmode);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", ecmode);
	if ((result = pclPciExec(hdl, CMD_LTC_ERROR)) != PCL_OK) UNLOCK_RETURN(hdl, pclLtcError, result);
	if (ecval != NULL) *ecval = PCI_ReadDATA0(hPCI);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.ltcError.use = TRUE;
	gPciCard.v[card].persist.ltcError.ecmode = ecmode;
#endif
	UNLOCK_RETURN(hdl, pclLtcError, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclVitcError(	// $18, get vitc error counter
	int hdl,				// in:	handle from pclOpen()
	int ecmode,				// in:	error counter mode, like PCL_EMODE_xxx
	int *ecval				// out:	error counter, 8 bit binary
) {
	int result;
#	define pclVitcErrorMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclVitcError, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, ecmode);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", ecmode);
	if ((result = pclPciExec(hdl, CMD_VITC_ERROR)) != PCL_OK) UNLOCK_RETURN(hdl, pclVitcError, result);
	if (ecval != NULL) *ecval = PCI_ReadDATA0(hPCI);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.vitcError.use = TRUE;
	gPciCard.v[card].persist.vitcError.ecmode = ecmode;
#endif
	UNLOCK_RETURN(hdl, pclVitcError, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclFlyWheel(	// $19, set flywheel mode
	int hdl,				// in:	handle from pclOpen()
	int fwmode				// in:	fly wheel mode, like PCL_FWMODE_xxx
) {
	int result;
#	define pclFlyWheelMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclFlyWheel, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, fwmode);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", fwmode);
	if ((result = pclPciExec(hdl, CMD_FLY_WHEEL)) != PCL_OK) UNLOCK_RETURN(hdl, pclFlyWheel, result);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.flyWheel.use = TRUE;
	gPciCard.v[card].persist.flyWheel.fwmode = fwmode;
#endif
	UNLOCK_RETURN(hdl, pclFlyWheel, PCL_OK);
}

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclAtcLtcError(	/* @85,	$22, get atc ltc error counter */
	int hdl,					/* in:	handle from pclOpen() */
	int ecmode,					/* in:	error counter mode, like PCL_ECMODE_xxx */
	int *ecval					/* out:	error counter, 8 bit binary */
) {
	int result;
#	define pclAtcLtcErrorMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclAtcLtcError, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, ecmode);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", ecmode);
	if ((result = pclPciExec(hdl, CMD_ATC_LTC_ERROR)) != PCL_OK) UNLOCK_RETURN(hdl, pclAtcLtcError, result);
	if (ecval != NULL) *ecval = PCI_ReadDATA0(hPCI);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.atcLtcError.use = TRUE;
	gPciCard.v[card].persist.atcLtcError.ecmode = ecmode;
#endif
	UNLOCK_RETURN(hdl, pclAtcLtcError, PCL_OK);
}

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclAtcVitcError(	/* @86,	$23, get atc vitc1 error counter */
	int hdl,					/* in:	handle from pclOpen() */
	int ecmode,					/* in:	error counter mode, like PCL_EMODE_xxx */
	int *ecval					/* out:	error counter, 8 bit binary */
) {
	int result;
#	define pclAtcVitcErrorMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclAtcVitcError, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, ecmode);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", ecmode);
	if ((result = pclPciExec(hdl, CMD_ATC_VITC_ERROR)) != PCL_OK) UNLOCK_RETURN(hdl, pclAtcVitc1Error, result);
	if (ecval != NULL) *ecval = PCI_ReadDATA0(hPCI);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.atcVitc1Error.use = TRUE;
	gPciCard.v[card].persist.atcVitc1Error.ecmode = ecmode;
#endif
	UNLOCK_RETURN(hdl, pclAtcVitc1Error, PCL_OK);
}

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclAtcVitc2Error(	/* @87,	$24, get atc vitc2 error counter */
	int hdl,					/* in:	handle from pclOpen() */
	int ecmode,					/* in:	error counter mode, like PCL_EMODE_xxx */
	int *ecval					/* out:	error counter, 8 bit binary */
) {
	int result;
#	define pclAtcVitc2ErrorMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclAtcVitc2Error, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, ecmode);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", ecmode);
	if ((result = pclPciExec(hdl, CMD_ATC_VITC2_ERROR)) != PCL_OK) UNLOCK_RETURN(hdl, pclAtcVitc2Error, result);
	if (ecval != NULL) *ecval = PCI_ReadDATA0(hPCI);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.atcVitc2Error.use = TRUE;
	gPciCard.v[card].persist.atcVitc2Error.ecmode = ecmode;
#endif
	UNLOCK_RETURN(hdl, pclAtcVitc2Error, PCL_OK);
}

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclHancLtcError(	/* @88,	$25, get hanc ltc error counter */
	int hdl,					/* in:	handle from pclOpen() */
	int ecmode,					/* in:	error counter mode, like PCL_ECMODE_xxx */
	int *ecval					/* out:	error counter, 8 bit binary */
) {
	int result;
#	define pclHancLtcErrorMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclHancLtcError, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, ecmode);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", ecmode);
	if ((result = pclPciExec(hdl, CMD_HANC_LTC_ERROR)) != PCL_OK) UNLOCK_RETURN(hdl, pclHancLtcError, result);
	if (ecval != NULL) *ecval = PCI_ReadDATA0(hPCI);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.hancLtcError.use = TRUE;
	gPciCard.v[card].persist.hancLtcError.ecmode = ecmode;
#endif
	UNLOCK_RETURN(hdl, pclHancLtcError, PCL_OK);
}

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclHancVitcError(	/* @89,	$26, get hanc vitc error counter */
	int hdl,					/* in:	handle from pclOpen() */
	int ecmode,					/* in:	error counter mode, like PCL_EMODE_xxx */
	int *ecval					/* out:	error counter, 8 bit binary */
) {
	int result;
#	define pclHancVitcErrorMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclHancVitcError, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, ecmode);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", ecmode);
	if ((result = pclPciExec(hdl, CMD_HANC_VITC_ERROR)) != PCL_OK) UNLOCK_RETURN(hdl, pclHancVitcError, result);
	if (ecval != NULL) *ecval = PCI_ReadDATA0(hPCI);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.hancVitcError.use = TRUE;
	gPciCard.v[card].persist.hancVitcError.ecmode = ecmode;
#endif
	UNLOCK_RETURN(hdl, pclHancVitcError, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetVideoStandard(	/* @91,	$4B, get video standard */
	int hdl,						/* in:	handle from pclOpen() */
	int _PCL_OUT *system,			/* out:	video system, like PCL_VSYSTEM_xxx */
	int _PCL_OUT *progressive		/* out:	0 = interlaced, else progressive video mode */
) {
	int result;
#	define pclGetVideoStandardMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetVideoStandard, (), hPCI, card);

	if ((result = pclPciExec(hdl, CMD_GET_VIDEO_STANDARD)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetVideoStandard, result);
	if (system != NULL) *system = PCI_ReadDATA0(hPCI);
	if (progressive != NULL) *progressive = PCI_ReadDATA1(hPCI);

	UNLOCK_RETURN(hdl, pclGetVideoStandard, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetPayloadIdentifier(	/* @105, $50, get payload identifier (SMPTE 352M) */
	int hdl,							/* in:	handle from pclOpen() */
	BYTE _PCL_OUT *pbPayloadIdentifier,	/* out:	payload identifier, 4 bytes */
	int _PCL_OUT *flags					/* out:	S352M flags, like PCL_S352M_xxx */
) {
	int result;
#	define pclGetPayloadIdentifierMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetPayloadIdentifier, (), hPCI, card);

	if ((result = pclPciExec(hdl, CMD_GET_PAYLOAD_IDENTIFIER)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetPayloadIdentifier, result);
	if (pbPayloadIdentifier) {
		pbPayloadIdentifier[0] = PCI_ReadDATA0(hPCI);
		pbPayloadIdentifier[1] = PCI_ReadDATA1(hPCI);
		pbPayloadIdentifier[2] = PCI_ReadDATA2(hPCI);
		pbPayloadIdentifier[3] = PCI_ReadDATA3(hPCI);
	}
	if (flags) {
		*flags = PCI_ReadDATA4(hPCI);
	}

	UNLOCK_RETURN(hdl, pclGetPayloadIdentifier, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclSetUsePayloadIdentifier(	/* @106, $51, get payload identifier (SMPTE 352M) */
	int hdl,								/* in:	handle from pclOpen() */
	BYTE _PCL_IN flags						/* in:	S352M flags, like PCL_SET_xxx */
) {
	int result;
#	define pclSetUsePayloadIdentifierMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclSetUsePayloadIdentifier, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, flags);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", flags);
	if ((result = pclPciExec(hdl, CMD_SET_USE_PAYLOAD_IDENTIFIER)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetPayloadIdentifier, result);

	UNLOCK_RETURN(hdl, pclSetUsePayloadIdentifier, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclSetDID(	/* @92,	$49, set DID of anc register */
	int hdl,				/* in:	handle from pclOpen() */
	int _PCL_IN ancreg,		/* in:	anc register, like PCL_ANCREG_xxx */
	int _PCL_IN did			/* in:	DID */
) {
	int result;
#	define pclSetDIDMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclSetDID, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, ancreg);
	PCI_WriteCMDD1(hPCI, did);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X %02X", did, ancreg);
	if ((result = pclPciExec(hdl, CMD_SET_DID)) != PCL_OK) UNLOCK_RETURN(hdl, pclSetDID, result);

	UNLOCK_RETURN(hdl, pclSetDID, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclSetSDID(	/* @93,	$4a, set SDID of anc register */
	int hdl,				/* in:	handle from pclOpen() */
	int _PCL_IN ancreg,		/* in:	anc register, like PCL_ANCREG_xxx */
	int _PCL_IN sdid		/* in:	SDID */
) {
	int result;
#	define pclSetSDIDMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclSetSDID, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, ancreg);
	PCI_WriteCMDD1(hPCI, sdid);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X %02X", sdid, ancreg);
	if ((result = pclPciExec(hdl, CMD_SET_SDID)) != PCL_OK) UNLOCK_RETURN(hdl, pclSetSDID, result);

	UNLOCK_RETURN(hdl, pclSetSDID, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetVideoRaster(	/* @94,	$4C, get video raster informations */
	int hdl,						/* in:	handle from pclOpen() */
	int _PCL_OUT *activeWords,		/* out:	active words / line */
	int _PCL_OUT *totalWords,		/* out:	total words / line */
	int _PCL_OUT *activeLines,		/* out:	active lines / frame */
	int _PCL_OUT *totalLines		/* out:	total lines / field */
) {
	int result;
#	define pclGetVideoRasterMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetVideoRaster, (), hPCI, card);

	if ((result = pclPciExec(hdl, CMD_GET_VIDEO_RASTER)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetVideoRaster, result);
	if (activeWords != NULL) *activeWords = ((WORD) PCI_ReadDATA1(hPCI) << 8) | PCI_ReadDATA0(hPCI);
	if (totalWords  != NULL) *totalWords  = ((WORD) PCI_ReadDATA3(hPCI) << 8) | PCI_ReadDATA2(hPCI);
	if (activeLines != NULL) *activeLines = ((WORD) PCI_ReadDATA5(hPCI) << 8) | PCI_ReadDATA4(hPCI);
	if (totalLines  != NULL) *totalLines  = ((WORD) PCI_ReadDATA7(hPCI) << 8) | PCI_ReadDATA6(hPCI);

	UNLOCK_RETURN(hdl, pclGetVideoRaster, PCL_OK);
}

/*-------------------------------------------------------------------------*/

#ifdef PCL_MULTI_MODE

int _PCL_FDEF pclGetMultiModeMulti(	/* @101, get multi operating mode */
	int hdl,						/* in:	handle from pclOpen() */
	int _PCL_OUT *multi				/* out:	multi operating mode, like PCL_MULTI_MODE_xxx */
) {
	PCI_GET_HANDLE_MULTI(hdl, pclGetMultiMode, (hdl, multi), hPCI, card);

	if (pgPciHdl->v[hdl].multi.main) {
		*multi = PCL_MULTI_MODE_MASTER;
	} else {
		*multi = PCL_MULTI_MODE_SLAVE;
	}
	return(PCL_OK);
}

#endif // PCL_MULTI_MODE

/*-------------------------------------------------------------------------*/

#ifdef PCL_MULTI_MODE

int _PCL_FDEF pclGetMultiMode(	/* @101, get multi operating mode */
	int hdl,					/* in:	handle from pclOpen() */
	int _PCL_OUT *multi			/* out:	multi operating mode, like PCL_MULTI_MODE_xxx */
) {
	CHECK_PTR(multi);
	PCI_GET_HANDLE(hdl, pclGetMultiMode, (hdl, multi), hPCI, card);

	*multi = PCL_MULTI_MODE_NONE;
	return(PCL_OK);
}

#else // PCL_MULTI_MODE

int _PCL_FDEF pclGetMultiMode(	/* @101, get multi operating mode */
	int hdl,					/* in:	handle from pclOpen() */
	int _PCL_OUT *multi			/* out:	multi operating mode, like PCL_MULTI_MODE_xxx */
) {
	DO_UNSUPPORTED_1(GetMultiMode, multi);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

#if defined(_Windows) || defined(_WIN32)

int _PCL_FDEF pclGetDriverInformation(		// @78, get driver information
	int hdl,								// in:	handle from pclOpen()
	PCL_DRIVER_INFORMATION _PCL_OUT *info	// out:	information data structure
) {
	CHECK_PTR(info);
#	define pclGetDriverInformationMulti pclMultiNotSupported
	PCI_GET_HANDLE(hdl, pclGetDriverInformation, (), hPCI, card);
//	if (info->size < sizeof(*info)) return(PCL_ERR_RANGE);

#define NEEDED_SIZE(_type, _member) \
	( \
		offsetof(_type, _member) \
	+	SIZEOF_MEMBER(_type, _member) \
	)

	if (info->size >= NEEDED_SIZE(PCL_DRIVER_INFORMATION, hWD)) {
		info->hWD = hPCI->hWD;
	}

	if (info->size >= NEEDED_SIZE(PCL_DRIVER_INFORMATION, sWdVersion)) {
		WD_VERSION ver;
		BZERO(ver);
		WD_Version(hPCI->hWD, &ver);

		info->sWdVersion.dwVer = ver.dwVer;
		strncpy_s(info->sWdVersion.cVer, sizeof(info->sWdVersion.cVer), ver.cVer, sizeof(info->sWdVersion.cVer));
	}

	if (info->size >= NEEDED_SIZE(PCL_DRIVER_INFORMATION, wSubsystemDeviceID)) {
		info->wVendorID = gPciCard.v[card].wVendorID;
		info->wDeviceID = gPciCard.v[card].wDeviceID;
		info->wSubsystemVendorID = gPciCard.v[card].wSubsystemVendorID;
		info->wSubsystemDeviceID = gPciCard.v[card].wSubsystemDeviceID;
	}
	if (info->size >= NEEDED_SIZE(PCL_DRIVER_INFORMATION, chipVer)) {
		info->intHandler = pgPciHdl->v[hdl].intHandler.p;
		info->chipVer = PCI_ReadVERSION(hPCI);
	}
	if (info->size >= NEEDED_SIZE(PCL_DRIVER_INFORMATION, perf)) {
		info->perf.freq = gPciCard.v[card].perf.freq;
		info->perf.abs = gPciCard.v[card].perf.abs;
	}
	return(PCL_OK);
}

#endif

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetIdata(	// @79, $38, get idata
	int hdl,				// in:	handle from pclOpen()
	int addr,				// in:	idata address
	int _PCL_OUT *idata		// out:	byte read from idata
) {
	int result;
	CHECK_PTR(idata);

#	define pclGetIdataMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetIdata, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, addr);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", addr);
	if ((result = pclPciExec(hdl, CMD_GET_IDATA)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetIdata, result);
	*idata = PCI_ReadDATA0(hPCI);

	UNLOCK_RETURN(hdl, pclGetIdata, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetPdata(	// @80, $3A, get pdata
	int hdl,				// in:	handle from pclOpen()
	int addr,				// in:	pdata address
	int _PCL_OUT *pdata		// out:	byte read from pdata
) {
	int result;
	CHECK_PTR(pdata);

#	define pclGetPdataMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetPdata, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, addr);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", addr);
	if ((result = pclPciExec(hdl, CMD_GET_PDATA)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetPdata, result);
	*pdata = PCI_ReadDATA0(hPCI);

	UNLOCK_RETURN(hdl, pclGetPdata, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetXdata(	// @83, $3C, get xdata
	int hdl,				// in:	handle from pclOpen()
	int addr,				// in:	xdata address
	int _PCL_OUT *xdata		// out:	byte read from xdata
) {
	int result;
	CHECK_PTR(xdata);

#	define pclGetXdataMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetXdata, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, addr >> 8);
	PCI_WriteCMDD1(hPCI, addr);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %04X", addr);
	if ((result = pclPciExec(hdl, CMD_GET_XDATA)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetXdata, result);
	*xdata = PCI_ReadDATA0(hPCI);

	UNLOCK_RETURN(hdl, pclGetXdata, PCL_OK);
}

/***************************************************************************/

#ifdef PCL_OPT_VBI				/* read vbi lines */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetVbiLines(	/* @95, setup vbi reader */
	int hdl,					/* in:	handle from pclOpen() */
	int line1,					/* in:	first line */
	int lines					/* in:	number of lines */
) {
//	int result;
	CHECK_RANGE(line1, 0, 255 - PCI_LOG_LINES_COUNT_M, PCL_ERR_RANGE);
	CHECK_RANGE(lines, 1, PCI_LOG_LINES_COUNT_M, PCL_ERR_RANGE);

#	define pclSetVbiLinesMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclSetVbiLines, (), hPCI, card);

	PCI_WriteLogLine(hPCI, line1);
	lines |= 0
	|	PCI_LOG_LINES_BLOCK_M
	|	PCI_LOG_LINES_NO_HANC_M
	|	PCI_LOG_LINES_Y_ONLY_M;
	PCI_WriteLogLines(hPCI, lines);

	PCI_WriteLogControl(hPCI, 0/*PCI_LOG_CONTROL_NEXT_Y_M*/);
	PCI_WriteLogPixel(hPCI, 0);
	PCI_WriteLogDone(hPCI, 0);

	UNLOCK_RETURN(hdl, pclSetVbiLines, PCL_OK);
}

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetVbiData(	/* @96, get vbi data */
	int hdl,					/* in:	handle from pclOpen() */
	int lines,					/* in:	number of lines to read */
	int pixels,					/* in:	number of pixels per line */
	BYTE _PCL_OUT *vector[]		/* out:	pointer to array of pointers to pixel data */
) {
	CHECK_RANGE(lines, 1, PCI_LOG_LINES_COUNT_M, PCL_ERR_RANGE);
//	CHECK_RANGE(pixels, 720, 1920, PCL_ERR_RANGE);
	if (lines * pixels > PCI_LOG_ADDR_MAX) {
		PCI_ERR("Parameter 'lines * pixels' (%d * %d = %d) is greater than %d\n", lines, pixels, lines * pixels, PCI_LOG_ADDR_MAX);
		return PCL_ERR_RANGE;
	}
	CHECK_RANGE(lines * pixels, 0, PCI_LOG_ADDR_MAX, PCL_ERR_RANGE);
	CHECK_PTR(vector);
	for (int line = 0; line < lines; line++) CHECK_PTR(vector[line]);

#	define pclGetVbiDataMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetVbiData, (), hPCI, card);

	if (lines * pixels > gPciCard.v[card].logAddr) {
		PCI_ERR("Parameter 'lines * pixels' (%d * %d = %d) is greater than %d\n", lines, pixels, lines * pixels, gPciCard.v[card].logAddr);
		UNLOCK_RETURN(hdl, pclGetVbiData, PCL_ERR_RANGE);
	}

	PCI_WriteLogAddr(hPCI, 0);
	for (int line = 0; line < lines; line++) {

		// dummy read
		PCI_ReadLogBYTE2(hPCI);	// y

		// block read
		PCI_ReadLogMultiBYTE2(hPCI, pixels, vector[line]);	// y
	}

	UNLOCK_RETURN(hdl, pclGetVbiData, PCL_OK);
}

/*-------------------------------------------------------------------------*/

#else // PCL_OPT_VBI

int _PCL_FDEF pclSetVbiLines(	/* @95, setup vbi reader */
	int hdl,					/* in:	handle from pclOpen() */
	int line1,					/* in:	first line */
	int lines					/* in:	number of lines */
) {
	DO_UNSUPPORTED_2(SetVbiLines, line1, lines);
}

int _PCL_FDEF pclGetVbiData(	/* @96, get vbi data */
	int hdl,					/* in:	handle from pclOpen() */
	int lines,					/* in:	number of lines to read */
	int pixels,					/* in:	number of pixels per line */
	BYTE _PCL_OUT *vector[]		/* out:	pointer to array of pointers to pixel data */
) {
	DO_UNSUPPORTED_3(SetVbiLines, lines, pixels, vector);
}

#endif // PCL_OPT_VBI

/***************************************************************************/

#ifdef PCL_OPT_ANC				/* read user defined ancillary data */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetAncDID(		/* @97, $4D, setup anc reader */
	int hdl,					/* in:	handle from pclOpen() */
	int did,					/* in:	did */
	int sdid					/* in:	sdid */
) {
	int result;

	CHECK_RANGE(did, 0, 255, PCL_ERR_RANGE);
	CHECK_RANGE(sdid, 0, 255, PCL_ERR_RANGE);

#	define pclSetAncDIDMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclSetAncDID, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, did);
	PCI_WriteCMDD1(hPCI, sdid);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %04X", sdid, did);
	if ((result = pclPciExec(hdl, CMD_SET_ANC_DID)) != PCL_OK) UNLOCK_RETURN(hdl, pclSetAncDID, result);

	UNLOCK_RETURN(hdl, pclSetAncDID, PCL_OK);
}

/*-------------------------------------------------------------------------*/

typedef struct {
	BYTE did;		// 0	did
	BYTE sdid;		// 1	sdid
	BYTE dc;		// 2	dc
	BYTE addrH;		// 3	startadresse daten
	BYTE addrL;		// 4
	BYTE error;		// 5	error flags
} PCL_ANC_DESC;

int _PCL_FDEF pclGetAncData(	/* @98, $4E, get anc data */
	int hdl,					/* in:	handle from pclOpen() */
	int size,					/* in:	size of anc data buffer (vector) */
	BYTE _PCL_OUT *did,			/* out:	pointer to did */
	BYTE _PCL_OUT *sdid,		/* out:	pointer to sdid */
	BYTE _PCL_OUT *dc,			/* out:	pointer to data count */
	BYTE _PCL_OUT *vector,		/* out:	pointer to anc data */
	BYTE _PCL_OUT *chk,			/* out:	pointer to checksum */
	BYTE _PCL_OUT *error		/* out:	pointer to error flags */
) {
	int result;

	CHECK_PTR(vector);
	CHECK_RANGE(size, 0, 255, PCL_ERR_RANGE);

#	define pclGetAncDataMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetAncData, (), hPCI, card);

	// get anc data pointer
	if ((result = pclPciExec(hdl, CMD_GET_ANC_DATA)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetAncData, result);
	int adesc = PCI_ReadDATA2(hPCI) | PCI_ReadDATA3(hPCI) << 8;

	// set ancram start address for descriptor
	PCI_WriteAncAddr(hPCI, adesc);

	// read descriptor
	PCL_ANC_DESC desc;
	PCI_ReadAncMultiBYTE(hPCI, sizeof(desc), &desc.did);

	if (did)   *did   = desc.did;
	if (sdid)  *sdid  = desc.sdid;
	if (dc)    *dc    = desc.dc;
	if (error) *error = desc.error;

	int addr = desc.addrH << 8 | desc.addrL;
	if (size > desc.dc) size = desc.dc;
	if ((addr + size) > 512) addr = 512 - size;

	// set ancram start address for data
	PCI_WriteAncAddr(hPCI, addr);

	// read data
	PCI_ReadAncMultiBYTE(hPCI, size, vector);

	// calculate checksum
	if (chk) {
		BYTE check = 0;

		check = desc.did + desc.sdid + desc.dc;
		for (int i = 0; i < desc.dc; i++) {
			check += vector[i];
		}
		*chk = check;
	}

	UNLOCK_RETURN(hdl, pclGetAncData, result);
}

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetAncDataTab(		/* @107, get anc data, extended version */
	int hdl,						/* in:	handle from pclOpen() */
	int size,						/* in:	size of anc data buffer (vector) */
	int count,						/* in:	number of packets in buffer */
	int _PCL_OUT *first,			/* out:	pointer to number of first packet */
	int _PCL_OUT *packets,			/* out:	pointer to packet counter */
	BYTE _PCL_OUT *vector[],		/* out:	pointer to array of pointers to anc data */
	PCL_ANC_INFO _PCL_OUT *info		/* out:	pointer to array of info struct */
) {
	CHECK_PTR(vector);
	CHECK_RANGE(size, 0, 255, PCL_ERR_RANGE);
	int result;

#	define pclGetAncDataTabMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetAncDataTab, (), hPCI, card);

	enum {
		NELT_CMD_DATA	= 11,
		ANC_DESC_MAX	= (NELT_CMD_DATA - 2) / 2,
	};
	int index = pgPciHdl->v[hdl].anc.index;

	if (index >= count) {
		UNLOCK_RETURN(hdl, pclGetAncDataTab, PCL_ERR_RANGE);
	}

	int cnt = index + 1;
	while (index < cnt) {
		// get anc data pointer
		int adesc[ANC_DESC_MAX];
		int frst;
		PCI_WriteCMDD0(hPCI, index);
		PCI_WriteCMDD1(hPCI, 0);	// reserved for future use
		APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", index);
		if ((result = pclPciExec(hdl, CMD_GET_ANC_DATA_TAB)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetAncDataTab, result);
		frst = PCI_ReadDATA0(hPCI);
		cnt  = PCI_ReadDATA1(hPCI);
		int cdesc = cnt;
//		if (cdesc > NELT(adesc)) cdesc = NELT(adesc);

		if (cdesc == 0) {
			UNLOCK_RETURN(hdl, pclGetAncDataTab, PCL_NO_NEWD);
		}
		if (cdesc >= 1) {
			adesc[0] = PCI_ReadDATA2(hPCI) | PCI_ReadDATA3(hPCI) << 8;
		}
		if (cdesc >= 2) {
			adesc[1] = PCI_ReadDATA4(hPCI) | PCI_ReadDATA5(hPCI) << 8;
		}
		if (cdesc >= 3) {
			adesc[2] = PCI_ReadDATA6(hPCI) | PCI_ReadDATA7(hPCI) << 8;
		}
		if (cdesc >= 4) {
			adesc[3] = PCI_ReadDATA8(hPCI) | PCI_ReadDATA9(hPCI) << 8;
		}
		if (cdesc >= ANC_DESC_MAX + 1) {
			assert(FALSE);
			cdesc = ANC_DESC_MAX;
		}

		for (int i = frst; i < frst + cdesc; i++) {
			if (i >= count) break;

			// set ancram start address for descriptor
			assert(i - frst < NELT(adesc));
/*			if (hPCI->bPcieV2) {
				PCI_WriteAncAddr(hPCI, adesc[i - frst] - 1);//quirk
			} else
*/			{
				PCI_WriteAncAddr(hPCI, adesc[i - frst]);
			}

			// read descriptor
			PCL_ANC_DESC desc;
			PCI_ReadAncMultiBYTE(hPCI, sizeof(desc), &desc.did);

			if (info) {
				assert(i < count);
				info[i].did   = desc.did;
				info[i].sdid  = desc.sdid;
				info[i].dc    = desc.dc;
				info[i].error = desc.error;
			}

			int addr = desc.addrH << 8 | desc.addrL;

			if (size > desc.dc) size = desc.dc;
			if ((addr + size) > 512) addr = 512 - size;

			// set ancram start address for data
/*			if (hPCI->bPcieV2) {
				PCI_WriteAncAddr(hPCI, addr - 1);//quirk
			} else
*/			{
				PCI_WriteAncAddr(hPCI, addr);
			}

			// read data
			assert(i < count);
			PCI_ReadAncMultiBYTE(hPCI, size, vector[i]);

			index++;
		}
	}
	if (first)		*first   =         pgPciHdl->v[hdl].anc.index;
	if (packets)	*packets = index - pgPciHdl->v[hdl].anc.index;
	pgPciHdl->v[hdl].anc.index = index;

	UNLOCK_RETURN(hdl, pclGetAncDataTab, result);
}

/*-------------------------------------------------------------------------*/

#else // PCL_OPT_ANC

int _PCL_FDEF pclSetAncDID(		/* @97, $4D, setup anc reader */
	int hdl,					/* in:	handle from pclOpen() */
	int did,					/* in:	did */
	int sdid					/* in:	sdid */
) {
	DO_UNSUPPORTED_2(SetVbiLines, did, sdid);
}

int _PCL_FDEF pclGetAncData(	/* @98, $4E, get anc data */
	int hdl,					/* in:	handle from pclOpen() */
	int size,					/* in:	size of anc data buffer (vector) */
	BYTE _PCL_OUT *did,			/* out:	pointer to did */
	BYTE _PCL_OUT *sdid,		/* out:	pointer to sdid */
	BYTE _PCL_OUT *dc,			/* out:	pointer to data count */
	BYTE _PCL_OUT *vector,		/* out:	pointer to anc data */
	BYTE _PCL_OUT *chk,			/* out:	pointer to checksum */
	BYTE _PCL_OUT *error		/* out:	pointer to did */
) {
	DO_UNSUPPORTED_7(pclGetAncData, size, did, sdid, dc, vector, chk, error);
}

int _PCL_FDEF pclGetAncDataTab(		/* @107, get anc data, extended version */
	int hdl,						/* in:	handle from pclOpen() */
	int size,						/* in:	size of anc data buffer (vector) */
	int count,						/* in:	number of packets in buffer */
	int _PCL_OUT *first,			/* out:	pointer to number of first packet */
	int _PCL_OUT *packets,			/* out:	pointer to packet counter */
	BYTE _PCL_OUT *vector[],		/* out:	pointer to array of pointers to anc data */
	PCL_ANC_INFO _PCL_OUT *info		/* out:	pointer to array of info struct */
) {
	DO_UNSUPPORTED_6(pclGetAncDataTab, size, count, first, packets, vector, info);
}

#endif // PCL_OPT_ANC

/***************************************************************************/

#ifdef PCL_OPT_GPI_IN			/* gpi inputs */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetGpi(	/* @99, $4F, get status of gpi in/out */
	int hdl,				/* in:	handle from pclOpen() */
	WORD _PCL_OUT *gpii,	/* out:	status of gpi inputs like PCL_GPI_xxx */
	WORD _PCL_OUT *gpio		/* out:	status of gpi outputs like PCL_GPI_xxx */
) {
	int result;

#	define pclGetGpiMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetGpi, (), hPCI, card);

	if ((result = pclPciExec(hdl, CMD_GET_GPI)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetGpi, result);
	if (gpii) *gpii = PCI_ReadDATA0(hPCI);
	if (gpio) *gpio = PCI_ReadDATA1(hPCI);

	UNLOCK_RETURN(hdl, pclGetGpi, PCL_OK);
}

/*-------------------------------------------------------------------------*/

#endif // PCL_OPT_GPI_IN

/***************************************************************************/

#ifdef PCL_OPT_GPI_IN_EVNT		/* gpi input events */

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclSetGpiEvent(	/* @108, $53, set event handling of gpi inputs */
	int hdl,					/* in:	handle from pclOpen() */
	int evnt,					/* in:	gpi event like PCL_GPI_EVNT_xxx */
	int reg						/* in:	registers number like PCL_REG_xxx */
) {
	int result;

#	define pclSetGpiEventMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclSetGpiEvent, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, evnt);
	PCI_WriteCMDD1(hPCI, reg);
	if ((result = pclPciExec(hdl, CMD_SET_GPI_EVENT)) != PCL_OK) UNLOCK_RETURN(hdl, pclSetGpiEvent, result);

	UNLOCK_RETURN(hdl, pclSetGpiEvent, PCL_OK);
}

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetGpiEvent(	/* @109, $54, get event of gpi input */
	int hdl,					/* in:	handle from pclOpen() */
	int gpi,					/* in:	gpi input like PCL_GPI_xxx */
	PCL_TC _PCL_OUT *time,		/* out:	timecode from register, bcd */
	int _PCL_OUT *bits,			/* out:	misc. bits of register,
										like PCL_BITS_xxx_xxx */
	PCL_TC _PCL_OUT *user,		/* out:	user-bits from register, bcd */
	int _PCL_OUT *reg			/* out:	registers number like PCL_REG_xxx */
) {
	int result;

#	define pclGetGpiEventMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetGpiEvent, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, gpi);
	if ((result = pclPciExec(hdl, CMD_GET_GPI_EVENT)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetGpiEvent, result);
	if (time != NULL) {
		*((BYTE *) time    ) = (BYTE) PCI_ReadDATA0(hPCI);
		*((BYTE *) time + 1) = (BYTE) PCI_ReadDATA1(hPCI);
		*((BYTE *) time + 2) = (BYTE) PCI_ReadDATA2(hPCI);
		*((BYTE *) time + 3) = (BYTE) PCI_ReadDATA3(hPCI);
	}
	if (bits != NULL) *bits = PCI_ReadDATA4(hPCI);
	if (user != NULL) {
		*((BYTE *) user    ) = (BYTE) PCI_ReadDATA5(hPCI);
		*((BYTE *) user + 1) = (BYTE) PCI_ReadDATA6(hPCI);
		*((BYTE *) user + 2) = (BYTE) PCI_ReadDATA7(hPCI);
		*((BYTE *) user + 3) = (BYTE) PCI_ReadDATA8(hPCI);
	}
	if (reg != NULL) *reg = PCI_ReadDATA9(hPCI);

	UNLOCK_RETURN(hdl, pclGetGpiEvent, PCL_OK);
}

/*-------------------------------------------------------------------------*/

int _PCL_FDEF pclGetGpiEvents(	/* @110, $55, get events of gpi inputs */
	int hdl,					/* in:	handle from pclOpen() */
	WORD _PCL_OUT *gpii,		/* out:	status of gpi inputs like PCL_GPI_xxx */
	WORD _PCL_OUT *change		/* out:	gpi change status like PCL_GPI_xxx */
) {
	int result;

#	define pclGetGpiEventsMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetGpiEvents, (), hPCI, card);

	if ((result = pclPciExec(hdl, CMD_SET_GPI_IRQ)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetGpiEvents, result);
	if (gpii)	*gpii	= PCI_ReadDATA0(hPCI);
	if (change)	*change	= PCI_ReadDATA1(hPCI);

	UNLOCK_RETURN(hdl, pclGetGpiEvents, PCL_OK);
}

/*-------------------------------------------------------------------------*/

#endif /* PCL_OPT_GPI_IN_EVNT */

/***************************************************************************/

#ifdef PCL_OPT_GPI_OUT			/* event driven gpi outputs */

int _PCL_FDEF pclSetGpiEventTime(	/* @81, $28/$2C, set gpi event time */
	int hdl,						/* in:	handle from pclOpen() */
	int gpi,						/* in:	number of gpi output (1 or 2) */
	int reg,						/* in:	registers number like PCL_REG_xxx */
	int mode,						/* in:	registers number like PCL_GPIM_xxx */
	PCL_TC time						/* in:	timecode of event, bcd */
) {
	int result;

	CHECK_RANGE(reg, PCL_REG_MIXED, PCL_REG_MIXED, PCL_ERR_RANGE);
	CHECK_RANGE(gpi, 1, 2, PCL_ERR_RANGE);

#	define pclSetGpiEventTimeMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclSetGpiEventTime, (), hPCI, card);

	assert(CMD_GPI_OUT_2 > CMD_GPI_OUT_1);
	// gpi == 1 -> cmd = CMD_GPI_OUT_1
	// gpi == 2 -> cmd = CMD_GPI_OUT_2
	int cmd = CMD_GPI_OUT_1 - (CMD_GPI_OUT_2 - CMD_GPI_OUT_1) + gpi * (CMD_GPI_OUT_2 - CMD_GPI_OUT_1);

	PCI_WriteCMDD0(hPCI, *((BYTE *) &time + 3));	// hour
	PCI_WriteCMDD1(hPCI, *((BYTE *) &time + 2));	// minute
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X %02X", *((BYTE *) &time + 2), *((BYTE *) &time + 3));
	if ((result = pclPciExec(hdl, cmd + 0)) != PCL_OK) UNLOCK_RETURN(hdl, pclSetGpiEventTime, result);

	PCI_WriteCMDD0(hPCI, *((BYTE *) &time + 1));	// second
	PCI_WriteCMDD1(hPCI, *((BYTE *) &time + 0));	// frame
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X %02X", *((BYTE *) &time + 0), *((BYTE *) &time + 1));
	if ((result = pclPciExec(hdl, cmd + 1)) != PCL_OK) UNLOCK_RETURN(hdl, pclSetGpiEventTime, result);

	PCI_WriteCMDD0(hPCI, *((BYTE *) &mode + 0));
	PCI_WriteCMDD1(hPCI, *((BYTE *) &mode + 1));
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X %02X", *((BYTE *) &mode + 1), *((BYTE *) &mode + 0));
	if ((result = pclPciExec(hdl, cmd + 2)) != PCL_OK) UNLOCK_RETURN(hdl, pclSetGpiEventTime, result);

#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.setGpiEventTime.use = TRUE;
	gPciCard.v[card].persist.setGpiEventTime.gpi = gpi;
	gPciCard.v[card].persist.setGpiEventTime.reg = reg;
	gPciCard.v[card].persist.setGpiEventTime.mode = mode;
	gPciCard.v[card].persist.setGpiEventTime.time = time;
#endif

	UNLOCK_RETURN(hdl, pclSetGpiEventTime, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetGpiEventStatus(	/* @82, $28/$2C, get gpi event status */
	int hdl,						/* in:	handle from pclOpen() */
	int gpi,						/* in:	number of gpi output (1 or 2) */
	int _PCL_OUT *status,			/* out:	status like PCL_GPIS_xxx */
	int _PCL_OUT *count				/* out:	pulse counter */
) {
	int result;

	CHECK_RANGE(gpi, 1, 2, PCL_ERR_RANGE);

#	define pclGetGpiEventStatusMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetGpiEventStatus, (), hPCI, card);

	assert(CMD_GPI_OUT_2 > CMD_GPI_OUT_1);
	// gpi == 1 -> cmd = CMD_GPI_OUT_1
	// gpi == 2 -> cmd = CMD_GPI_OUT_2
	int cmd = CMD_GPI_OUT_1 - (CMD_GPI_OUT_2 - CMD_GPI_OUT_1) + gpi * (CMD_GPI_OUT_2 - CMD_GPI_OUT_1);

	if ((result = pclPciExec(hdl, cmd + 3)) != PCL_OK) UNLOCK_RETURN(hdl, pclGetGpiEventStatus, result);

	if (status != 0) *status = PCI_ReadDATA0(hPCI);
	if (count  != 0) *count  = PCI_ReadDATA1(hPCI);

	UNLOCK_RETURN(hdl, pclGetGpiEventStatus, PCL_OK);
}

#endif

//---------------------------------------------------------------------------

#ifndef PCL_PASCAL

PCL_TC _PCL_FDEF pclTc2Bcd(	// convert binary timecode to packed bcd
	int hours,				// in:	hours of binary timecode
	int minutes,			// in:	minutes of binary timecode
	int seconds,			// in:	seconds of binary timecode
	int frames				// in:	frames of binary timecode
) {
	return(
		((long) HEX2BCD(hours)   << 24) |
		((long) HEX2BCD(minutes) << 16) |
		((long) HEX2BCD(seconds) <<  8) |
		((long) HEX2BCD(frames)       )
	);
}

#endif

//---------------------------------------------------------------------------

int _PCL_FDEF pclBcd2Tc(	// convert packed bcd timecode to binary
	PCL_TC time,			// in:	packed bcd timecode
	int _PCL_OUT *hours,	// out:	hours of binary timecode
	int _PCL_OUT *minutes,	// out:	minutes of binary timecode
	int _PCL_OUT *seconds,	// out:	seconds of binary timecode
	int _PCL_OUT *frames	// out:	frames of binary timecode
) {
	if (hours	!= NULL) *hours	  = BCD2HEX(HIBYTE(HIWORD(time)));
	if (minutes != NULL) *minutes = BCD2HEX(LOBYTE(HIWORD(time)));
	if (seconds != NULL) *seconds = BCD2HEX(HIBYTE(LOWORD(time)));
	if (frames  != NULL) *frames  = BCD2HEX(LOBYTE(LOWORD(time)));

	return(PCL_OK);
}

//---------------------------------------------------------------------------

#ifndef PCL_PASCAL

long _PCL_FDEF pclTc2Frames(	// convert packed bcd timecode to long
	PCL_TC time,				// in:	packed bcd timecode
	int frate					// in:	frame rate (25 or 30)
) {
	return(
		(long) BCD2HEX(LOBYTE(LOWORD(time)))                   +
		(long) BCD2HEX(HIBYTE(LOWORD(time))) * frate           +
		(long) BCD2HEX(LOBYTE(HIWORD(time))) * frate * 60      +
		(long) BCD2HEX(HIBYTE(HIWORD(time))) * frate * 60 * 60
	);
}

#endif

//---------------------------------------------------------------------------

#if defined PCL_OPT_REALTIME && !defined PCL_PASCAL

#define PCL_PCI_REALTIME_TICK_NANO 9079

//---------------------------------------------------------------------------

PCL_TC _PCL_FDEF pclTcInc(PCL_TC time, unsigned rate)
{
	time += 0x00000001L;
	if ((time & 0x0000000FL) > 0x00000009L) {
		time += 0x00000010L - 0x0000000AL;
	}
	if ((time & 0x000000FFL) > HEX2BCD(rate - 1)) {
		time += 0x00000100L - HEX2BCD(rate);

		if ((time & 0x00000F00L) > 0x00000900L) {
			time += 0x00001000L - 0x00000A00L;
		}
		if ((time & 0x0000FF00L) > 0x00005900L) {
			time += 0x00010000L - 0x00006000L;

			if ((time & 0x000F0000L) > 0x00090000L) {
				time += 0x00100000L - 0x000A0000L;
			}
			if ((time & 0x00FF0000L) > 0x00590000L) {
				time += 0x01000000L - 0x00600000L;

				if ((time & 0x0F000000L) > 0x09000000L) {
					time += 0x10000000L - 0x0A000000L;
				}
				if ((time & 0xFF000000L) > 0x23000000L) {
					time += (PCL_TC) (0x00000000L - 0x24000000L);
				}
			}
		}
	}
	return(time);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclGetRealtime(	// @102, $15, get realtime
	int hdl,					// in:	handle from pclOpen()
	PCL_REALTIME _PCL_OUT *tm	// out:	realtime from pcl
) {
	PCL_TC tim, user;
	WORD tick;
	BYTE dval;
	long micro;
#	define pclGetRealtimeMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclGetRealtime, (), hPCI, card);

	BYTE ack;
	static BYTE lastAck;
	int framesPerSecond;

	CHECK_PTR(tm);
	tm->ltc.valid = PCL_RTVAL_INVALID;

	ack = PCI_ReadACK(hPCI);
	if (ack == lastAck) {
		PCI_ERR("pclGetRealtime(): ACK was not changed (%2x)\n", ack);
		UNLOCK_RETURN(hdl, pclGetRealtime, PCL_ERR_TIMEOUT);
	}
	lastAck = ack;

	*((BYTE *) &tick    ) = (BYTE) PCI_ReadDATA4(hPCI);
	*((BYTE *) &tick + 1) = (BYTE) PCI_ReadDATA5(hPCI);

	*((BYTE *) &tim     ) = (BYTE) PCI_ReadDATA3(hPCI);
	*((BYTE *) &tim  + 1) = (BYTE) PCI_ReadDATA2(hPCI);
	*((BYTE *) &tim  + 2) = (BYTE) PCI_ReadDATA1(hPCI);
	*((BYTE *) &tim  + 3) = (BYTE) PCI_ReadDATA0(hPCI);

	*((BYTE *) &user    ) = (BYTE) PCI_ReadDATA9(hPCI);
	*((BYTE *) &user + 1) = (BYTE) PCI_ReadDATA8(hPCI);
	*((BYTE *) &user + 2) = (BYTE) PCI_ReadDATA7(hPCI);
	*((BYTE *) &user + 3) = (BYTE) PCI_ReadDATA6(hPCI);

	dval                  = (BYTE) PCI_ReadDATAA(hPCI);

#if defined _NTDDK_
	{
		static LARGE_INTEGER t;
		TIME_FIELDS tf;

		RtlTimeToTimeFields(&t, &tf);
		tm->system.day   = tf.Day;
		tm->system.month = tf.Month;
		tm->system.year  = tf.Year;
		tm->system.hour  = tf.Hour;
		tm->system.min   = tf.Minute;
		tm->system.sec   = tf.Second;
		tm->system.milli = tf.Milliseconds;

		PCI_WriteCMDW(hPCI, 0x15);
		KeQuerySystemTime(&t);
	}
#elif defined WIN32
	{
		static SYSTEMTIME t;
		int priority;
		DWORD clas;
		HANDLE thread;
		HANDLE process;
		BOOL b;

		tm->system.day   = t.wDay;
		tm->system.month = t.wMonth;
		tm->system.year  = t.wYear;
		tm->system.hour  = t.wHour;
		tm->system.min   = t.wMinute;
		tm->system.sec   = t.wSecond;
		tm->system.milli = t.wMilliseconds;

		process = GetCurrentProcess();
		clas = GetPriorityClass(process);

		thread = GetCurrentThread();
		priority = GetThreadPriority(thread);

		b = SetPriorityClass(process, REALTIME_PRIORITY_CLASS);
		b = SetThreadPriority(thread, THREAD_PRIORITY_TIME_CRITICAL);

		PCI_WriteCMDW(hPCI, 0x15);
//		Sleep(200);
		GetSystemTime(&t);

		b = SetPriorityClass(process, clas);
		b = SetThreadPriority(thread, priority);
	}
#elif defined LINUX
	{

	}
#else
	{
		static struct time t;
		static struct date d;
		struct tm st;
		time_t tt;
		struct tm *pt;

		st.tm_sec   = t.ti_sec;
		st.tm_min   = t.ti_min;
		st.tm_hour  = t.ti_hour;
		st.tm_mday  = d.da_day;
		st.tm_mon   = d.da_mon - 1;
		st.tm_year  = d.da_year - 1900;
		st.tm_isdst = 0;

		tt = mktime(&st);
		pt = gmtime(&tt);

		tm->system.day   = pt->tm_mday;
		tm->system.month = pt->tm_mon + 1;
		tm->system.year  = pt->tm_year + 1900;
		tm->system.hour  = pt->tm_hour;
		tm->system.min   = pt->tm_min;
		tm->system.sec   = pt->tm_sec;
		tm->system.milli = t.ti_hund * 10;

		PCI_WriteCMDW(hPCI, 0x15);
		gettime(&t);
		getdate(&d);
	}
#endif

	if (tick >= 0x1200) {
		UNLOCK_RETURN(hdl, pclGetRealtime, PCL_NO_NEWD);
	}
	if (dval == 0) {
		tm->ltc.valid |= PCL_RTVAL_DATE;
	}

	micro = ((long) tick * PCL_PCI_REALTIME_TICK_NANO) / 1000;

	tm->ltc.milli = (WORD) (micro / 1000);
	tm->ltc.micro = (WORD) (micro % 1000);

	{
		static int lastSecond, lastFrame, frate = PCL_FRAME_UNKNOWN, newFrate, change = 0;
		int second, frame;

		frame  = BCD2HEX((int) (tim)      & 0xFF);
		second = BCD2HEX((int) (tim >> 8) & 0xFF);

		if (second == lastSecond + 1) {
			switch (newFrate) {
			case PCL_FRAME_25:
				if (lastFrame < 25) change++; else change = -1;
				break;
			case PCL_FRAME_30DF:
				if (lastFrame > 25) change++; else change = -1;
				break;
			default:
				change = -1;
				break;
			}
			if (change == -1) {
				if (lastFrame < 25) newFrate = PCL_FRAME_25;
				if (lastFrame > 25) newFrate = PCL_FRAME_30DF;
			}
			if (change >= 5) {
				frate = newFrate;
				change = 0;
			}
		}
		lastSecond = second;
		lastFrame = frame;
		tm->ltc.frame = frate;
	}
/*
	{
		tim = 0x01020304;
		tm->ltc.milli = 5;
	}
*/
	if (tm->ltc.frame == PCL_FRAME_30DF) {
		__int64 milli, frame, min, min10;
		int minute = BCD2HEX((int) (tim >> 16) & 0xFF) + BCD2HEX((int) (tim >> 24) & 0xFF) * 60;
#		define milliPerFrame30 1001
#		define milliPerDay (24l*60*60*1000)
#		define framesPerDay (2589408l)
		milli = tm->ltc.milli;
		frame = (
			BCD2HEX((int) (tim)       & 0xFF) +
			BCD2HEX((int) (tim >> 8)  & 0xFF) * 30 +
			BCD2HEX((int) (tim >> 16) & 0xFF) * 30 * 60 +
			BCD2HEX((int) (tim >> 24) & 0xFF) * 30 * 60 * 60
		);
		min   = minute;
/*#ifdef LINUX
		min10 = minute;
		do_div(min10, 10);
#else*/
		min10 = minute / 10;
//#endif
		frame -= (min - min10) * 2;

		{
			// Die Berechnung in der folgenden Zeile basiert auf der
			// NTSC-Nennfrequenz.  Sie fhrt um Mitternacht zu
			// einem Sprung der LTC-Zeit von 86ms.
//			milli = milli + (frame * milliPerFrame30) / 30;

			// Die folgende Zeile umgeht dies, indem sie nicht von der
			// NTSC-Nennfrequenz ausgeht, sondern den Tag in 'framesPerDay'
			// gleiche Teile einteilt
/*#ifdef LINUX
			milli += (frame * milliPerDay);
			do_div(milli, framesPerDay);
#else*/
			milli = milli + (frame * milliPerDay) / framesPerDay;
//#endif
		}
/*#ifdef LINUX
		tm->ltc.milli = (int) do_div(milli, 1000);	// milli /= 1000;
		tm->ltc.sec   = (int) do_div(milli, 60);	// milli /= 60;
		tm->ltc.min   = (int) do_div(milli, 60);	// milli /= 60;
		tm->ltc.hour  = (int) do_div(milli, 24);
#else*/
		tm->ltc.milli = (int) (milli % 1000);	milli /= 1000;
		tm->ltc.sec   = (int) (milli % 60);	milli /= 60;
		tm->ltc.min   = (int) (milli % 60);	milli /= 60;
		tm->ltc.hour  = (int) (milli % 24);
//#endif
	} else {
		switch (tm->ltc.frame) {
		case PCL_FRAME_24:
			framesPerSecond = 24;
			break;
		case PCL_FRAME_25:
		default:
			framesPerSecond = 25;
			break;
		case PCL_FRAME_30:
			framesPerSecond = 30;
			break;
		}
		if (tm->ltc.milli * framesPerSecond >= 1000) {
			tim = pclTcInc(tim, framesPerSecond);
			if (tim == 0) {
				tm->ltc.valid &= ~PCL_RTVAL_DATE;
			}
			tm->ltc.milli -= 1000 / framesPerSecond;
		}
		tm->ltc.milli += (BCD2HEX((int) tim & 0xFF) * 1000) / framesPerSecond;	tim >>= 8;
		tm->ltc.sec    = BCD2HEX((int) tim & 0xFF);					tim >>= 8;
		tm->ltc.min    = BCD2HEX((int) tim & 0xFF);					tim >>= 8;
		tm->ltc.hour   = BCD2HEX((int) tim & 0xFF);
	}

	tm->ltc.year  = BCD2HEX((int) user & 0xFF) + 1900;	user >>= 8;
	tm->ltc.month = BCD2HEX((int) user & 0xFF);			user >>= 8;
	tm->ltc.day   = BCD2HEX((int) user & 0xFF);			user >>= 8;
	tm->ltc.tzone = ((int) user >> 1) & 3;
	if (user & 0x01) tm->ltc.valid |= PCL_RTVAL_TIME;

	if (tm->ltc.year < 1990) tm->ltc.year += 100; // y2k
	if (tm->ltc.frame == PCL_FRAME_UNKNOWN) {
		tm->ltc.valid = PCL_RTVAL_INVALID;
		UNLOCK_RETURN(hdl, pclGetRealtime, PCL_NO_NEWD);
	}

	UNLOCK_RETURN(hdl, pclGetRealtime, PCL_OK);
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclSetRealtimeMode(	// @103, $1F, set realtime mode
	int hdl,						// in:	handle from pclOpen()
	int mode						// in:	mode for realtime reader
) {
	int result;
#	define pclSetRealtimeModeMulti pclMultiNotSupported
	PCI_GET_HANDLE_LOCK(hdl, pclSetRealtimeMode, (), hPCI, card);

	PCI_WriteCMDD0(hPCI, mode);
	APPEND_DEBUG_LOG(hdl, "  CMD Data: %02X", mode);
	if ((result = pclPciExec(hdl, CMD_SET_REALTIME)) != PCL_OK) UNLOCK_RETURN(hdl, pclSetRealtimeMode, result);
#ifdef PCL_WAKEUP_PERSISTENT
	gPciCard.v[card].persist.setRealtimeMode.use = TRUE;
	gPciCard.v[card].persist.setRealtimeMode.mode = mode;
#endif
	UNLOCK_RETURN(hdl, pclSetRealtimeMode, PCL_OK);
}

//---------------------------------------------------------------------------
/*
int _PCL_FDEF pclGetRealtimeMonitor(
//	int hdl,
	PCL_REALTIME_MONITOR _PCL_OUT *rtm	// out:	realtime monitor data from pcl
) {
	CHECK_PTR(rtm);
	GetSharedMem(rtm);
	return(PCL_OK);
}
*/
#endif

//---------------------------------------------------------------------------

BOOL pclPciLock(int hdl)
{
	assert(pgPciHdl);
	int card = pgPciHdl->v[hdl].card;
	CHECK_CARD(card);
	APPEND_DEBUG_LOG(hdl, "> pclPciLock(hdl:%d)", hdl);

#ifdef LINUX
#ifdef PCL_USE_SHM_MUTEX
	{
		unsigned int toutMs = 2000; // Timeout
		struct timeval tstart, tstop, tdiff;
		int result = 0;
		gettimeofday(&tstart, NULL);

		while ((result = pthread_mutex_trylock(&gPciCardLocal.v[card].pMutex -> mutex))) {
			if (result == EOWNERDEAD || result == -EOWNERDEAD) {
				APPEND_DEBUG_LOG(-1, "Mutex owner died. Recovering.");
				result = pthread_mutex_consistent(&gPciCardLocal.v[card].pMutex -> mutex);
				break;
			} // end if
			gettimeofday(&tstop, NULL);
			timersub(&tstop, &tstart, &tdiff);
			unsigned int msec = tdiff.tv_sec * 1000 + tdiff.tv_usec / 1000;
			if (msec >= toutMs) break;
		} // end while
		if (result != 0) return FALSE;

//		pthread_mutex_lock(&gPciCardLocal.v[card].pMutex -> mutex);
	}
#elif defined(PCL_USE_SYSV_SEM)
	int suc = lockTimeout(gSemId, card + SEM_FIRST_CARD, 2, 0);
	if (suc) return FALSE;
#endif
#endif // LINUX

#ifdef PCL_FAST_MUTEX
//	assert(gPciCard.v[card].mutex);
	if (!gPciCardLocal.v[card].mutex.enter(PCL_TIMEOUT)) {
		APPEND_DEBUG_LOG(hdl, "< pclLock(hdl:%d) = FALSE", hdl);
		return(FALSE);
	}
#endif // PCL_FAST_MUTEX
#ifdef PCL_METERED_SECTION
	DWORD result = EnterMeteredSection(
		pgPciHdl->v[hdl].section,	// Pointer to a metered section
		METERED_SECTION_TIMEOUT	// Time-out interval in milliseconds
	);
#endif // PCL_METERED_SECTION

	APPEND_DEBUG_LOG(hdl, "  pclPciLock(hdl:%d)", hdl);

	assert(!gPciCardLocal.v[card].locked);

#ifdef PCL_METERED_SECTION
	switch (result) {
	default:
		assert(FALSE);
		// no break;
	case WAIT_TIMEOUT:
		return(FALSE);

	case WAIT_OBJECT_0:
		break;
	}
#endif // PCL_METERED_SECTION

	gPciCardLocal.v[card].locked = TRUE;
	APPEND_DEBUG_LOG(hdl, "< pclPciLock(hdl:%d) = TRUE", hdl);
	return(TRUE);
}

//---------------------------------------------------------------------------

void pclPciUnlock(int hdl)
{
	assert(pgPciHdl);
	int card = pgPciHdl->v[hdl].card;
	CHECK_CARD(card);
	assert(gPciCardLocal.v[card].locked);
	gPciCardLocal.v[card].locked = FALSE;

	APPEND_DEBUG_LOG(hdl, "> pclPciUnlock(hdl:%d)", hdl);

#ifdef PCL_METERED_SECTION
	BOOL result = LeaveMeteredSection(
		pgPciHdl->v[hdl].section,	// Pointer to a metered section
		1,						// Amount to add to current count
		NULL					// Address of previous count
	);
#endif // PCL_METERED_SECTION
#ifdef PCL_FAST_MUTEX
//	assert(gPciCard.v[card]._mutex);
	gPciCardLocal.v[card].mutex.leave();
#endif // PCL_FAST_MUTEX

#ifdef LINUX
#if defined(PCL_USE_SHM_MUTEX)
	pthread_mutex_unlock(&gPciCardLocal.v[card].pMutex -> mutex);
#elif defined(PCL_USE_SYSV_SEM)
	unlock(gSemId, card + SEM_FIRST_CARD);
#endif
#endif // LINUX
	APPEND_DEBUG_LOG(hdl, "< pclPciUnlock(hdl:%d)", hdl);
}

//---------------------------------------------------------------------------

#if WD_VER < 600
void PCI_IntHandlerRoutine(
	PCI_HANDLE hPCI,			// in:	handle from pclPciOpen()
	PCI_INT_RESULT *pciResult
)
#else
void PCI_IntHandlerRoutine(
	WDC_DEVICE_HANDLE hDev,		// in:	handle
	PCI_INT_RESULT *pciResult
)
#endif
{
	int card;
#if WD_VER < 600
	APPEND_DEBUG_LOG(0, "> --- PCI_IntHandlerRoutine(hDev:%p, pciResult:%p, *pciResult:%d)", hPCI, pciResult, *pciResult);
#else
	APPEND_DEBUG_LOG(0, "> --- PCI_IntHandlerRoutine(hDev:%p, pciResult:%p, *pciResult:0x%x)", hDev, pciResult, *pciResult);
#endif
//	if (!pciResult)//NEVER
	for (card = 0; card < PCL_CARDS; card++) {
		if (
#ifdef LINUX
			gPciCard.v[card].hPCI == 0
#else
			gPciCard.v[card].hPCI == NULL
#endif
		) continue;	// card not in use

		if (
#if WD_VER < 600
			gPciCard.v[card].hPCI != hPCI
#else
			gPciCard.v[card].hDev != hDev
#endif
		) continue;	// interrupt was requested by another card

		for (;;) {
			int cardFlags = PCI_ReadINTFLGL(gPciCard.v[card].hPCI);
			if (gPciCard.v[card].irq16) cardFlags |= PCI_ReadINTFLGH(gPciCard.v[card].hPCI) << 8;

			if (!cardFlags)
				break;

			int ackHdl = -1;
			int ack = 0;

			int hdl;
			assert(pgPciHdl);
			for (hdl = 0; hdl < PCL_HANDLES; hdl++) {
				TPciIntHandler intHandler = pgPciHdl->v[hdl].intHandler;
				if (intHandler.p == NULL
				||	card != pgPciHdl->v[hdl].card)
				{
					continue;
				}

				int flags = cardFlags & pgPciHdl->v[hdl].imask;
#ifdef PCL_OPT_ANC				/* read user defined ancillary data */
				int flagsAnc = cardFlags & pgPciHdl->v[hdl].anc.imask;
				if (flagsAnc & (PCL_IMASK_V1 | PCL_IMASK_V2)) {
					pgPciHdl->v[hdl].anc.index = 0;
				}
#endif
				if (flags) {
					APPEND_DEBUG_LOG(hdl, "  PCI_IntHandlerRoutine hdl:%d, cardFlags:0x%04X, imask:0x%04X)", hdl, cardFlags, pgPciHdl->v[hdl].imask);

					int mask;
					for (mask = 0x8000; mask != 0; mask >>= 1) {
						if ((flags & mask) == 0)
							continue;

						PCL_INT_RESULT intResult;

						intResult.counter = pciResult->dwCounter;
						intResult.lost    = pciResult->dwLost;
#if WD_VER < 600
						intResult.stopped = pciResult->fStopped;
#else
						intResult.stopped = pciResult->waitResult != INTERRUPT_RECEIVED;
#endif
						ASSERT_STATIC(sizeof(intResult) == sizeof(PCL_INT_RESULT));
						ASSERT_STATIC(sizeof(intHandler.p) == sizeof(intHandler.f));
						assert(intHandler.p != NULL);
						APPEND_DEBUG_LOG(hdl, "  PCI_IntHandlerRoutine -> intHandler(hdl:%d, mask:0x%04X, intResult:%d)", hdl, mask, intResult);
						intHandler.f(hdl, mask, &intResult);
						APPEND_DEBUG_LOG(hdl, "  PCI_IntHandlerRoutine <- intHandler(hdl:%d, mask:0x%04X, intResult:%d)", hdl, mask, intResult);
					}
				}
				ack |= flags;
#ifdef PCL_OPT_ANC				/* read user defined ancillary data */
				ack |= flagsAnc;
#endif
				ackHdl = hdl;
				APPEND_DEBUG_LOG(hdl, "  PCI_IntHandlerRoutine ackHdl:%d ack:0x%04X flags:0x%04X", ackHdl, ack, flags);
			} // for hdl
			if (!ack) break;
			pclIntAck(ackHdl, ack);
		} // for (;;)

		break;	// card was found and handled, work is completed
	} // for card
#if WD_VER < 600
	APPEND_DEBUG_LOG(0, "< --- PCI_IntHandlerRoutine(hDev:%p, pciResult:%p, *pciResult:%d)", hPCI, pciResult, *pciResult);
#else
	APPEND_DEBUG_LOG(0, "< --- PCI_IntHandlerRoutine(hDev:%p, pciResult:%p, *pciResult:0x%x)", hDev, pciResult, *pciResult);
#endif
}

//---------------------------------------------------------------------------

int _PCL_FDEF pclIntEnable(	// enable or disable interrupts
	int hdl,				// in:	handle from pclPciOpen()
	BOOL enable				// in:	enable or disable interrupt
) {
	APPEND_DEBUG_LOG(hdl, "> pclIntEnable(hdl:%d, enable:%d)", hdl, enable);

#	define pclIntEnableMulti pclMultiNotSupported
	PCI_GET_HANDLE(hdl, pclIntEnable, (), hPCI, card);
//	assert(gPciCard.v[card].locked);

	if (enable) {
		if (!PCI_IntIsEnabled(hPCI)) {
			PCI_ERR("");
			if (!IntEnable(
				hdl,
				PCI_IntHandlerRoutine
			)) {
				if (!PCI_ErrorString[0])
					PCI_ERR("pclIntEnable(): Enable of Interrupt failed\n");
				APPEND_DEBUG_LOG(hdl, "< pclIntEnable() PCL_ERR_IOCTL");
				return(PCL_ERR_IOCTL);
			}
#if WD_VER < 600
//			assert(INTCSR_ACK_9030 == hPCI->Int.Trans[0].Data.Dword);
#endif
			PCI_WriteEnable(hdl);
		}
	} else {
		if (PCI_IntIsEnabled(hPCI)) {
			PCI_WriteDisable(hdl);
			if (PCI_IntIsEnabled(hPCI)) {
				PCI_IntDisable(hPCI);
			}
		}
	}
	APPEND_DEBUG_LOG(hdl, "< pclIntEnable()");
	return(PCL_OK);
}

//---------------------------------------------------------------------------

BOOL _PCL_FDEF pclPciExec(
	int hdl,
	BYTE cmd
) {
	BYTE ack, cmdr;
#ifdef LINUX
	struct timeval tstart, tstop, tdiff;
#else
	DWORD ticks;
#endif
	unsigned int msec;

	assert(pgPciHdl);
	int card = pgPciHdl->v[hdl].card;
	CHECK_CARD(card);

	assert(card >= 0 && card < NELT(gPciCard.v));
	PCI_HANDLE hPCI = gPciCard.v[card].hPCI;
	assert(gPciCardLocal.v[card].locked);

	APPEND_DEBUG_LOG(hdl, "> pclPciExec(hdl:%d, cmd:$%02X)", hdl, cmd);

#ifdef LINUX
	assert(hPCI != 0);
#else
	assert(hPCI != NULL);
#endif
	ack = PCI_ReadACK(hPCI);
	PCI_WriteCMDW(hPCI, cmd);

#ifdef LINUX
	gettimeofday(&tstart, NULL);
#else
	ticks = GetTickCount();
#endif

	while (ack == PCI_ReadACK(hPCI)) {
		_MM_PAUSE;
#ifdef LINUX
		gettimeofday(&tstop, NULL);
		timersub(&tstop, &tstart, &tdiff);
		msec = tdiff.tv_sec * 1000 + tdiff.tv_usec / 1000;
#else
		msec = (GetTickCount() - ticks);
#endif // LINUX
#ifndef NDEBUG
		if (cmd == CMD_SET_ANC_DID) {
			cmd = cmd;
		}
#endif
		if (msec > PCL_TIMEOUT) {
			if (ack != PCI_ReadACK(hPCI)) break; // last chance
			PCI_ERR("pclPciExec(): Sent CMD 0x%02X, no ACK received\n", cmd);
			APPEND_DEBUG_LOG(hdl, "< pclPciExec(hdl:%d, cmd:$%02X) = TIMEOUT", hdl, cmd);
			return(PCL_ERR_TIMEOUT);
		} // end if
//		Sleep(0);
	}
	if ((cmdr = PCI_ReadCMDR(hPCI)) != cmd) {
		PCI_ERR("pclPciExec(): Sent CMD 0x%02X, received CMDR 0x%02X\n", cmd, cmdr);
		APPEND_DEBUG_LOG(hdl, "< pclPciExec(hdl:%d, cmd:$%02X) = UNKNOWN CMD", hdl, cmd);
		return(PCL_ERR_UK_CMD);
	}
	APPEND_DEBUG_LOG(hdl, "< pclPciExec(hdl:%d, cmd:$%02X) - ack:%02X - data:%02X %02X  %02X %02X %02X %02X  %02X  %02X %02X %02X %02X",
		hdl, cmd, ack,
		PCI_ReadDATAA(hPCI),
		PCI_ReadDATA9(hPCI),
		PCI_ReadDATA8(hPCI),
		PCI_ReadDATA7(hPCI),
		PCI_ReadDATA6(hPCI),
		PCI_ReadDATA5(hPCI), PCI_ReadDATA4(hPCI),
		PCI_ReadDATA3(hPCI), PCI_ReadDATA2(hPCI), PCI_ReadDATA1(hPCI), PCI_ReadDATA0(hPCI));
	return(PCL_OK);
}

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int _PCL_FDEF pclMultiIgnored()
{
	return(PCL_OK);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------

#ifdef PCL_MULTI_MODE

int _PCL_FDEF pclMultiNotSupported()
{
	assert(FALSE);
	return(PCL_ERR_MULTI_NOT_SUPP);
}

#endif // PCL_MULTI_MODE

//---------------------------------------------------------------------------
