/**********************************************\
*                                              *
*  Linux Kernel Driver for PCL PCI(e) Devices  *
*                                              *
*        Part of the shared library            *
*            (WinDriver-Clone)                 *
*                                              *
\**********************************************/

// $Id: pci_lib_linux.c 2123 2023-10-05 14:32:37Z J.Loh $
// Version: 5.45 (c) 1995-2023 PLURA Europe GmbH

#include "params.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
//#include <stdlib.h> // exit()
#include <unistd.h> // usleep
#include <strings.h>
#include <assert.h>
#include <signal.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/file.h>	// flock
//#include <sys/types.h>
//#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


#include "pci_lib_linux.h"
#include "avpcl.h"
#include "avpcldrv.h"
#include "funcs.h"

#ifndef PATHTODEVICEFILES
#error Path not defined. Please do so in the Makefile.
#endif
//---------------------------------------------------------------------------

// error string
char PCI_ErrorString[1024];

const char *pclErrorString;

// Int result
PCI_INT_RESULT gIntResult = { 0, 0, FALSE };

//---------------------------------------------------------------------------

int sprintf_s(
   char *buffer,
   size_t sizeOfBuffer,
   const char *format, ... 
) {
	int result;

	va_list args;
	va_start(args, format);
	result = vsnprintf(buffer, sizeOfBuffer, format, args);
	va_end(args);

	return result;
}

//---------------------------------------------------------------------------

void PCI_ERR(const char *format, ...)
{
	va_list args;

	va_start(args, format);
	vsnprintf(PCI_ErrorString, sizeof(PCI_ErrorString), format, args);
	va_end(args);
}

//---------------------------------------------------------------------------

void Sleep(unsigned long timeout)
{
	usleep(timeout * 1000);
}

//---------------------------------------------------------------------------

// 0x00  r  Data 0
BYTE PCI_ReadDATA0 (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATA0, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x01  r  Data 1
BYTE PCI_ReadDATA1 (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATA1, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x02  r  Data 2
BYTE PCI_ReadDATA2 (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATA2, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x03  r  Data 3
BYTE PCI_ReadDATA3 (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATA3, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x04  r  Data 4
BYTE PCI_ReadDATA4 (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATA4, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x05  r  Data 5
BYTE PCI_ReadDATA5 (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATA5, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x06  r  Data 6
BYTE PCI_ReadDATA6 (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATA6, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x07  r  Data 7
BYTE PCI_ReadDATA7 (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATA7, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x08  r  Data 8
BYTE PCI_ReadDATA8 (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATA8, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x09  r  Data 9
BYTE PCI_ReadDATA9 (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATA9, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x0A  r  Data A (MSB)
BYTE PCI_ReadDATAA (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READDATAA, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x0C  r  Command Response
BYTE PCI_ReadCMDR (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READCMDR, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x0D  r  Acknowledge
BYTE PCI_ReadACK (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READACK, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

// 0x0E w Command
void PCI_WriteCMDW (PCI_HANDLE hPCI, BYTE data)
{
	ioctl(hPCI, AVPCL_IOC_WRITECMD, &data);
}

//---------------------------------------------------------------------------

// 0x0F w Command Data 0
void PCI_WriteCMDD0 (PCI_HANDLE hPCI, BYTE data)
{
	ioctl(hPCI, AVPCL_IOC_WRITECMDD0, &data);
}

//---------------------------------------------------------------------------

// 0x10 w Command Data 1
void PCI_WriteCMDD1 (PCI_HANDLE hPCI, BYTE data)
{
	ioctl(hPCI, AVPCL_IOC_WRITECMDD1, &data);
}

//---------------------------------------------------------------------------

// 0x11 w Config
void PCI_WriteCFG (PCI_HANDLE hPCI, BYTE data)
{
	ioctl(hPCI, AVPCL_IOC_WRITECFG, &data);
}

//---------------------------------------------------------------------------

void PCI_WriteINT_ACK (PCI_HANDLE hPCI)
{
	ioctl(hPCI, AVPCL_IOC_WRITEINTACK, 0);
}

//---------------------------------------------------------------------------

void PCI_WriteINTCSR (PCI_HANDLE hPCI, DWORD data)
{
	ioctl(hPCI, AVPCL_IOC_WRITEINTCSR, &data);
}

//---------------------------------------------------------------------------
// 0x1F rw Version Number
BYTE PCI_ReadVERSION (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READVERSION, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

BYTE PCI_ReadINTFLGL (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READINTFLGL, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

BYTE PCI_ReadINTFLGH (PCI_HANDLE hPCI)
{
	BYTE result;
	if (ioctl(hPCI, AVPCL_IOC_READINTFLGH, &result) == 0) {
		return result;
	} // end if
	else return PCL_ERR_IOCTL;
}

//---------------------------------------------------------------------------

BOOL PCI_IntIsEnabled (PCI_HANDLE hPCI)
{
	int card = 0;
	for (card = 0; card < PCL_CARDS; card++) {
		if (gPciCard.v[card].hPCI == hPCI) {
//			APPEND_DEBUG_LOG(0, "IsEnable: card:%i hIntThread:%p", card, gPciCard.v[card].hIntThread);
			if (gPciCard.v[card].hIntThread != NULL) {
				return TRUE;
			} else {
				return FALSE;
			} // end else
		} // end if
	} // end for
	return FALSE;
}

//---------------------------------------------------------------------------

int PCI_GetDeviceId(PCI_HANDLE hPCI, WORD _PCL_OUT *result) {
	if (ioctl(hPCI, AVPCL_RETR_DEVID, result) == 0) {
		return PCL_OK;
	} else {
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_GetDeviceId()

//---------------------------------------------------------------------------

int PCI_GetSubsystemDeviceId(PCI_HANDLE hPCI, WORD _PCL_OUT *result) {
	if (ioctl(hPCI, AVPCL_RETR_SUBSYSDEVID, result) == 0) {
		return PCL_OK;
	} else {
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_GetSubsystemDeviceId()

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_GetBus(PCI_HANDLE hPCI, BYTE _PCL_OUT *bus) {
	if (ioctl(hPCI, AVPCL_RETR_BUS, bus) == 0) {
		return PCL_OK;
	} else {
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_GetBus()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_GetDeviceInfo(PCI_HANDLE hPCI, TDeviceInfo _PCL_OUT *deviceInfo) {
	if (ioctl(hPCI, AVPCL_RETR_DEVINFO, deviceInfo) == 0) {
		return PCL_OK;
	} else {
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_GetDeviceInfo()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_GetDeviceInfoBridge(PCI_HANDLE hPCI, TDeviceInfo _PCL_OUT *deviceInfo) {
	if (ioctl(hPCI, AVPCL_BRIDGE_RETR_DEVINFO, deviceInfo) == 0) {
		return PCL_OK;
	} else {
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_GetDeviceInfoBridge()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_GetDeviceIdBridge(PCI_HANDLE hPCI, WORD _PCL_OUT *result) {
	if (ioctl(hPCI, AVPCL_BRIDGE_RETR_DEVID, result) == 0) {
		return PCL_OK;
	} else {
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_GetDeviceIdBridge()
#endif

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_ReadOffset8 (PCI_HANDLE hPCI, BYTE bar, UINT offset, BYTE _PCL_OUT *result) {
	avpcl_read_bar_t readBar = { bar, offset, 1, result };

	int res = ioctl(hPCI, AVPCL_READ_BAR, &readBar);
	if (res == 0) {
		return PCL_OK;
	} else {
		perror("PCI_ReadOffset8");
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_ReadOffset8()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_ReadOffset8Bridge (PCI_HANDLE hPCI, BYTE bar, UINT offset, BYTE _PCL_OUT *result) {
	avpcl_read_bar_t readBar = { bar, offset, 1, result };

	int res = ioctl(hPCI, AVPCL_BRIDGE_READ_BAR, &readBar);
	if (res == 0) {
		return PCL_OK;
	} else {
		perror("PCI_ReadOffset8Bridge");
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_ReadOffset8Bridge()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_ReadOffset32(PCI_HANDLE hPCI, BYTE bar, UINT offset, DWORD _PCL_OUT *result) {
	avpcl_read_bar_t readBar = { bar, offset, 4, result };

	int res = ioctl(hPCI, AVPCL_READ_BAR, &readBar);
	if (res == 0) {
		return PCL_OK;
	} else {
		perror("PCI_ReadOffset32");
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_ReadOffset32()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_ReadOffset32Bridge(PCI_HANDLE hPCI, BYTE bar, UINT offset, DWORD _PCL_OUT *result) {
	avpcl_read_bar_t readBar = { bar, offset, 4, result };

	int res = ioctl(hPCI, AVPCL_BRIDGE_READ_BAR, &readBar);
	if (res == 0) {
		return PCL_OK;
	} else {
		perror("PCI_ReadOffset32Bridge");
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_ReadOffset32Bridge()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_WriteOffset8 (PCI_HANDLE hPCI, BYTE bar, UINT offset, BYTE data) {
	avpcl_write_bar_t writeBar = { bar, offset, 1, &data };

	int res = ioctl(hPCI, AVPCL_WRITE_BAR, &writeBar);
	if (res == 0) {
		return PCL_OK;
	} else {
		perror("PCI_WriteOffset8");
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_WriteOffset8()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_WriteOffset8Bridge(PCI_HANDLE hPCI, BYTE bar, UINT offset, BYTE data) {
	avpcl_write_bar_t writeBar = { bar, offset, 1, &data };

	int res = ioctl(hPCI, AVPCL_BRIDGE_WRITE_BAR, &writeBar);
	if (res == 0) {
		return PCL_OK;
	} else {
		perror("PCI_WriteOffset8Bridge");
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_WriteOffset8Bridge()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_WriteOffset32(PCI_HANDLE hPCI, BYTE bar, UINT offset, DWORD data) {
	avpcl_write_bar_t writeBar = { bar, offset, 4, &data };

	int res = ioctl(hPCI, AVPCL_WRITE_BAR, &writeBar);
	if (res == 0) {
		return PCL_OK;
	} else {
		perror("PCI_WriteOffset32");
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_WriteOffset32()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_WriteOffset32Bridge(PCI_HANDLE hPCI, BYTE bar, UINT offset, DWORD data) {
	avpcl_write_bar_t writeBar = { bar, offset, 4, &data };

	int res = ioctl(hPCI, AVPCL_BRIDGE_WRITE_BAR, &writeBar);
	if (res == 0) {
		return PCL_OK;
	} else {
		perror("PCI_WriteOffset32Bridge");
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_WriteOffset32Bridge()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_ConfigDump(PCI_HANDLE hPCI, UINT offset, UINT size, void _PCL_OUT *result) {
	avpcl_read_config_t readConfig;
	readConfig.offset = offset;
	readConfig.size = size;
	readConfig.result = result;
	int res = ioctl(hPCI, AVPCL_READ_CONFIG, &readConfig);
	if (res == 0) {
		return PCL_OK;
	} else {
		perror("PCI_ConfigDump");
//		printf("PCI_ConfigDump: res=%i\n", res);
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_ConfigDump()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

#ifdef PCL_OPT_FLASH
int PCI_ConfigDumpBridge(PCI_HANDLE hPCI, UINT offset, UINT size, void _PCL_OUT *result) {
	avpcl_read_config_t readConfig;
	readConfig.offset = offset;
	readConfig.size = size;
	readConfig.result = result;
	int res = ioctl(hPCI, AVPCL_BRIDGE_READ_CONFIG, &readConfig);
	if (res == 0) {
		return PCL_OK;
	} else {
		perror("PCI_ConfigDumpBridge");
//		printf("PCI_ConfigDumpBridge: res=%i\n", res);
		return PCL_ERR_IOCTL;
	} // end else
} // end PCI_ConfigDump()
#endif // PCL_OPT_FLASH

//---------------------------------------------------------------------------

BOOL PCI_IsBridge(PCI_HANDLE hPCI) {
	return !ioctl(hPCI, AVPCL_TEST_BRIDGE, NULL);
} // end PCI_IsBridge()

//---------------------------------------------------------------------------

const char *PCL_CardFilename(int card) {
	switch (card) {
		case 0: return PATHTODEVICEFILES"avpcl0";
		case 1: return PATHTODEVICEFILES"avpcl1";
		case 2: return PATHTODEVICEFILES"avpcl2";
		case 3: return PATHTODEVICEFILES"avpcl3";
		case 4: return PATHTODEVICEFILES"avpcl4";
		case 5: return PATHTODEVICEFILES"avpcl5";
		case 6: return PATHTODEVICEFILES"avpcl6";
		case 7: return PATHTODEVICEFILES"avpcl7";
		case 8: return PATHTODEVICEFILES"avpcl8";
		case 9: return PATHTODEVICEFILES"avpcl9";
		default: return "";
	} // end switch
}

//---------------------------------------------------------------------------

void PCI_IntDisable (PCI_HANDLE hPCI)
{
#ifdef USE_SIGNAL
	PCI_UnRegisterInt(hPCI);
#else
#endif
	int card = 0;
	for (card = 0; card < PCL_CARDS; card++) {
		if (gPciCard.v[card].hPCI == hPCI) {
			flock(hPCI, LOCK_UN | LOCK_NB);
			gPciCard.v[card].hIntThread = NULL;
//			APPEND_DEBUG_LOG(0, "IntDisable: card:%i hIntThread:%p", card, gPciCard.v[card].hIntThread);
			break;
		} // end if
	} // end for
}

//---------------------------------------------------------------------------

void PCI_WriteEnable(int hdl)
{
	APPEND_DEBUG_LOG(hdl, "> PCI_WriteEnable(hdl:%d)", hdl);
	assert(pgPciHdl);
	int card = pgPciHdl->v[hdl].card;
	CHECK_CARD(card);
	PCI_HANDLE hPCI = gPciCard.v[card].hPCI;

	switch (gPciCard.v[card].wDeviceID) {
	case PCI_DEFAULT_DEVICE_ID_PCIE:
		PCI_WriteINTCSR(hPCI, INTCSR_LINTI1_ENABLE_9056 | INTCSR_LINTI1_ENABLE_9056_M);
		break;
	case PCI_DEFAULT_DEVICE_ID_5V:
	case PCI_DEFAULT_DEVICE_ID_3V:
		PCI_WriteINTCSR(hPCI, INTCSR_ACK_9030 | INTCSR_LINTI1_ENABLE_9030_M);
		break;
	default:
		DEBUG_PRINTF("Error: PCI_WriteEnable: DeviceID:%i\n", gPciCard.v[card].wDeviceID);
		assert(	gPciCard.v[card].wDeviceID == PCI_DEFAULT_DEVICE_ID_PCIE ||
				gPciCard.v[card].wDeviceID == PCI_DEFAULT_DEVICE_ID_5V ||
				gPciCard.v[card].wDeviceID == PCI_DEFAULT_DEVICE_ID_3V);
		break;
	}
	APPEND_DEBUG_LOG(hdl, "< PCI_WriteEnable()");
}

//---------------------------------------------------------------------------

void PCI_WriteDisable(int hdl)
{
	assert(pgPciHdl);
	int card = pgPciHdl->v[hdl].card;
	PCI_HANDLE hPCI = gPciCard.v[card].hPCI;

	switch (gPciCard.v[card].wDeviceID) {
	case PCI_DEFAULT_DEVICE_ID_PCIE:
		PCI_WriteINTCSR(hPCI, INTCSR_LINTI1_ENABLE_9056 & ~INTCSR_LINTI1_ENABLE_9056_M);
		break;
	default:	assert(FALSE);
	case PCI_DEFAULT_DEVICE_ID_5V:
	case PCI_DEFAULT_DEVICE_ID_3V:
		PCI_WriteINTCSR(hPCI, INTCSR_ACK_9030 & ~INTCSR_LINTI1_ENABLE_9030_M);
		break;
	}
}

//---------------------------------------------------------------------------

#ifdef USE_SIGNAL
void SignalHandler(int signum, siginfo_t *info, void *unused) {
//	assert(FALSE);
}
#else
//---------------------------------------------------------------------------

void *interruptTask(void *arg) {
	long card = (long)arg;
	assert(card < NELT(gPciCard.v));

	// open file
	int fd = -1;
	assert(0 <= card && card <= 9);
	fd = open(PCL_CardFilename(card), O_RDONLY);

	if (fd < 0) {
		assert(FALSE);
		return 0;
	} // end if

	char buf[2];
	memset(buf, 0, sizeof(buf));

	APPEND_DEBUG_LOG(0, "Interrupt Task for card %i started.", card);
	while (gPciCard.v[card].hIntThread != NULL) {
		if (read(fd, buf, sizeof(buf)) == 0) {
			APPEND_DEBUG_LOG(0, "Interrupt Task for card %i: read.", card);
			if (!gPciCard.v[card].hIntThread) break;
			gIntResult.dwCounter++;
			APPEND_DEBUG_LOG(0, "Interrupt Task for card %i: exec.", card);
			gPciCard.v[card].hIntThread(gPciCard.v[card].hPCI, &gIntResult);
			APPEND_DEBUG_LOG(0, "Interrupt Task for card %i: return.", card);
		} else {
			break;
		} // end else
	} // end while

	APPEND_DEBUG_LOG(0, "Interrupt Task for card %i finished.", card);
	close(fd);
	pthread_exit(NULL);
	return 0;
} // end interruptTask()

#endif

//---------------------------------------------------------------------------

BOOL IntEnable (int hdl, PCI_INT_HANDLER funcIntHandler)
{
	assert(pgPciHdl);
	long card = pgPciHdl->v[hdl].card;
	PCI_HANDLE hPCI = gPciCard.v[card].hPCI;
//	PCI_ADDR addrSpace;

	// Check if interrupt is already enabled
	if (PCI_IntIsEnabled (hPCI)) {
		pclErrorString = "Interrupt is already enabled.";
		return FALSE;
	} // end if

	// test if card is locked
	int res = flock(hPCI, LOCK_EX | LOCK_NB);
	if (res) {
		switch (errno) {
			case EBADF:
				PCI_ERR("IntEnable: Bad file descriptor.");
				return FALSE;
			case EWOULDBLOCK:
				PCI_ERR("IntEnable: Card locked by another process.");
				return FALSE;
			default:
				PCI_ERR("IntEnable: Unable to lock card (%i).", res);
				return FALSE;
		} // end switch
	} // end if

	gPciCard.v[card].hIntThread = funcIntHandler;
//	APPEND_DEBUG_LOG(0, "IntEnable: card:%i hIntThread:%p", card, gPciCard.v[card].hIntThread);

	// Start thread if not done already. pthread_kill(id, 0) checks if the thread exists
	if (gPciCard.v[card].intThreadId == 0 || pthread_kill(gPciCard.v[card].intThreadId, 0) == ESRCH)
		pthread_create(&gPciCard.v[card].intThreadId, NULL, &interruptTask, (void*)card);

#ifdef USE_SIGNAL
	// Register for signal action
	struct sigaction act;
//	act.sa_handler = SignalHandler;
	act.sa_sigaction = SignalHandler;
	act.sa_flags = SA_SIGINFO;
	sigaction(AV_SIGNAL, &act, NULL);
//	DEBUG_PRINTF("My PID:%i\n", getpid());

	PCI_RegisterInt(hPCI); // tell pid to driver
#else
#endif

	return TRUE;
}

//---------------------------------------------------------------------------
