/***************************************************************************\
*																			*
*	Demo and Test Program for PCL PCI(e) cards								*
*																			*
*	Copyright (c) 1995-2023 by PLURA Europe GmbH							*
*																			*
*	This program reads several data from the PCL PCI(e) in interrupt mode	*
*																			*
\***************************************************************************/

// $Id: IntTest.cpp 2123 2023-10-05 14:32:37Z J.Loh $

#define VER_S	"5.45"	// Version

//---------------------------------------------------------------------------

#ifdef _WIN32
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <ctype.h>

#elif defined(LINUX)

// Linux compile command:
// gcc -o IntTest IntTest.cpp -DLINUX -L. -lavpcl32 -lncurses

#include <curses.h>
#include <stdlib.h>
//#include <assert.h>

#ifndef __USE_BSD
#	define __USE_BSD
#	include <string.h>
#	include <unistd.h>
#	undef __USE_BSD
#else
#	include <string.h>
#	include <unistd.h>
#endif

#endif // defined(LINUX)

#include "avpcl.h"
#undef	PCL_OPT_GPI_IN_EVNT		/* gpi input events */

//---------------------------------------------------------------------------

#if _MSC_VER >= 1400 // MSVC 2005/2008
#	define sscanf sscanf_s
#endif

//---------------------------------------------------------------------------

#ifdef LINUX
#	define printfw printw
#	define _strdup strdup
#	define _getch getch
#else
#	define printfw printf
#endif

//---------------------------------------------------------------------------

void gotoxy(int x, int y) // set cursor to new position of the screen
{
#ifdef _WIN32
	HANDLE hStdout;
	COORD coordCursor; // new cursor position coordinates

	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);

	coordCursor.X = (short) (x - 1);
	coordCursor.Y = (short) (y - 1);

	SetConsoleCursorPosition(
		hStdout, // handle of console screen buffer
		coordCursor // new cursor position coordinates
	);
#elif defined(LINUX)
	move(y-1, x-1);
#endif
}

//---------------------------------------------------------------------------

void clrscr(void) // clear screen
{
#ifdef _WIN32
	DWORD dummy;
	COORD Home = { 0, 0 };
	HANDLE hStdout;
	CONSOLE_SCREEN_BUFFER_INFO csbiInfo;	// Console information

	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo( hStdout, &csbiInfo );

	FillConsoleOutputCharacter(
		hStdout, ' ', csbiInfo.dwSize.X * csbiInfo.dwSize.Y,
		Home, &dummy
	);
#elif defined(LINUX)
	initscr();		// start ncurses mode
	refresh();
#endif
}


//---------------------------------------------------------------------------

void clreol(void) // clear from cursor position to end of line
{
#ifdef _WIN32
	DWORD dummy;
	HANDLE hStdout;
	CONSOLE_SCREEN_BUFFER_INFO csbiInfo;	// Console information

	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo( hStdout, &csbiInfo );

	FillConsoleOutputCharacter(
		hStdout, ' ', csbiInfo.dwSize.X - csbiInfo.dwCursorPosition.X,
		csbiInfo.dwCursorPosition, &dummy
	);
#elif defined(LINUX)
	int cx, cy, cols, rows;
	(void)rows;
	getmaxyx(stdscr, rows, cols);
	getyx(stdscr, cy, cx);
	int n;
	for (n = 0; n < cols - cx; n++) {
		printw(" ");
	} // end for
	move(cy, cx);
#endif
}

//---------------------------------------------------------------------------

void cursorVisible(BOOL show) // hide or show cursor
{
#ifdef _WIN32
	HANDLE hStdout;
	CONSOLE_CURSOR_INFO ConsoleCursorInfo;

	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);

	GetConsoleCursorInfo(hStdout, &ConsoleCursorInfo);
	ConsoleCursorInfo.bVisible = show;
	SetConsoleCursorInfo(hStdout, &ConsoleCursorInfo);
#elif defined(LINUX)
	curs_set(show);
#endif
}

//---------------------------------------------------------------------------

#if 0 //def LINUX

void Sleep(long ms)
{
	usleep(ms * 1000);
}

#endif

//---------------------------------------------------------------------------

BOOL legalBase(int base) // is base a legal address of a PCI card?
{
	if (base > PCL_BASE_PCI_OFS && base <= PCL_BASE_PCI_LAST) return(TRUE);
	return(FALSE);
}

//---------------------------------------------------------------------------

const char *err2msg(int result) // convert PCL error number to string
{
	return(
		result == PCL_NO_NEWD       ? "No new data" :
		result == PCL_OK            ? "OK" :
		result == PCL_ERR_INV_PORT  ? "Invalid port id" :
		result == PCL_ERR_NOT_FOUND ? "PCL hardware not found" :
		result == PCL_ERR_NO_HANDLE ? "No free handle found" :
		result == PCL_ERR_INV_HDL   ? "Invalid handle" :
		result == PCL_ERR_NOTOPEN   ? "Handle not opened by pclOpen()" :
		result == PCL_ERR_NULL      ? "Null pointer assignment" :
		result == PCL_ERR_TIMEOUT   ? "Timeout" :
		result == PCL_ERR_INV_REG   ? "Register not available" :
		result == PCL_ERR_INV_BANK  ? "Invalid transmit bank number" :
		result == PCL_ERR_RANGE     ? "Parameter out of range" :
		result == PCL_ERR_INIT      ? "Can't initialize library" :
		result == PCL_ERR_IOCTL     ? "Error in device driver while IOCTL" :
		result == PCL_ERR_VER       ? "Wrong Version of device driver" :
		result == PCL_ERR_NOT_SUPP	? "Function not supported" :
		result == PCL_ERR_UK_CMD	? "Unknown command" :
		result == PCL_ERR_OPEN		? "Can't open device driver" :
									  "(Unknown)"
	);
}

//---------------------------------------------------------------------------

void fatal(				// handle fatal errors
	BOOL fullscreen,	// fullscreen mode
	char *reason,		// string describing failure
	int result			// return code
) {
#ifdef LINUX
	(void)fullscreen;
	cbreak();
#else
	if (fullscreen) gotoxy(1, 15);
#endif
	printfw("Fatal Error while %s: %s (%d)"
	,	reason
	,	err2msg(result)
	,	result
	);
	printfw("\n%s\n"
		"Press any key to continue..."
	,	pclGetErrorString()
	);
	_getch();
#ifdef LINUX
	endwin();
#endif
	exit(1);
}

//---------------------------------------------------------------------------

void warning(			// show warning
	BOOL fullscreen,	// fullscreen mode
	int result			// return code
) {
	char *s;

	if (fullscreen) gotoxy(1, 20);
	printfw("%s", err2msg(result));
	clreol();
	s = _strdup(pclGetErrorString());
	if (s[0] != '\0') s[strlen(s) - 1] = '\0';
	printfw("\n%s", s);
	clreol();
	free(s);
}

//---------------------------------------------------------------------------

void PCL_CALLBACK intHandler(		// interrupt handler
	int hdl,						// handle from pclOpen()
	int flag,						// interrupt flag to handle
	const PCL_INT_RESULT *intResult	// interrupt result
) {
	int result;						// result from pclXxx() functions
	DWORD time, user;				// time code information
	int newd, bits, frame, stat;	// time code information

	if (intResult->stopped) {	// interrupt was disabled while servicing
		return;					// don't handle interrupt
	}

	switch (flag) {	// what interrupt source to handle

	case PCL_IMASK_MIXED:	// handle mixed register interrupt
		if ((result = pclGetRegister(hdl, PCL_REG_MIXED, &time, &bits,
			&user, &frame, &newd)) == PCL_OK
		) {	// get time code from mixed register
			gotoxy(19, 3);
			printfw( // show time and user information
				"%2x:%02x:%02x:%02x  %08x  %s"
			,	HIBYTE(HIWORD(time))
			,	LOBYTE(HIWORD(time))
			,	HIBYTE(LOWORD(time))
			,	LOBYTE(LOWORD(time))
			,	user
			,	newd & PCL_NEWD_VITC ? "VITC" :
				newd & PCL_NEWD_LTC  ? "LTC " :
									   "?   "
			);
		} else {
			warning(TRUE, result);	// error while pclGetRegister()
		}
		break;

	case PCL_IMASK_VITC:	// handle VITC interrupt
		if ((result = pclGetRegister(hdl, PCL_REG_VITC, &time, &bits,
			&user, &frame, &newd)) == PCL_OK
		) {	// get time code from VITC register
			gotoxy(19, 5);
			printfw( // show time and user information
				"%2x:%02x:%02x:%02x  %08x  V%d   Rate:%s"
			,	HIBYTE(HIWORD(time))
			,	LOBYTE(HIWORD(time))
			,	HIBYTE(LOWORD(time))
			,	LOBYTE(LOWORD(time))
			,	user
			,	((bits & (
					frame == PCL_FRAME_25
				?	PCL_BITS_VITC_PF
				:	PCL_BITS_VITC_NF
				)) ? 2 : 1)
			,	frame == PCL_FRAME_24   ? "24  " :
				frame == PCL_FRAME_25   ? "25  " :
				frame == PCL_FRAME_30   ? "30  " :
				frame == PCL_FRAME_30DF ? "30df" :
										  "--  "
			);
		} else {
			warning(TRUE, result);	// error while pclGetRegister()
		}
		break;

	case PCL_IMASK_LTC:	// handle LTC interrupt
		if ((result = pclGetRegister(hdl, PCL_REG_LTC, &time, &bits,
			&user, &frame, &newd)) == PCL_OK
		) {	// get time code from LTC register
			gotoxy(19, 6);
			printfw( // show time and user information
				"%2x:%02x:%02x:%02x  %08x       Rate:%s"
			,	HIBYTE(HIWORD(time))
			,	LOBYTE(HIWORD(time))
			,	HIBYTE(LOWORD(time))
			,	LOBYTE(LOWORD(time))
			,	user
			,	frame == PCL_FRAME_24   ? "24  " :
				frame == PCL_FRAME_25   ? "25  " :
				frame == PCL_FRAME_30   ? "30  " :
				frame == PCL_FRAME_30DF ? "30df" :
										  "--  "
			);
		} else {
			warning(TRUE, result);	// error while pclGetRegister()
		}
		break;

	case PCL_IMASK_ATC_LTC:		// handle ATC LTC interrupt
		if ((result = pclGetRegister(hdl, PCL_REG_ATC_LTC, &time, &bits,
			&user, &frame, &newd)) == PCL_OK
		) {	// get time code from ATC LTC register
			gotoxy(19, 7);
			printfw( // show time and user information
				"%2x:%02x:%02x:%02x  %08x       Rate:%s"
			,	HIBYTE(HIWORD(time))
			,	LOBYTE(HIWORD(time))
			,	HIBYTE(LOWORD(time))
			,	LOBYTE(LOWORD(time))
			,	user
			,	frame == PCL_FRAME_24   ? "24  " :
				frame == PCL_FRAME_25   ? "25  " :
				frame == PCL_FRAME_30   ? "30  " :
				frame == PCL_FRAME_30DF ? "30df" :
										  "--  "
			);
			if ((result = pclGetRegStatus(hdl, PCL_REG_ATC_LTC, &stat)) == PCL_OK) {
				printfw("  DBB1:");
				switch (stat & PCL_STAT_DBB1) {
				case PCL_STAT_DBB1_LTC:		printfw("LTC  ");	break;
				default:					printfw("%02x   ", stat & PCL_STAT_DBB1);	break;
				}
				printfw(
					// show status information
					" %c%c"
				,	stat & PCL_STAT_DBB2_TC_VALIDITY ? 'V' : ' '
				,	stat & PCL_STAT_DBB2_PROCESS_BIT ? 'P' : ' '
				);
			} else {
				warning(TRUE, result);
			}
		} else {
			warning(TRUE, result);	// error while pclGetRegister()
		}
		break;

	case PCL_IMASK_ATC_VITC:	// handle ATC VITC interrupt
		if ((result = pclGetRegister(hdl, PCL_REG_ATC_VITC, &time, &bits,
			&user, &frame, &newd)) == PCL_OK
		) {	// get time code from ATC VITC register
			gotoxy(19, 8);
			printfw( // show time and user information
				"%2x:%02x:%02x:%02x  %08x  V%d   Rate:%s"
			,	HIBYTE(HIWORD(time))
			,	LOBYTE(HIWORD(time))
			,	HIBYTE(LOWORD(time))
			,	LOBYTE(LOWORD(time))
			,	user
			,	((bits & (
					frame == PCL_FRAME_25
				?	PCL_BITS_VITC_PF
				:	PCL_BITS_VITC_NF
				)) ? 2 : 1)
			,	frame == PCL_FRAME_24   ? "24  " :
				frame == PCL_FRAME_25   ? "25  " :
				frame == PCL_FRAME_30   ? "30  " :
				frame == PCL_FRAME_30DF ? "30df" :
										  "--  "
			);
			if ((result = pclGetRegStatus(hdl, PCL_REG_ATC_VITC, &stat)) == PCL_OK) {
				printfw("  DBB1:");
				switch (stat & PCL_STAT_DBB1) {
				case PCL_STAT_DBB1_VITC1:	printfw("VITC1");	break;
				case PCL_STAT_DBB1_VITC2:	printfw("VITC2");	break;
				default:					printfw("%02x   ", stat & PCL_STAT_DBB1);	break;
				}
				printfw(
					// show status information
					" %c%c Line %2d%c"
				,	stat & PCL_STAT_DBB2_TC_VALIDITY ? 'V' : ' '
				,	stat & PCL_STAT_DBB2_PROCESS_BIT ? 'P' : ' '
				,	(stat & PCL_STAT_DBB2_VITC_LINE_SELECT) >> 8
				,	stat & PCL_STAT_DBB2_VITC_LINE_DUP ? 'D' : ' '
				);
			} else {
				warning(TRUE, result);
			}
		} else {
			warning(TRUE, result);	// error while pclGetRegister()
		}
		break;

	case PCL_IMASK_V1:	// handle V1 interrupt (odd video field)
		gotoxy(19, 10);
		printfw("V1  (odd)");	// show video field

		int system, progressive;
		if ((result = pclGetVideoStandard(hdl, &system, &progressive)) == PCL_OK) {
			printfw("  %s %s"
			,	system == PCL_VSYSTEM_UNKNOWN_SD		? "(Unknown SD)           " :
				system == PCL_VSYSTEM_625I_50			? " 625i50                " :
				system == PCL_VSYSTEM_525I_59			? " 525i59.94             " :
				system == PCL_VSYSTEM_1080I_PSF_50		? "1080i50 or psF25       " :
				system == PCL_VSYSTEM_1080I_PSF_59_60	? "1080i59.94 or 60 or psF" :
				system == PCL_VSYSTEM_1080PSF_23_24		? "1080psF23.98 or 24     " :
				system == PCL_VSYSTEM_720P_50			? " 720p50                " :
				system == PCL_VSYSTEM_720P_59_60		? " 720p59.94 or 60       " :
				system == PCL_VSYSTEM_1080P_25			? "1080p25                " :
				system == PCL_VSYSTEM_1080P_29_30		? "1080p29.97 or 30       " :
				system == PCL_VSYSTEM_1080P_23_24		? "1080p23.98 or 24       " :
				system == PCL_VSYSTEM_720P_23_24		? " 720p23.98 or 24       " :
				system == PCL_VSYSTEM_720P_25			? " 720p25                " :
				system == PCL_VSYSTEM_720P_29_30		? " 720p29.97 or 30       " :
				system == PCL_VSYSTEM_1035I_59_60		? "1035i59.94 or 60       " :
				system == PCL_VSYSTEM_UNKNOWN_HD		? "(Unknown HD)           " :
				system == PCL_VSYSTEM_UNKNOWN_3G		? "(Unknown 3G)           " :
				system == PCL_VSYSTEM_1080P_47_48		? "1080p47.95 or 48, Lev.A" :
				system == PCL_VSYSTEM_1080P_50			? "1080p50, Level A       " :
				system == PCL_VSYSTEM_1080P_59_60		? "1080p59.94 or 60, Lev.A" :
				system == PCL_VSYSTEM_DL_1080P_47_48	? "1080p47.95 or 48, Lev.B" :
				system == PCL_VSYSTEM_DL_1080P_47		? "1080p47.95, Level B    " :
				system == PCL_VSYSTEM_DL_1080P_59		? "1080p59.94, Level B    " :
				system == PCL_VSYSTEM_DL_1080P_60		? "1080p60, Level B       " :
				system == PCL_VSYSTEM_1080I_PSF_59		? "1080i59.94 or psF29.97 " :
				system == PCL_VSYSTEM_1080I_PSF_60		? "1080i60 or psF30       " :
				system == PCL_VSYSTEM_1035I_59			? "1035i59.94             " :
				system == PCL_VSYSTEM_1035I_60			? "1035i60                " :
				// The following video standards can only be detected
				// with the help of SMPTE 352M Payload Identifier packets.
				system == PCL_VSYSTEM_720P_23			? " 720p23.98             " :
				system == PCL_VSYSTEM_720P_24			? " 720p24                " :
				system == PCL_VSYSTEM_720P_29			? " 720p29.97             " :
				system == PCL_VSYSTEM_720P_30			? " 720p30                " :
				system == PCL_VSYSTEM_720P_59			? " 720p59.94             " :
				system == PCL_VSYSTEM_720P_60			? " 720p60                " :
				system == PCL_VSYSTEM_1080I_50			? "1080i50                " :
				system == PCL_VSYSTEM_1080I_59			? "1080i59.94             " :
				system == PCL_VSYSTEM_1080I_60			? "1080i60                " :
				system == PCL_VSYSTEM_1080PSF_23		? "1080psF23.98           " :
				system == PCL_VSYSTEM_1080PSF_24		? "1080psF24              " :
				system == PCL_VSYSTEM_1080PSF_25		? "1080psF25              " :
				system == PCL_VSYSTEM_1080PSF_29		? "1080psF29.97           " :
				system == PCL_VSYSTEM_1080PSF_30		? "1080psF30              " :
				system == PCL_VSYSTEM_1080P_23			? "1080p23.98             " :
				system == PCL_VSYSTEM_1080P_24			? "1080p24                " :
				system == PCL_VSYSTEM_1080P_29			? "1080p29.97             " :
				system == PCL_VSYSTEM_1080P_30			? "1080p30                " :
				system == PCL_VSYSTEM_1080P_47			? "1080p47.95, Level A    " :
				system == PCL_VSYSTEM_1080P_48			? "1080p48, Level A       " :
				system == PCL_VSYSTEM_1080P_59			? "1080p59.94, Level A    " :
				system == PCL_VSYSTEM_1080P_60			? "1080p60, Level A       " :
				system == PCL_VSYSTEM_DL_1080P_48		? "1080p48, Level B       " :
				system == PCL_VSYSTEM_DL_1080P_50		? "1080p50, Level B       " :
				system == PCL_VSYSTEM_625P_50			? " 625p50                " :
				system == PCL_VSYSTEM_525P_59			? " 525p59.94             " :
				system == PCL_VSYSTEM_UNKNOWN			? "(Unknown)	          " :
				system == PCL_VSYSTEM_1080I_47_48_T		? "1080i47.95 or 48 or psF" :
				system == PCL_VSYSTEM_1080P_47_48_T		? "1080p47.95 or 48       " :
				system == PCL_VSYSTEM_1080P_50_T		? "1080p50                " :
				system == PCL_VSYSTEM_1080P_59_60_T		? "1080p59.94 or 60       " :
														  "(Undefined)            "
			,	progressive ? "P" : " "
			);
		}

		break;

	case PCL_IMASK_V2:	// handle V2 interrupt (even video field)
		gotoxy(19, 10);
		printfw("V2 (even)");	// show video field

		int activeWords;		// active words / line
		int totalWords;			// total words / line
		int activeLines;		// active lines / field
		int totalLines;			// total lines / frame
		if ((result = pclGetVideoRaster(hdl, &activeWords, &totalWords, &activeLines, &totalLines)) == PCL_OK) {
			gotoxy(19, 11);
			printfw("Active Words:%4d, Lines:%4d - Total Words:%4d, Lines:%4d"
			,	activeWords		// active words / line
			,	activeLines		// active lines / field
			,	totalWords		// total words / line
			,	totalLines		// total lines / frame
			);
		}

		break;

#ifdef PCL_OPT_GPI_IN_EVNT		// gpi input events
	case PCL_IMASK_GPI:
	{
		WORD gpii, change;
		int reg;
		if ((result = pclGetGpiEvents(hdl, &gpii, &change)) == PCL_OK) {
			if (change & PCL_GPI_1) {
				if ((result = pclGetGpiEvent(hdl, PCL_GPI_1, &time, &bits, &user, &reg)) == PCL_OK) {
					gotoxy(62, 3);
					printfw("%s"
					,	gpii & PCL_GPI_1 ? "    " : "GPI1"
					);
					gotoxy(55, 4);
					printfw("%2x:%02x:%02x:%02x"
					,	HIBYTE(HIWORD(time))
					,	LOBYTE(HIWORD(time))
					,	HIBYTE(LOWORD(time))
					,	LOBYTE(LOWORD(time))
					);
					gotoxy(58, 5);
					printfw("%08lx"
					,	user
					);
				}
			}
			if (change & PCL_GPI_2) {
				if ((result = pclGetGpiEvent(hdl, PCL_GPI_2, &time, &bits, &user, &reg)) == PCL_OK) {
					gotoxy(75, 3);
					printfw("%s"
					,	gpii & PCL_GPI_2 ? "    " : "GPI2"
					);
					gotoxy(68, 4);
					printfw("%2x:%02x:%02x:%02x"
					,	HIBYTE(HIWORD(time))
					,	LOBYTE(HIWORD(time))
					,	HIBYTE(LOWORD(time))
					,	LOBYTE(LOWORD(time))
					);
					gotoxy(71, 5);
					printfw("%08lx"
					,	user
					);
				}
			}
		}
		break;
	}

#endif // PCL_OPT_GPI_IN_EVNT

	default:	// unknown interrupt source
		break;	// don't handle interrupt
	}
}

//---------------------------------------------------------------------------

int main( // main function
	int argn,
	char **argc
) {
#ifdef LINUX
	initscr();
	raw();			// disable line buffering
	keypad(stdscr, TRUE);	// make F1, F2, ... accessible
	noecho();		// don't echo on getch
	halfdelay(1);		// wait 1/10 th second for key press
#endif // LINUX

	int result;						// return code of pclXxx() functions
	int hdl;						// handle to access pcl
	int base = PCL_BASE_PCI_1;		// address of pcl card
	BOOL running = TRUE;
	int ident, bytea, byteb, bytec, firmware, dll;
	BOOL hasAncRead = FALSE;

	if (argn > 1) {
		if        (sscanf(argc[1], "%d",   &base) == 1 && legalBase(base)) {
		} else if (sscanf(argc[1], "%x",   &base) == 1 && legalBase(base)) {
		} else if (sscanf(argc[1], "0x%x", &base) == 1 && legalBase(base)) {
		} else if (sscanf(argc[1], "0X%x", &base) == 1 && legalBase(base)) {
		} else if (sscanf(argc[1], "$%x",  &base) == 1 && legalBase(base)) {
		} else {
#ifdef LINUX
			cbreak();
#endif
			printfw(
				"Error in parameter base_address: %s\n"
				"Please try:  IntTest base_address\n\n"
				"Press any key to continue..."
			,	argc[1]
			);
			_getch();
			exit(1);
		}
	}
	if ((result = pclOpen(&hdl, base)) != PCL_OK) {
									// open pcl card and get handle
		printfw(
			"Unable to find PCL card at address %d (0x%X)\n"
			"Please try:  IntTest base_address\n\n"
		,	base
		,	base
		);
		fatal(FALSE, "pclOpen", result);
	}

	// the following code initializes the PCL board
	if ((result = pclReset(hdl, PCL_IMODE_NORMAL)) < PCL_OK) {
		fatal(FALSE, "pclReset", result);
	}

	{ // get some version informations from pcl
		if ((result = pclGetVersion(hdl, &ident, &firmware, &dll))
			== PCL_OK
		) {	// get version information
			if ((result = pclGetVerSpecial(hdl, &bytea, &byteb, &bytec))
				!= PCL_OK
			) {	// get special version information
				bytea = byteb = bytec = 0;
				warning(FALSE, result);
			}
		} else {
			ident = firmware = dll = bytea = byteb = bytec = 0;
			warning(FALSE, result);
		}

		if (dll < PCL_VER_DLL) {	// check if DLL version is o.k.
			gotoxy(1, 15);
			printfw(
				"Warning: Version of DLL is V%d.%02d, "
				"but has to be V%d.%02d or higher.\n"
			,	dll >> 8, dll & 0xFF
			,	PCL_VER_DLL >> 8, PCL_VER_DLL & 0xFF
			);
		}
	}

	// is the card able to read atc?
	switch (bytec) {
	case PCL_ID_PCI_HD:
	case PCL_ID_PCI_D_V2:
	case PCL_ID_PCIE_HD:
	case PCL_ID_PCIE_D:
	case PCL_ID_PCIE_3G:
	case PCL_ID_PCIE_3G_V2:
		hasAncRead = TRUE;
	}

	// build screen mask
	cursorVisible(FALSE);	// hide cursor
	clrscr(); // first clear screen
	gotoxy(1,1);
	printfw(
		"PLURA PCL Cards - IntTest Version " VER_S " - Card 0x%x\n"
		"\n"
		"Mixed     (int) : --:--:--:--  --------\n"
		"\n"
		"VITC      (int) : --:--:--:--  --------\n"
		"LTC       (int) : --:--:--:--  --------\n"
	,	base
	);
	gotoxy(1, 22);
	printfw(
		"Version         : --\n"
		"\n"
		"eX)it (esc)    D)isable Interrupts    E)nable Interrupts\n"
//	,	base
	);

	if (hasAncRead) {
		gotoxy(1, 7);
		printfw(
			"ATC LTC   (int) : --:--:--:--  --------\n"
			"ATC VITC  (int) : --:--:--:--  --------\n"
			"\n"
			"\n"
			"Video Geometry  : --\n"
//			"SMPTE 352M      : --\n"
		);
	}

	switch (bytec) {
	case PCL_ID_PCI_L:
	case PCL_ID_PCI_TS:
	case PCL_ID_PCIE_L:
	case PCL_ID_PCIE_L_V2:
	case PCL_ID_PCIE_TS:
	case PCL_ID_PCIE_TS_V2:
		break;
	default:
		gotoxy(1, 10);
		printfw(
			"Video           : --\n"
		);
	}

	// show version informations
	gotoxy(19, 22);

	// pcl identifier is coded in bit 6 and 7 of
	//	firmware version
	printfw("%s"
	,	(firmware & 0xC0) == 0x00 ? "PCL6 or PCL3"	:
		(firmware & 0xC0) == 0x40 ? "PCL PCI"		:
		(firmware & 0xC0) == 0x80 ? "PCL5 or PCL10"	:
		(firmware & 0xC0) == 0xC0 ? "PCL12"			:
									"unknown"
	);
	switch (bytec) {
	case PCL_ID_PCI_LV_D:	printfw(" LV or D");	break;
	case PCL_ID_PCI_LV:		printfw(" LV");			break;
	case PCL_ID_PCI_D:		printfw(" D");			break;
	case PCL_ID_PCI_L:		printfw(" L");			break;
	case PCL_ID_PCI_V:		printfw(" V");			break;
	case PCL_ID_PCI_TS:		printfw(" TS");			break;
	case PCL_ID_PCI_D_V2:	printfw(" D (v2)");		break;
	case PCL_ID_PCI_L_V2:	printfw(" L (v2)");		break;
	case PCL_ID_PCI_HD:		printfw(" HD");			break;
	case PCL_ID_PCIE_L:		printfw("e L");			break;
	case PCL_ID_PCIE_L_V2:	printfw("e L (v2)");	break;
	case PCL_ID_PCIE_LV:	printfw("e LV");		break;
	case PCL_ID_PCIE_LV_V2:	printfw("e LV (v2)");	break;
	case PCL_ID_PCIE_TS:	printfw("e TS");		break;
	case PCL_ID_PCIE_TS_V2:	printfw("e TS (v2)");	break;
	case PCL_ID_PCIE_HD:	printfw("e HD");		break;
	case PCL_ID_PCIE_D:		printfw("e D");			break;
	case PCL_ID_PCIE_3G:	printfw("e 3G");		break;
	case PCL_ID_PCIE_3G_V2:	printfw("e 3G (v2)");	break;
	default:
		printfw(", %s (%c%c/%c)"
		,	ident ? "special" : "no special"
		,	byteb, bytea, bytec	// special oem version of pcl?
		);
		break;
	}

	// pcl firmware version
	printfw(", Firmware: V%d", firmware & 0x3F);

	switch (bytec) {
	case PCL_ID_PCI_HD:
	case PCL_ID_PCI_D_V2:
	case PCL_ID_PCI_L_V2:
	case PCL_ID_PCIE_L:
	case PCL_ID_PCIE_L_V2:
	case PCL_ID_PCIE_LV:
	case PCL_ID_PCIE_LV_V2:
	case PCL_ID_PCIE_TS:
	case PCL_ID_PCIE_TS_V2:
	case PCL_ID_PCIE_HD:
	case PCL_ID_PCIE_D:
	case PCL_ID_PCIE_3G:
	case PCL_ID_PCIE_3G_V2:
		printfw(".%c%c", byteb, bytea);	break;
	}

	// version of device driver
	printfw(", DLL: V%d.%02d", dll >> 8, dll & 0xFF);
	gotoxy(1, 20);

	// set the vitc reader mode.  we want to read from line 6 to 22
	if ((result = pclVitcSetup(hdl, 6, 22, PCL_VLINE_BLOCK)) < PCL_OK) {
		warning(TRUE, result);
	}

	// set the frame rate to auto
	if ((result = pclTcFrames(hdl, PCL_TCFR_AUTO)) < PCL_OK) {
		warning(TRUE, result);
	}

#ifdef PCL_OPT_GPI_IN_EVNT		/* gpi input events */

	if ((result = pclSetGpiEvent(hdl, PCL_GPI_EVNT_ENABLE_HL_1, PCL_REG_MIXED)) <= PCL_OK) {
		warning(TRUE, result);
	}
	if ((result = pclSetGpiEvent(hdl, PCL_GPI_EVNT_ENABLE_LH_1, PCL_REG_MIXED)) <= PCL_OK) {
		warning(TRUE, result);
	}
	if ((result = pclSetGpiEvent(hdl, PCL_GPI_EVNT_ENABLE_HL_2, PCL_REG_MIXED)) <= PCL_OK) {
		warning(TRUE, result);
	}
	if ((result = pclSetGpiEvent(hdl, PCL_GPI_EVNT_ENABLE_LH_2, PCL_REG_MIXED)) <= PCL_OK) {
		warning(TRUE, result);
	}

#	define PCL_IMASK ( \
		PCL_IMASK_V1 | \
		PCL_IMASK_V2 | \
		PCL_IMASK_MIXED | \
		PCL_IMASK_VITC | \
		PCL_IMASK_LTC | \
		PCL_IMASK_ATC_LTC | \
		PCL_IMASK_ATC_VITC | \
		PCL_IMASK_GPI \
	)										// PCL interrupts to enable

#else // PCL_OPT_GPI_IN_EVNT

#	define PCL_IMASK (0 \
	|	PCL_IMASK_V1 \
	|	PCL_IMASK_V2 \
	|	PCL_IMASK_MIXED \
	|	PCL_IMASK_VITC \
	|	PCL_IMASK_LTC \
	|	PCL_IMASK_ATC_LTC \
	|	PCL_IMASK_ATC_VITC \
	)										// PCL interrupts to enable

#endif // PCL_OPT_GPI_IN_EVNT

	if ((result = pclIntMask(hdl, 0)) < PCL_OK) {
		fatal(TRUE, "pclIntMask", result);
	}
	if ((result = pclSetIntHandler(hdl, intHandler)) < PCL_OK) {
		fatal(TRUE, "pclSetIntHandler", result);
	}
	if ((result = pclIntMask(hdl, PCL_IMASK)) < PCL_OK) {
		fatal(TRUE, "pclIntMask", result);
	}

	// the following might be enabled for time critical interrupt service
/*
	if ((result = pclSetIntPriority(hdl, THREAD_PRIORITY_TIME_CRITICAL)) < PCL_OK) {
		warning(TRUE, result);
	}
*/
	while (running) { // main program loop
		int ch;

		switch (ch = _getch()) {	// wait until there is
									//	a keystroke waiting
			case 'd': case 'D':	// disable all interrupts
				if ((result = pclIntMask(hdl, 0)) < PCL_OK) {
					warning(TRUE, result);
				}
				break;
			case 'e': case 'E':	// enable all interrupts
				if ((result = pclIntMask(hdl, PCL_IMASK)) < PCL_OK) {
					warning(TRUE, result);
				}
				break;
			case 'x': case 'X': case '\x1b':
				running = FALSE;
				break;
		}
	}

	// it's important to close the device if we don't need it anymore
	pclClose(hdl);
	gotoxy(1, 24);
	cursorVisible(TRUE);	// show cursor
#ifdef LINUX
	endwin();
#endif
	return(0);
}
